/*
 * Filename: NSApplication.m
 * Project:  NSXKit
 * Module:   Main application components
 *
 * Function: The NSApplication class manages the main event loop of
 *           the program, keeps track of the application's windows
 *           and which one is the key window.
 *
 * Author:   Pascal Forget
 * Date:     January 1996
 *
 * Copyright (C) 1996 Free Software Foundation, Inc.
 */
#include "NSApplication.h"
#include "NSWindow.h"
#include "NSImage.h"
#include "NSResponder.h"
#include <unistd.h>
#include <X11/keysymdef.h>
#include "PXKEventFactory.h"
#include "misc.h"

NSWindow *keyWindow;
NSWindow *mainWindow;
 
static char *progname; /* name this program was invoked by */

NSMutableArray *_allWindows; /* The app's windows are registered */

/*
 * Current event settings stored here as we don't want to
 * waste CPU cycles creating / destroying these variables
 * each time we go through a cycle in the event loop
 */
XEvent xEvent;
NSEvent *_currentEvent;
NSEventType eventType;
NSEvent *NULL_EVENT = nil;
unsigned short eventModifierFlags;
NSPoint eventLocation;
unsigned int eventFlags;
NSTimeInterval eventTime;
KeySym keysym;
unsigned short keyCode;

// Global strings
NSString *NSModalPanelRunLoopMode = @"ModalPanelMode";
NSString *NSEventTrackingRunLoopMode = @"EventTrackingMode";

//
// Global Exception Strings 
//
NSString *NSAbortModalException = @"AbortModal";
NSString *NSAbortPrintingException = @"AbortPrinting";
NSString *NSAppKitIgnoredException = @"AppKitIgnored";
NSString *NSAppKitVirtualMemoryException = @"AppKitVirtualMemory";
NSString *NSBadBitmapParametersException = @"BadBitmapParameters";
NSString *NSBadComparisonException = @"BadComparison";
NSString *NSBadRTFColorTableException = @"BadRTFColorTable";
NSString *NSBadRTFDirectiveException = @"BadRTFDirective";
NSString *NSBadRTFFontTableException = @"BadRTFFontTable";
NSString *NSBadRTFStyleSheetException = @"BadRTFStyleSheet";
NSString *NSBrowserIllegalDelegateException = @"BrowserIllegalDelegate";
NSString *NSColorListIOException = @"ColorListIO";
NSString *NSColorListNotEditableException = @"ColorListNotEditable";
NSString *NSDraggingException = @"Draggin";
NSString *NSFontUnavailableException = @"FontUnavailable";
NSString *NSIllegalSelectorException = @"IllegalSelector";
NSString *NSImageCacheException = @"ImageCache";
NSString *NSNibLoadingException = @"NibLoading";
NSString *NSPPDIncludeNotFoundException = @"PPDIncludeNotFound";
NSString *NSPPDIncludeStackOverflowException = @"PPDIncludeStackOverflow";
NSString *NSPPDIncludeStackUnderflowException = @"PPDIncludeStackUnderflow";
NSString *NSPPDParseException = @"PPDParse";
NSString *NSPasteboardCommunicationException = @"PasteboardCommunication";
NSString *NSPrintOperationExistsException = @"PrintOperationExists";
NSString *NSPrintPackageException = @"PrintPackage";
NSString *NSPrintingCommunicationException = @"PrintingCommunication";
NSString *NSRTFPropertyStackOverflowException = @"RTFPropertyStackOverflow";
NSString *NSTIFFException = @"TIFF";
NSString *NSTextLineTooLongException = @"TextLineTooLong";
NSString *NSTextNoSelectionException = @"TextNoSelection";
NSString *NSTextReadException = @"TextRead";
NSString *NSTextWriteException = @"TextWrite";
NSString *NSTypedStreamVersionException = @"TypedStreamVersion";
NSString *NSWindowServerCommunicationException = @"WindowServerCommunication";
NSString *NSWordTablesReadException = @"WordTablesRead";
NSString *NSWordTablesWriteException = @"WordTablesWrite";

// Application notifications
NSString *NSApplicationDidBecomeActiveNotification 
              = @"ApplicationDidBecomeActive";
NSString *NSApplicationDidFinishLaunchingNotification 
              = @"ApplicationDidFinishLaunching";
NSString *NSApplicationDidHideNotification = @"ApplicationDidHide";
NSString *NSApplicationDidResignActiveNotification 
              = @"ApplicationDidResignActive";
NSString *NSApplicationDidUnhideNotification = @"ApplicationDidUnhide";
NSString *NSApplicationDidUpdateNotification = @"ApplicationDidUpdate";
NSString *NSApplicationWillBecomeActiveNotification 
              = @"ApplicationWillBecomeActive";
NSString *NSApplicationWillFinishLaunchingNotification 
              = @"ApplicationWillFinishLaunching";
NSString *NSApplicationWillHideNotification = @"ApplicationWillHide";
NSString *NSApplicationWillResignActiveNotification 
              = @"ApplicationWillResignActive";
NSString *NSApplicationWillUnhideNotification = @"ApplicationWillUnhide";
NSString *NSApplicationWillUpdateNotification = @"ApplicationWillUpdate";



@implementation NSApplication

- createSystemImages;
{
    return self;
}

- init;
{
    [super init];
    initializeDrawingEngine();
    [self createSystemImages];

    _allWindows = [[NSMutableArray alloc] init];

    _eventQueue = [[NSMutableArray alloc] init];

    // NULL EVENT
    NULL_EVENT = [[NSEvent alloc] init];
    
    return self;
}

+ (NSApplication *)sharedApplication;
{
    if (NSApp == nil) {  
	NSApp = [[NSApplication alloc] init]; 
    }

    return NSApp; /* NSApp global variable declared in NSGlobal.h */
}

- (void)activateIgnoringOtherApps:(BOOL)flag;
{
    isActive = YES; /* primitive implementation while we wait for DPS */
}

- (void)deactivate;
{
    isActive = NO;
}

- (void)finishLaunching;
{
    isActive = YES;
    
    // Should send notifications here
}

- (BOOL)isActive;
{
    return isActive;
}

- (void)abortModal;
{
    /* not implemented */
}

- (NSModalSession)beginModalSessionForWindow:(NSWindow *)theWindow;
{
    return 0; /* Not implemented */
}

- (void)endModalSession:(NSModalSession)session;
{
}

- (BOOL)isRunning;
{
    return isRunning;
}

- (void)eventProcessing
{
    return;
}

- (void)run;
{
    NSEvent *event;
    NSEventType type;    
    NSAutoreleasePool *eventLoopPool;
    
    [self finishLaunching];
    
    isRunning = YES;
   
    /*
     * This is the event loop
     */
    
    while (isRunning)  {

	if (isActive == NO) {
	    continue;
	}
	
//	eventLoopPool = [[NSAutoreleasePool alloc] init];
	eventLoopPool = [NSAutoreleasePool new];

	if ([_eventQueue count] == 0) {
	  event = nextEvent();
	} else {
	  event = [_eventQueue lastObject];
	  [_eventQueue removeLastObject];
	}

        if (event == NULL_EVENT) {
	  dbug(msg("Got NULL event!\n"));
        }

	// _currentEvent = event;
	

	// This event processing should be moved to
	// self sendEvent:

	if ((event) && (mainWindow != nil)) {
	  // does nothing
	  // [self eventProcessing];
	    
	  type = [event type];

	  // This is pish
	  switch (type) {
	  case NSMouseMoved : {
	    [mainWindow mouseMoved:event];
	    break;
	  }
	  case NSLeftMouseDragged : {
	    [mainWindow mouseDragged:event];
	    break;
	  }
	  case NSRightMouseDragged : {
	    [mainWindow rightmouseDragged:event];
	    break;
	  }

	  case NSLeftMouseUp : {
	    [mainWindow mouseUp:event];
	  }
	  case NSLeftMouseDown : {
	    [mainWindow mouseDown:event];
	    break;
	  }
	  
	  case NSRightMouseDown : {
	    [mainWindow rightMouseDown:event];
	    break;
	  }

	  case NSRightMouseUp : {
	    [mainWindow rightMouseUp:event];
	    break;
	  }

	  case NSMouseEntered : {
	    [mainWindow mouseEntered:event];
	    break;
	  }

	  case NSMouseExited : {
	    [mainWindow mouseExited:event];
	    break;
	  }

	  case NSKeyDown : {
	    [mainWindow keyDown:event];
	    break;
	  }

	  case NSKeyUp : {
	    [mainWindow keyUp:event];
	    break;
	  }

	  case NSFlagsChanged : {
	    [mainWindow flagsChanged:event];
	    break;
	  }

	  case NSPeriodic : {
	    /* dunno */
	  }

	  case NSCursorUpdate : {
	    /* dunno */
	  }


	  }

	}
	
	[eventLoopPool release];
 
    } /* end event loop */
}

- (int)runModalForWindow:(NSWindow *)theWindow;
{
    return -1; /* Not implemented */
}

- (int)runModalsession:(NSModalSession)session;
{
    return -1;
}

- (void)sendEvent:(NSEvent *)theEvent;
{
    /* Not implemented.  Called from the run method??? */
}

- (void)stop:(id)sender;
{
    isRunning = NO;
}

- (void)stopModal;
{
}

- (void)stopModalWithCode:(int)returnCode;
{
}

- (NSEvent *)currentEvent;
{
    return _currentEvent;
}

- (void)discardsEventsMatchingEventMask:(unsigned int)eventMask
		       beforeEvent:(NSEvent *)lastEvent;
{
}

- (NSEvent *)nextEventMatchingMask:(unsigned int)mask
			 untilDate:(NSDate *)expiration
			    inMode:(NSString *)mode
			   dequeue:(BOOL)flag;
{
    BOOL noMatch = YES;
    NSEvent *event;
    NSDate *now;
    
    while (noMatch)
    {
	now = [NSDate dateWithTimeIntervalSinceNow: 0];
  
	if ([now compare: expiration] == NSOrderedDescending)
	{
	    return nil;
	}

	event = nextEvent();
	
	if ((mask & NSLeftMouseDragged) &&
	    ([event type] == NSLeftMouseDragged))
	{
	    noMatch = NO;
	}
      	
	if ((mask & NSRightMouseDragged) &&
	    ([event type] == NSRightMouseDragged))
	{
	    noMatch = NO;
	}
      
	if ((mask & NSLeftMouseUp) &&
	    ([event type] == NSLeftMouseUp))
	{
	    noMatch = NO;
	}
	    
	if ((mask & NSRightMouseUp) &&
	    ([event type] == NSRightMouseUp))
	{
	    noMatch = NO;
	}
	    	
	if ((mask & NSMouseEntered) &&
	    ([event type] == NSMouseEntered))
	{
	    noMatch = NO;
	}
	
	if ((mask & NSMouseMoved) &&
	    ([event type] == NSMouseMoved))
	{
	    noMatch = NO;
	}

	if ((mask & NSMouseExited) &&
	    ([event type] == NSMouseExited))
	{
	    noMatch = NO;
	}

	if ((mask & NSKeyDown) &&
	    ([event type] == NSKeyDown))
	{
	    noMatch = NO;
	}

	if ((mask & NSKeyUp) &&
	    ([event type] == NSKeyUp))
	{
	    noMatch = NO;
	}
	
	if ((mask & NSPeriodic) &&
	    ([event type] == NSPeriodic))
	{
	    noMatch = NO;
	}
	    
	if ((mask & NSFlagsChanged) &&
	    ([event type] == NSFlagsChanged))
	{
	    noMatch = NO;
	}

	if ((mask & NSCursorUpdate) &&
	    ([event type] == NSCursorUpdate))
	{
	    noMatch = NO;
	}
    }
    
    if (flag == NO) // Keep this event around
    {
	[self postEvent: event atStart: YES];
    }
   
    return event;
}

- (void)postEvent:(NSEvent *)event atStart:(BOOL)flag;
{
    if (event == nil)
    {
	return;
    }
    
    if (flag == YES)
    {
	[_eventQueue addObject: event];
    } else {
	[_eventQueue insertObject: event atIndex: 0];
    }
}

- (BOOL)sendAction:(SEL)aSelector to:(id)aTarget from:(id)sender;
{
    return NO; /* Not implemented */
}

- (id)targetForAction:(SEL)aSelector;
{
    return nil; /* Not implemented */
}

- (BOOL)tryToPerform:(SEL)aSelector with:(id)anObject;
{
    return NO; /* Not implemented */
}

- (void)setApplicationIconImage:(NSImage *)anImage;
{
    if (appIcon != nil)
    {
	[appIcon release];
    }
    
    appIcon = [anImage retain];
}

- (NSImage *)applicationIconImage;
{
    return appIcon;
}

- (void)hide:(id)sender;
{
    isHidden = YES;

    /* Iconize / hide all windows */
}
    
- (BOOL)isHidden;
{
    return isHidden;
}

- (void)unhide:(id)sender;
{
    isHidden = NO;

    /* Display all the app windows */
}

- (void)unhideWithoutActivation;
{
    [self unhide]; /* default implementation to be changed eventually */
}

- (NSWindow *)keyWindow;
{
    return keyWindow;
}
    
- (NSWindow *)mainWindow;
{
    return mainWindow;
}

- (NSWindow *)makeWindowsPerform:(SEL)aSelector inOrder:(BOOL)flag;
{
    if (flag)
    {
	[_allWindows makeObjectsPerform:aSelector];
    }
    else
    {
	[_allWindows makeObjectsPerformFromTheEnd:aSelector];
    }
    return nil; /* which window am I supposed to return? */
}

- (void)miniaturizeAll:(id)sender;
{
    [_allWindows makeObjectsPerform:@selector(miniaturize)];
}

- (void)preventWindowOrdering;
{
}

- (void)setWindowsNeedUpdate:(BOOL)flag;
{
    windowsNeedUpdate = flag;
}

- (void)updateWindows;
{
    int i, count;
    NSWindow *aWindow;

    count = [_allWindows count];

    for (i=0; i<count; i++)
    {
	aWindow = [_allWindows objectAtIndex:i];

	if ([aWindow isVisible])
	{
	    [aWindow update];
	}
    }
}
    
- (NSArray *)windows;
{
    return [NSArray arrayWithArray:_allWindows];
}

- (NSWindow *)windowWithWindowNumber:(int)windowNum;
{
    int i, count;
    NSWindow *aWindow;

    count = [_allWindows count];

    for (i=0; i<count; i++)
    {
	aWindow = [_allWindows objectAtIndex:i];

	if ([aWindow windowNumber] == windowNum)
	{
	    return aWindow;
	}
    }
    return nil; /* No window was found having windowNum */
}    

- (void)terminate:(id)sender;
{
    exit(0);
}

@end

@implementation NSApplication (NonOpenStep)

- (void)registerWindow:(NSWindow *)newWindow;
{
    [_allWindows addObject:newWindow];
}

- (void)removeWindow:(NSWindow *)defunctWindow;
{
    [_allWindows removeObject:defunctWindow];
}

- (void)setKeyWindow:(NSWindow *)aWindow;
{
    keyWindow = aWindow;
}

- (void)setMainWindow:(NSWindow *)aWindow;
{
    keyWindow = aWindow;
    mainWindow = aWindow;
}

- (void)updateKeyAndMainWindow;
{
    NSWindow *keyWindow = [self keyWindow];

#ifdef NSPANEL_H__
    if ([keyWindow isKindOfClass:[NSPanel class]] == NO) {
	mainWindow = keyWindow;
    }
#else /* class NSPanel not implemented yet */
    if (YES) {
	mainWindow = keyWindow;
    }
#endif
}

@end
