/* 
   NSAttributedString.m

   Implementation of AppKit additions for NSAttributedString class

   Copyright (C) 1997 Free Software Foundation, Inc.

   Written by: ANOQ of the Sun <anoq@vip.cybercity.dk>
   Date: November 1997
   
   This file is part of NSXKit

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// FIXME:
// Methods not implemented in NSAttributedString:
//   - initWithRTF:documentAttributes:docAttributes
//   - initWithRTFD:documentAttributes:docAttributes
//   - initWithRTFDFileWrapper:wrapper documentAttributes:docAttributes
//   - initWithPath:path documentAttributes:docAttributes
//   - RTFFromRange:aRange documentAttributes:docAttributes
//   - RTFDFileWrapperFromRange:aRange documentAttributes:docAttributes
//   - RTFDFromRange:aRange documentAttributes:docAttributes
//   - size
// Incomplete methods:
//   - drawInRect:  does not take line-sweep-direction into account
//       These next 3 methods doesn't take language characteristics into account
//   - nextWordFromIndex:forward:
//   - lineBreakBeforeIndex:withinRange:
//   - doubleClickAtIndex:
// Methods not implemented in NSMutableAttributedString
//   - updateAttachmentsFromPath:path
//   - setAlignment:
//   - applyFontTraits:
//   - fixFontAttributeInRange:
//   - fixParagraphStyleAttributeInRange:

#include "NSAttributedString.h"
#include <Foundation/NSException.h>
#include <Foundation/NSValue.h>
#include "PXKDrawing.h"

NSString *NSFontAttributeName = @"NSFont";
NSString *NSForegroundColorAttributeName = @"NSColor";
NSString *NSBackgroundColorAttributeName = @"NSBackgroundColor";
NSString *NSUnderlineStyleAttributeName = @"NSUnderline";
NSString *NSSuperscriptAttributeName = @"NSSuperScript";
NSString *NSBaselineOffsetAttributeName = @"NSBaselineOffset";
NSString *NSKernAttributeName = @"NSKern";
NSString *NSLigatureAttributeName = @"NSLigature";
NSString *NSParagraphStyleAttributeName = @"NSParagraphStyle";
NSString *NSAttachmentAttributeName = @"NSAttachment";

@implementation NSAttributedString (NSAttributedStringAdditions)

void _testMoveAttr(NSDictionary *fromDict,
                   NSString **keys,
                   id *objects,
                   int *count,
                   NSString *key)
{
  id attr;
  
  attr = [fromDict objectForKey:key];
  if(attr)
  {
    keys[*count] = key;
    objects[*count] = attr;
    (*count)++;
  }
}

//Creating an NSAttributedString
-(id)initWithRTF:(NSData *)rtfData documentAttributes:(NSDictionary **)docAttributes
{
  //I believe this must be implemented in a category of NSGAttributedString
  /*Initializes a new NSAttributedString by decoding the stream of RTF commands and data contained in rtfData. Also
  returns by reference in docAttributes a dictionary containing document-level attributes, as listed in the class 
  cluster description under RTF Document Attributes. docAttributes may be NULL, in which case no document attributes 
  are returned. Returns self, or nil if rtfData can't be decoded.*/
  return nil;
}

-(id)initWithRTFD:(NSData *)rtfdData documentAttributes:(NSDictionary **)docAttributes
{
  //I believe this must be implemented in a category of NSGAttributedString
  /*Initializes a new NSAttributedString by decoding the stream of RTFD commands and data contained in rtfdData. Also
  returns by reference in docAttributes a dictionary containing document-level attributes, as listed in the class 
  cluster description under RTF Document Attributes. docAttributes may be NULL, in which case no document attributes 
  are returned. Returns self, or nil if rtfData can't be decoded.
   initWithRTFDFileWrapper:documentAttributes:*/
  return nil;
}

-(id)initWithRTFDFileWrapper:(NSFileWrapper *)wrapper documentAttributes:(NSDictionary **)docAttributes
{
  //I believe this must be implemented in a category of NSGAttributedString
  /*Initializes a new NSAttributedString from wrapper an NSFileWrapper object containing an RTFD document. Also returns
  by reference in docAttributes a dictionary containing document-level attributes, as listed in the class cluster 
  description under RTF Document Attributes. docAttributes may be NULL, in which case no document attributes are 
  returned. Returns self, or nil if wrapper can't be interpreted as an RTFD document.
   lineBreakBeforeIndex:withinRange:*/
  return nil;
}

-(id)initWithPath:(NSString *)path documentAttributes:(NSDictionary **)docAttributes
{
  //I believe this must be implemented in a category of NSGAttributedString
  /*Initializes a new NSAttributedString from RTF or RTFD data contained in the file at path. Also returns by reference 
  in docAttributes a dictionary containing document-level attributes, as listed in the class cluster description under 
  RTF Document Attributes. docAttributes may be NULL, in which case no document attributes are returned. Returns 
  self, or nil if rtfData can't be decoded.*/
  return nil;
}

+ (NSAttributedString *)attributedStringWithAttachment:(NSTextAttachment *)attachment
{
  NSAttributedString *newStr;
  NSDictionary *attrDict;
  NSString *tmpStr;
  unichar attChar = NSAttachmentCharacter;
 
  attrDict = [NSDictionary dictionaryWithObjects:&attachment
    forKeys:&NSAttachmentAttributeName count:1];
  tmpStr = [[NSString alloc]
    initWithCharacters:&attChar length:1];
  [tmpStr autorelease];
  newStr = [[NSAttributedString alloc]
    initWithString:tmpStr attributes:attrDict];
  return [newStr autorelease];
}

//Retrieving attribute information
- (NSDictionary *)fontAttributesInRange:(NSRange)aRange
{
  NSString *keys[8];
  id objects[8];
  int cnt = 0;
  NSDictionary *allAttrs;
  
  if(NSMaxRange(aRange) > [self length])
  {
    [NSException raise:NSRangeException format:
      @"RangeError in method -fontAttributesInRange: in class NSAttributedString (AttributedStringAdditions)"];
  }
  allAttrs = [self attributesAtIndex:aRange.location effectiveRange:NULL];
  
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSFontAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSForegroundColorAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSBackgroundColorAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSUnderlineStyleAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSSuperscriptAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSBaselineOffsetAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSKernAttributeName);
  _testMoveAttr(allAttrs,keys,objects,&cnt,NSLigatureAttributeName);

  if(cnt)
  {
    return [NSDictionary dictionaryWithObjects:objects
      forKeys:keys count:cnt];
  }
  else
  {
    return [NSDictionary dictionary];
  }
}

- (NSDictionary *)rulerAttributesInRange:(NSRange)aRange
{
  NSDictionary *allAttrs;
  id attr;
  
  if(NSMaxRange(aRange) > [self length])
  {
    [NSException raise:NSRangeException format:
      @"RangeError in method -fontAttributesInRange: in class NSAttributedString (AttributedStringAdditions)"];
  }
  allAttrs = [self attributesAtIndex:aRange.location effectiveRange:NULL];
  
  attr = [allAttrs objectForKey:NSParagraphStyleAttributeName];
  if(attr)
  {
    return [NSDictionary dictionaryWithObjects:&attr
      forKeys:&NSParagraphStyleAttributeName count:1];
  }
  else
  {
    return [NSDictionary dictionary];
  }
}

- (BOOL)containsAttachments
{
  //Note: Key method... called in cluster's subclass

  NSRange effectiveRange = NSMakeRange(0,0);
  NSDictionary *attrs;

  while(NSMaxRange(effectiveRange) < [self length])
  {
    attrs = [self attributesAtIndex:NSMaxRange(effectiveRange)
      effectiveRange:&effectiveRange];
    if([attrs objectForKey:NSAttachmentAttributeName])
      return YES;
  }
  return NO;
}

//Calculating linguistic units
- (NSRange)doubleClickAtIndex:(unsigned int)index
{
  //FIXME: This method doesn't take language characteristics into account as it should
  NSRange result;
  int tmpLength;
  char *strPtr;
  char aChar;
  
  tmpLength = [self length];
  if (index>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSAttributedString (NSAttributedStringAdditions) in method - doubleClickAtIndex:"];
  }
  
  result = NSMakeRange(index,0);
  strPtr = (char *)[[self string] cString];

  while(result.location>0)
  {
    aChar = strPtr[result.location];
    if(aChar==' ' || aChar=='\t' || aChar=='\n')
      break;
    result.location--;
    result.length++;
  }
  while(NSMaxRange(result)<tmpLength)
  {
    aChar = strPtr[NSMaxRange(result)];
    if(aChar==' ' || aChar=='\t' || aChar=='\n')
      break;
    result.length++;
  }
  return result;
}

- (unsigned int)lineBreakBeforeIndex:(unsigned int)index withinRange:(NSRange)aRange
{
  //FIXME: This method doesn't take language characteristics into account as it should
  int tmpLength;
  char *strPtr;
  char aChar;
  
  tmpLength = [self length];
  if (index>tmpLength || aRange.location<0 || NSMaxRange(aRange)>tmpLength)
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSAttributedString (NSAttributedStringAdditions) in method - lineBreakBeforeIndex:"];
  }

  if(index<=aRange.location)
    return NSNotFound;
  if(index>NSMaxRange(aRange))
    index = NSMaxRange(aRange);

  strPtr = (char *)[[self string] cString];
  while(index>aRange.location)
  {
    index--;
    aChar = strPtr[index];
    if(aChar==' ' || aChar=='\t' || aChar=='\n')
      return index;
  }
  return NSNotFound;
}

- (unsigned int)nextWordFromIndex:(unsigned int)index forward:(BOOL)flag
{
  //FIXME: This method doesn't take language characteristics into account as it should
  int tmpLength;
  char *strPtr;
  char aChar;
  BOOL wordBroken = NO;
  BOOL afterWordBreak = NO;
  
  tmpLength = [self length];
  if (index>tmpLength)
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSAttributedString (NSAttributedStringAdditions) in method - nextWordFromIndex:"];
  }

  strPtr = (char *)[[self string] cString];

  if(flag)//It is faster to test this outside the loop
  {       //instead of doing it in each iteration...
    while(index<tmpLength)
    {
      aChar = strPtr[index];
      if(aChar==' ' || aChar=='\t' || aChar=='\n')
        wordBroken = YES;
      else if(wordBroken)
        return index;
      index++;
    }
  }
  else
  {
    while(index>=0)
    {
      aChar = strPtr[index];
      if(aChar==' ' || aChar=='\t' || aChar=='\n')
      {
        if(afterWordBreak)
          return index+1;
        wordBroken = YES;
      }
      else
      {
        if(wordBroken)
          afterWordBreak = YES;
      }
      index--;
    }
  }
  if(index<0)
    return 0;
  else if(index>tmpLength)
    return tmpLength;
  return index;
}

#ifdef 0
- (void)fillTextBuffer
{
  /*char *strPtr;
  int tmpBufferLoc;
  NSString *tmpStr;

  strPtr = textStartPtr + bufferDrawRange.location;
  tmpBufferLoc = bufferDrawRange.location + bufferDrawRange.length;
  
  while(*strPtr!=' ' && *strPtr!='\t' && *strPtr!='\n' && tmpBufferLoc<textDrawLength)
  {
    tmpBufferLoc++;
    strPtr++;
    if(selectedRange.location==tmpBufferLoc)
      break;
  }
  tmpStr = [[textContents
              substringFromRange:
                NSMakeRange(bufferDrawRange.location,
                            tmpBufferLoc-bufferDrawRange.location)]
              autorelease];
  if(*strPtr=='\n' || tmpBufferLoc>=textDrawLength)
  {
    makeLineFeed = YES;
    tmpBufferLoc++;
  }
  else
  {
    //Test if possible...
    stringWidth = pxkWidthOfString(tmpStr,defaultFont);
    if((stringWidth+currentTextPos.x)>drawSizeLimit.width)
    {
      if(possibleBreakLoc>0)
      {
        tmpBufferLoc = possibleBreakLoc;
        strPtr = textStartPtr + tmpBufferLoc;
        tmpStr = [[textContents
                    substringFromRange:
                      NSMakeRange(bufferDrawRange.location,tmpBufferLoc-bufferDrawRange.location)]
                    autorelease];
      }
      makeLineFeed = YES;
    }
    else
      possibleBreakLoc = tmpBufferLoc;
  }
  bufferDrawRange.length = (tmpBufferLoc - bufferDrawRange.location);
  if(makeLineFeed)
    possibleBreakLoc = 0;*/
}

- (void)drawTextBuffer
{
  /*NSString *tmpStr;
  
  tmpStr = [[textContents
              substringFromRange:bufferDrawRange]
              autorelease];
  pxkDrawString([tmpStr cString],currentTextPos);
  if(makeLineFeed)
  {
    currentTextPos.y += lineFeedHeight;
    currentTextPos.x = 0;
    makeLineFeed = NO;
  }
  else
    currentTextPos.x += stringWidth;
  bufferDrawRange.location += bufferDrawRange.length;
  bufferDrawRange.length = 0;*/
}

- (void)drawText
{
  NSLayInfo *layInfo;
  NSRun run;
  int cnt,maxCnt;
  NSTextBlock *block;
  int blockPos,blockCharsLeft;
  NSLineDesc lineDesc,charsCopied;
  int lineFeedHeight;
  char *textDest;

  layInfo = allocNSLayInfo();
  NS_DURING
    [textColor set];
    [defaultFont set];
    layInfo->rect.origin.x = 0;
    layInfo->rect.origin.y = 0;
    layInfo->rect.size = [self frame].size;
    memset(&run,sizeof(NSRun),0);
    run.font = defaultFont;
    pxkCalcLineBreaks(layInfo,&run,firstTextBlock,breakArray);
    maxCnt = breakArray->chunk.used/sizeof(NSLineDesc);
    block = firstTextBlock;
    blockPos = 0;
    lineFeedHeight = [defaultFont pointSize];
    layInfo->lays->lays[0].x = 0;
    layInfo->lays->lays[0].y = lineFeedHeight;
    for(cnt=0;cnt<maxCnt;cnt++)
    {
      //Get line length
      lineDesc = breakArray->breaks[cnt];
      //alloc mem for chars
      layInfo->chars = (NSCharArray *)NSChunkGrow((NSTextChunk *)layInfo->chars,sizeof(unsigned char)*(lineDesc+1));
      //Get chars
      charsCopied = 0;
      while(charsCopied<lineDesc && block!=NULL)
      {
        textDest = layInfo->chars->text + charsCopied;
        blockCharsLeft = block->chars - blockPos;
        if((lineDesc-charsCopied) < blockCharsLeft)
        {
          memmove(textDest,
                  block->text + blockPos,
                  lineDesc - charsCopied);
          blockPos += lineDesc - charsCopied;
          charsCopied = lineDesc;
        }
        else
        {
          if(blockCharsLeft > 0)
          {
            memmove(textDest,
                    block->text+blockPos,
                    blockCharsLeft);
            charsCopied += blockCharsLeft;
          }
          blockPos = 0;
          block = block->next;
        }
      }
      //Null-terminate chars
      layInfo->chars->text[lineDesc] = '\0';
      //Draw line
      NSScanALine(self,layInfo);
      NSDrawALine(self,layInfo);
      //Make line feed
      layInfo->lays->lays[0].y += lineFeedHeight;
    }
  NS_HANDLER
    freeNSLayInfo(layInfo);
    [localException raise];
  NS_ENDHANDLER
  /*lineFeedHeight = [defaultFont pointSize];
  currentTextPos.x = 0;
  currentTextPos.y = lineFeedHeight;

  [textColor set];
  [defaultFont set];
  bufferDrawRange.location = 0;
  bufferDrawRange.length = 0;
  textDrawLength = [textContents length];
  textStartPtr = (char *)[textContents cString];
  possibleBreakLoc = 0;
  makeLineFeed = NO;
  drawSizeLimit = [self frame].size;
  while(bufferDrawRange.location<textDrawLength)
  {
    [self fillTextBuffer];
    [self drawTextBuffer];
  }*/


  /*while(searchLoc<searchLength)
  {
    while(*strPtr!=' ' && *strPtr!='\t' && *strPtr!='\n' && searchLoc<searchLength)
    {
      searchLoc++;
      strPtr++;
      if(selectedRange.loaction==searchLoc)
    }
    tmpStr = [[textContents
                substringFromRange:
                  NSMakeRange(prevLocation,searchLoc-prevLocation)]
                autorelease];
    if(*strPtr=='\n' || searchLoc>=searchLength)
      newLine = YES;
    else
    {
      //Test if possible...
      if(pxkWidthOfString(tmpStr,defaultFont)>rect.size.width)
      {
        if(possibleBreak>0)
        {
          searchLoc = possibleBreak;
          strPtr = strStart + searchLoc;
          tmpStr = [[textContents
                      substringFromRange:
                        NSMakeRange(prevLocation,searchLoc-prevLocation)]
                      autorelease];
        }
        newLine = YES;
      }
      else
        possibleBreak = searchLoc;
    }
    searchLoc++;
    strPtr++;
    if(newLine)
    {
      pxkDrawString([tmpStr cString],rect.origin);
      prevLocation = searchLoc;
      rect.origin.y += fontHeight;
      possibleBreak = 0;
      newLine = NO;
    }
  }*/
}

- (void)drawAtPoint:(NSPoint)aPoint withAttributes:(NSDictionary *)attributes
{
  NSColor *color;
  NSFont *font;
  NSView *fView;
  NSPoint truePos = aPoint;
  BOOL flipped;
  int fontHeight,tmpBufferLoc,strBegin,strEnd;
  char *strPtr;
  NSString *tmpStr;

  fView = [NSView focusView];
  if(fView == nil)
  {
    [NSException raise:NSGenericException
      format:@"No view in focus in class NSString (NSStringAdditions) in method - drawAtPoint:withAttributes:"];
  }

  color = (NSColor *)[attributes objectForKey:
    NSForegroundColorAttributeName];
  if(color==nil)
    color = [NSColor blackColor];
  [color set];

  font = (NSFont *)[attributes objectForKey:
    NSFontAttributeName];
  if(font==nil)
    font = (NSFont *)[NSFont fontWithName:@"Helvetica" size:12];
  [font set];

  flipped = [fView isFlipped];
  fontHeight = [font pointSize];
  if(!flipped)
  {
    truePos.y = [fView frame].size.height - aPoint.y - fontHeight;
  }

  strPtr = (char *)[self cString];
  strBegin = 0;
  strEnd = [self length];
  tmpBufferLoc = 0;
  while(tmpBufferLoc<strEnd)
  {
    while(*strPtr!='\n' && tmpBufferLoc<strEnd)
    {
      tmpBufferLoc++;
      strPtr++;
    }
    printf("Before substring:%ld,%ld\n",(long)strBegin,(long)tmpBufferLoc);
    tmpStr = [self substringFromRange:NSMakeRange(strBegin,tmpBufferLoc-strBegin)];
    puts("After substring");
    pxkDrawString([tmpStr cString],truePos);
    if(flipped)
      truePos.y += fontHeight;
    else
      truePos.y -= fontHeight;
    tmpBufferLoc++;//Skip '/n'
    strPtr++;//Skip '/n'
    strBegin = tmpBufferLoc;
  }
}

- (void)__drawAtPoint:(NSPoint)point inRange:(NSRange)range
{
  NSFont *font;
  NSPoint truePos = point;
  NSRange effRange;
  NSString *text,*txtPiece;
  int length,strBegin,strEnd,tmpBufferLoc,textEdge,fontHeight;
  NSDictionary *attr;
  char *totalStrPtr,*strPtr;
  BOOL flipped;
  NSView *fView;
  NSColor *defColor;
  NSFont  *defFont;
  
  //This method shall NOT take line-sweep-direction into account, and it doesn't
  fView = [NSView focusView];
  if(!fView)
    [NSException raise:NSGenericException format:
      @"No focued view in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];

  textEdge = point.x;

  effRange = NSMakeRange(range.location,0);
  text = [self string];
  length = NSMaxRange(range);
  if(length>[text length])
  {
    [NSException raise:NSGenericException format:
      @"Range too long in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];
  }
  totalStrPtr = (char *)[text cString];
  flipped = [fView isFlipped];
  if(!flipped)
  {
    truePos.y = [fView frame].size.height - point.y - _txt_baseLineToBottom;
  }

  defColor = [NSColor blackColor];
  defFont = [NSFont fontWithName:@"Helvetica" size:12];
  while(NSMaxRange(effRange)<length)
  {
    attr = [self attributesAtIndex:NSMaxRange(effRange) effectiveRange:&effRange];
    if(NSMaxRange(effRange)>length)
      effRange.length = length - effRange.location;

    _txt_setAttributes(attr,&font,defFont,defColor);
    fontHeight = [font pointSize];
    
    strBegin = effRange.location;
    strPtr = totalStrPtr + strBegin;
    strEnd = NSMaxRange(effRange);
    tmpBufferLoc = strBegin;
    while(tmpBufferLoc<strEnd)
    {
      while(*strPtr!='\n' && tmpBufferLoc<strEnd)
      {
        tmpBufferLoc++;
        strPtr++;
      }
      txtPiece =
        [text substringFromRange:
        NSMakeRange(strBegin,tmpBufferLoc-strBegin)];
      pxkDrawString([txtPiece cString],truePos);
      
      if(tmpBufferLoc < strEnd)
      {
        //Forced line-feed
        if(flipped)
          truePos.y += fontHeight;
        else
          truePos.y -= fontHeight;
        truePos.x = textEdge;
        tmpBufferLoc++;//Skip '/n'
        strPtr++;//Skip '/n'
        strBegin = tmpBufferLoc;
      }
      else
        truePos.x += [font widthOfString:txtPiece];
    }
  }
}

#endif

void _txt_drawString(
  const char *cString,
  NSPoint pos,
  BOOL flipped,
  BOOL sweepLeftToRight,
  int fontHeight/* fontHeight is provided to speed things up */)
{
  if(!flipped)
  {
    pos.y -= 3;//Height from baseline to bottom of text,
               //use font and attributes here
  }
  else
  {
    pos.y += fontHeight;
  }
  pxkDrawString(cString,pos);
}

void _txt_setAttributes(
  NSDictionary *dict,
  NSFont **font,
  NSFont *_txt_defFont,
  NSColor *_txt_defColor)
{
  //FIXME: Add the other attributes to this function and maybe
  //       some more return-parameters like **font if nessecary

  NSColor *color;

  color = [dict objectForKey:NSForegroundColorAttributeName];
  if(color==nil)
    color = _txt_defColor;
  [color set];
    
  *font = [dict objectForKey:NSFontAttributeName];
  if(*font==nil)
    *font = _txt_defFont;
  [*font set];
}

- (void)__drawAtPoint:(NSPoint)point inRange:(NSRange)range
{
  NSFont *font;
  NSPoint truePos = point;
  NSRange effRange;
  NSString *text,*txtPiece;
  int strBegin,strEnd,tmpBufferLoc,attrBufferLoc;
  int textPieceEnd,textEdge,fontHeight,lineFeedHeight;
  NSDictionary *attr;
  char *totalStrPtr,*strPtr;
  BOOL flipped;
  BOOL doBreak = NO;
  NSView *fView;
  NSColor *defColor;
  NSFont  *defFont;
  
  //This method shall NOT take line-sweep-direction into account, and it doesn't
  fView = [NSView focusView];
  if(!fView)
    [NSException raise:NSGenericException format:
      @"No focued view in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];

  textEdge = point.x;

  strBegin = range.location;
  effRange = NSMakeRange(strBegin,0);
  text = [self string];
  strEnd = NSMaxRange(range);
  if(strEnd>[text length])
  {
    [NSException raise:NSGenericException format:
      @"Range too long in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];
  }
  totalStrPtr = (char *)[text cString];
  flipped = [fView isFlipped];
  if(!flipped)
  {
    truePos.y = [fView frame].size.height - point.y;
  }

  defColor = [NSColor blackColor];
  defFont = [NSFont fontWithName:@"Helvetica" size:12];

  tmpBufferLoc = strBegin;
  strPtr = totalStrPtr + strBegin;

  while(tmpBufferLoc<strEnd)
  {
    attrBufferLoc = tmpBufferLoc;
    if(NSMaxRange(effRange) <= attrBufferLoc)
    {
      attr = [self attributesAtIndex:NSMaxRange(effRange) effectiveRange:&effRange];
      _txt_setAttributes(attr,&font,defFont,defColor);
      fontHeight = [font pointSize];
    }
    lineFeedHeight = fontHeight;
    while(*strPtr!='\n' && tmpBufferLoc<strEnd)
    {
      tmpBufferLoc++;
      strPtr++;
    }
    while(effRange.location < tmpBufferLoc)
    {
      if(NSMaxRange(effRange) <= attrBufferLoc)
      {
        attr = [self attributesAtIndex:NSMaxRange(effRange) effectiveRange:&effRange];
        _txt_setAttributes(attr,&font,defFont,defColor);
        fontHeight = [font pointSize];
        if(lineFeedHeight<fontHeight)
          lineFeedHeight = fontHeight;
      }

      if(NSMaxRange(effRange)>=tmpBufferLoc)
      {
        textPieceEnd = tmpBufferLoc;
        doBreak = YES;
      }
      else
        textPieceEnd = NSMaxRange(effRange);

      txtPiece =
        [text substringFromRange:
        NSMakeRange(attrBufferLoc,textPieceEnd-attrBufferLoc)];
      _txt_drawString([txtPiece cString],truePos,flipped,YES,fontHeight);
      truePos.x += [font widthOfString:txtPiece];
      attrBufferLoc = textPieceEnd;
      if(doBreak)
      {
        doBreak = NO;
        break;
      }
    }
    if(tmpBufferLoc < strEnd)
    {
      //Forced line-feed
      if(flipped)
        truePos.y += lineFeedHeight;
      else
        truePos.y -= lineFeedHeight;
      truePos.x = textEdge;
      tmpBufferLoc++;//Skip '/n'
      strPtr++;//Skip '/n'
      strBegin = tmpBufferLoc;
    }
  }
}

//Drawing the string
-(void)drawAtPoint:(NSPoint)point
{
  //Note: This method shall NOT take line-sweep-direction into account - and it doesn't
  [self __drawAtPoint:point inRange:NSMakeRange(0,[self length])];
}

-(void)drawInRect:(NSRect)rect
{
  //FIXME: This method SHOULD take line-sweep-direction into account, but it doesn't do it in this implementation
  if(![NSView focusView])
    [NSException raise:NSGenericException format:
      @"No focued view in class NSAttributedString (NSAttributedStringAdditions) in method - drawInRect"];
  /*Draws the receiver with its font and other display attributes within rect in the currently focused NSView, clipping 
  the drawing to this rectangle. Text is drawn within rect according to its line sweep direction; for example, Arabic 
  text will begin at the right edge and potentially be clipped on the left.
  Don't invoke this method while no NSView is focused.
  See also: - lockFocus (NSView), - drawAtPoint:*/
}

- (NSSize)size
{
  /*Returns the bounding box of the marks that the receiver draws.
  See also: - drawAtPoint:, - drawInRect:*/
  return NSMakeSize(0,0);
}

//Generating RTF data
- (NSData *)RTFFromRange:(NSRange)aRange documentAttributes:(NSDictionary *)docAttributes
{
  /*Returns an NSData object that contains an RTF stream corresponding to the characters and attributes within aRange, 
  omitting all attachment attributes. Also writes the document-level attributes in docAttributes, as explained in the 
  class cluster description under RTF Document Attributes. If there are no document-level attributes, docAttributes 
  can be nil. Raises an NSRangeException if any part of aRange lies beyond the end of the
  receiver's characters.
  When writing data to the pasteboard, you can use the NSData object as the first argument to
  NSPasteboard's setData:forType: method, with a second argument of NSRTFPboardType.
  Although this method strips attachments, it leaves the attachment characters in the text itself.
  NSText's RTFFromRange: method, on the other hand, does strip attachment characters when extracting RTF.
  See also: - RTFDFromRange:documentAttributes:, - RTFDFileWrapperFromRange:documentAttributes:*/
  return nil;
}

- (NSFileWrapper *)RTFDFileWrapperFromRange:(NSRange)aRange documentAttributes:(NSDictionary *)docAttributes
{
  /*Returns an NSFileWrapper object that contains an RTFD document corresponding to the characters and attributes 
  within aRange. The file wrapper also includes the document-level attributes in docAttributes, as explained in the 
  class cluster description under RTF Document Attributes. If there are no document-level attributes, docAttributes 
  can be nil. Raises an NSRangeException if any part of aRange lies beyond the end of the
  receiver's characters.
  You can save the file wrapper using NSFileWrapper's writeToFile:atomically:updateFilenames: method.
  See also: - RTFDFromRange:documentAttributes:, - RTFFromRange:documentAttributes:*/
  return nil;
}

- (NSData *)RTFDFromRange:(NSRange)aRange documentAttributes:(NSDictionary *)docAttributes
{
  /*Returns an NSData object that contains an RTFD stream corresponding to the characters and attributes within aRange. 
  Also writes the document-level attributes in docAttributes, as explained in the class cluster description under 
  RTF Document Attributes. If there are no document-level attributes, docAttributes can be nil. Raises an 
  NSRangeException if any part of aRange lies beyond the end of the receiver's characters.
  When writing data to the pasteboard, you can use the NSData object as the first argument to
  NSPasteboard's setData:forType: method, with a second argument of NSRTFDPboardType.
  See also: - RTFFromRange:documentAttributes:, - RTFDFileWrapperFromRange:documentAttributes:*/
  return nil;
}

@end //NSAttributedString (NSAttributedStringAdditions)



@implementation NSMutableAttributedString (NSAttributedStringAdditions)

//Changing attributes
-(void)applyFontTraits:(NSFontTraitMask)mask range:(NSRange)aRange
{
  /*Apply the font attributes specified by mask to the characters in aRange. See the NSFontManager class specification 
  for a description of the font traits available. Raises an NSRangeException if any part of aRange lies beyond the 
  end of the receiver's characters.
  See also: - setAlignment:range:, - setAttributes:range:*/
}

-(void)setAlignment:(NSTextAlignment)alignment range:(NSRange)aRange
{
  /*Sets the alignment characteristic of the paragraph style attribute for the characters in aRange to alignment. When 
  attribute fixing takes place, this change will only affect paragraphs whose first character was included in aRange. 
  Raises an NSRangeException if any part of aRange lies beyond the end of the
  receiver's characters.
  See also: - addAttributes:range:, - applyFontTraits:range:, - fixParagraphStyleAttributeInRange:*/
}

-(void)subscriptRange:(NSRange)aRange
{
  NSRange effectiveRange = NSMakeRange(0,0);
  NSDictionary *attrs;
  int intNumber;
  id number;
  
  if(aRange.location<0 || NSMaxRange(aRange)>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSMutableAttributedString (NSAttributedStringAdditions) in method - subscriptRange:"];
  }

  while(NSMaxRange(effectiveRange) < [self length])
  {
    attrs = [self attributesAtIndex:NSMaxRange(effectiveRange)
      effectiveRange:&effectiveRange];
    number = [attrs objectForKey:NSSuperscriptAttributeName];
    if(number)
      intNumber = [number intValue] - 1;
    else
      intNumber = (-1);

    [self addAttribute:NSSuperscriptAttributeName
      value:[NSNumber numberWithInt:intNumber]
      range:effectiveRange];
  }
}

-(void)superscriptRange:(NSRange)aRange
{
  NSRange effectiveRange = NSMakeRange(0,0);
  NSDictionary *attrs;
  int intNumber;
  id number;
  
  if(aRange.location<0 || NSMaxRange(aRange)>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSMutableAttributedString (NSAttributedStringAdditions) in method - superscriptRange:"];
  }
  
  while(NSMaxRange(effectiveRange) < [self length])
  {
    attrs = [self attributesAtIndex:NSMaxRange(effectiveRange)
      effectiveRange:&effectiveRange];
    number = [attrs objectForKey:NSSuperscriptAttributeName];
    if(number)
      intNumber = [number intValue] + 1;
    else
      intNumber = 1;

    [self addAttribute:NSSuperscriptAttributeName
      value:[NSNumber numberWithInt:intNumber]
      range:effectiveRange];
  }
}

-(void)unscriptRange:(NSRange)aRange
{
  [self removeAttribute:NSSuperscriptAttributeName
    range:aRange];
}

//Updating attachment contents
-(void)updateAttachmentsFromPath:(NSString *)path
{
  /*Updates all attachments based on files contained in the RTFD file package at path.
  See also: - updateFromPath: (NSFileWrapper)*/
}

//Fixing attributes after changes
-(void)fixAttributesInRange:(NSRange)aRange
{
  [self fixAttachmentAttributeInRange:aRange];
  [self fixFontAttributeInRange:aRange];
  [self fixParagraphStyleAttributeInRange:aRange];
}

-(void)fixAttachmentAttributeInRange:(NSRange)aRange
{
  NSRange effectiveRange = NSMakeRange(0,0);
  NSRange tmpRange;
  NSDictionary *attrs;
  NSString *tmpStr;
  int cnt;
  unichar attChar = NSAttachmentCharacter;
  
  if(aRange.location<0 || NSMaxRange(aRange)>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSMutableAttributedString (NSAttributedStringAdditions) in method - fixAttachmentAttributeInRange:"];
  }

  tmpStr = [[NSString alloc]
    initWithCharacters:&attChar length:1];
  [tmpStr autorelease];

  while(NSMaxRange(effectiveRange) < [self length])
  {
    attrs = [self attributesAtIndex:NSMaxRange(effectiveRange)
      effectiveRange:&effectiveRange];
    if([attrs objectForKey:NSAttachmentAttributeName])
    {
      for(cnt=effectiveRange.location;cnt<NSMaxRange(effectiveRange);cnt++)
      {
        tmpRange = NSMakeRange(cnt,1);
        if([[[self string] substringFromRange:tmpRange] isEqual:tmpStr] == NO)
        {
          [self removeAttribute:NSAttachmentAttributeName
            range:tmpRange];
        }
      }
    }
  }
}

-(void)fixFontAttributeInRange:(NSRange)aRange
{
  if(aRange.location<0 || NSMaxRange(aRange)>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSMutableAttributedString (NSAttributedStringAdditions) in method - fixFontAttributeInRange:"];
  }
  /*Fixes the font attribute in aRange, assigning default fonts to characters with illegal fonts for their scripts and 
  otherwise correcting font attribute assignments. For example, Kanji characters in assigned a Latin font are 
  reassigned an appropriate Kanji font. Raises an NSRangeException if any part of aRange lies beyond the end of the 
  receiver's characters.
  See also: - fixParagraphStyleAttributeInRange:, - fixAttachmentAttributeInRange:,
  - fixAttributesInRange:*/
}

-(void)fixParagraphStyleAttributeInRange:(NSRange)aRange
{
  if(aRange.location<0 || NSMaxRange(aRange)>[self length])
  {
    [NSException raise:NSRangeException format:
      @"Out of range in class NSMutableAttributedString (NSAttributedStringAdditions) in method - fixParagraphStyleAttributeInRange:"];
  }
  /*Fixes the paragraph style attributes in aRange, assigning the first paragraph style attribute value in each 
  paragraph to all characters of the paragraph. This method extends the range as needed to cover the last paragraph 
  partially contained. A paragraph is delimited by any of these characters, the longest possible sequence being 
  preferred to any shorter:
  U+000D (\r or CR) U+2028 (Unicode line separator) U+000A (\n or LF) U+2029 (Unicode paragraph separator) \r\n, in 
  that order (also known as CRLF)
  Raises an NSRangeException if any part of aRange lies beyond the end of the
  receiver's characters.
  See also: - fixFontAttributeInRange:, - fixAttachmentAttributeInRange:, - fixAttributesInRange:*/
}

@end //NSMutableAttributedString