/* 
   NSBitmapImageRep.m

   Bitmap image representation.

   Copyright (C) 1996 Free Software Foundation, Inc.
   
   Written by:  Adam Fedor <fedor@colorado.edu>
   Date: Feb 1996
   Modified by: ANOQ of the Sun <anoq@vip.cybercity.dk>
   Date: November 1997
   
   This file is part of the NSXKit.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <math.h>
#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSData.h>
#include "NSBitmapImageRep.h"
#include "nsimage-tiff.h"
#include "PXKDrawing.h"

//FIXME:
//------
//In the NSBitmapImageRep class I allocate 32 bits per pixel for the
//internal storage format when the display is running 32bits per
//pixel. I do it to avoid an image-conversion from 24bit to 32bit
//when sending the image to the X-server. I don't know if this
//violates the OpenStep-specification. I haven't found any other
//ways to avoid this image-conversion, and the image-conversion
//takes a lot of time. The image-conversion has been written
//though and is in PXKDrawing.m in the function NSDrawBitmap().

/* Maximum number of planes */
#define MAX_PLANES 5

@implementation NSBitmapImageRep 

/* Given a TIFF image (from the libtiff library), load the image information
   into our data structure.  Reads the specified image. */
- _initFromImage: (TIFF *)image number: (int)imageNumber
{
  NSString* space;
  NSTiffInfo* info;
  NSTiffDesire desire;

  imageData = NULL;
  imagePlanes = NULL;
  info = NSTiffGetInfo(imageNumber, image);
  if (!info) 
    {
      [NSException raise:NSTIFFException format: @"Read invalid TIFF info"];
    }

  /* 8-bit RGB will be converted to 24-bit by the tiff routines, so account
     for this. */
  space = nil;
#ifdef HAVE_LIBTIFF
  switch(info->photoInterp) 
    {
    case PHOTOMETRIC_MINISBLACK: space = NSDeviceWhiteColorSpace; break;
    case PHOTOMETRIC_MINISWHITE: space = NSDeviceBlackColorSpace; break;
    case PHOTOMETRIC_RGB: space = NSDeviceRGBColorSpace; break;
    case PHOTOMETRIC_PALETTE:
      space = NSDeviceRGBColorSpace; 
      info->samplesPerPixel = 3;
      break;
    default:
      break;
    }
#endif

  printf("Image dimensions: %ldx%ld bitsPerSample:%ld samplesPerPixel:%ld\n",
    (long)info->width,(long)info->height,(long)info->bitsPerSample,
    (long)info->samplesPerPixel);
  if(info->planarConfig == PLANARCONFIG_SEPARATE)
    puts("Planar data!");
  desire.bitsPerSample = pxkFVRBitsPerSample();
  desire.samplesPerPixel = pxkFVRSamplesPerPixel();
  desire.wantsAlpha = ((info->samplesPerPixel > 3) || pxkFVRHasAlpha()); //Never discard any information...
  desire.planar = pxkFVRIsPlanar();
  [self initWithBitmapDataPlanes: NULL
		pixelsWide: info->width
		pixelsHigh: info->height
		bitsPerSample: desire.bitsPerSample
		samplesPerPixel: desire.samplesPerPixel
		hasAlpha: desire.wantsAlpha
		isPlanar: desire.planar
		colorSpaceName: space
		bytesPerRow: 0
		bitsPerPixel: 0];
  compression = info->compression;

  if (NSTiffRead(imageNumber, image, NULL, &desire, [self bitmapData]))
  {
    [NSException raise:NSTIFFException format: @"Read invalid TIFF image"];
  }
  
  return self;
}

+ (id) imageRepWithData: (NSData *)tiffData
{
  NSBitmapImageRep* imageRep = nil;

  imageRep = [[self class] alloc];
  if(!imageRep)
  {
    [NSException raise: NSAppKitVirtualMemoryException
      format: @"Out of virtual memory for new NSBitmapImageRep"];
  }
  [imageRep autorelease];
  [imageRep initWithData:tiffData];
  return imageRep;
}

+ (NSArray *) imageRepsWithData: (NSData *)tiffData
{
  int images;
  TIFF*       image;
  NSTiffInfo* info;
  NSMutableArray* array;

  puts("NSBitmapImageRep: + imageRepsWithData:(NSData *)tiffData");
  image = NSTiffOpenData((char *)[tiffData bytes], [tiffData length], 
			 "r", NULL);
  if (!image)
    [NSException raise:NSTIFFException format: @"Read invalid TIFF data"];
               
  
  array = [[NSMutableArray arrayWithCapacity:1] autorelease];
  images = 0;
  while ((info = NSTiffGetInfo(images, image))) 
    {
      NSBitmapImageRep* imageRep;

      OBJC_FREE(info);
      imageRep = [[[[self class] alloc]
		   _initFromImage: image number: images] autorelease];
      [array addObject: imageRep];
      images++;
    }
  NSTiffClose(image);

  return array;
}

/* Loads only the default (first) image from the TIFF image contained in
   data. */
- (id) initWithData: (NSData *)tiffData
{
  TIFF 	*image;

  imageData = NULL;
  imagePlanes = NULL;
  image = NSTiffOpenData((char *)[tiffData bytes], [tiffData length], 
			 "r", NULL);
  if (!image)
    {
      [NSException raise:NSTIFFException format: @"Read invalid TIFF data"];
    }

  [self _initFromImage:image number: 0];
  NSTiffClose(image);
  return self;
}

- (id) initWithFocusedViewRect: (NSRect)rect
{
  NSString *space;

  space = NSDeviceRGBColorSpace;
  
  [self initWithBitmapDataPlanes: NULL
		pixelsWide: rect.size.width
		pixelsHigh: rect.size.height
		bitsPerSample: pxkFVRBitsPerSample()
		samplesPerPixel: pxkFVRSamplesPerPixel()
		hasAlpha: pxkFVRHasAlpha()
		isPlanar: pxkFVRIsPlanar()
		colorSpaceName: space
		bytesPerRow: 0
		bitsPerPixel: 0];
  imageData = pxkFVRGrab(rect);
  //Init imagePlanes...
  return self;
}

/* This is the designated initializer */
/* Note: If data is actaully passed to us in planes, we DO NOT own this
   data and we DO NOT copy it. Just assume that it will always be available.
*/
- (id) initWithBitmapDataPlanes: (unsigned char **)planes
		pixelsWide: (int)width
		pixelsHigh: (int)height
		bitsPerSample: (int)bps
		samplesPerPixel: (int)spp
		hasAlpha: (BOOL)alpha
		isPlanar: (BOOL)isPlanar
		colorSpaceName: (NSString *)colorSpaceName
		bytesPerRow: (int)rowBytes
		bitsPerPixel: (int)pixelBits;
{
  if (!bps || !spp || !width || !height) 
    {
      [NSException raise: NSInvalidArgumentException
        format: @"Required arguments not specified creating NSBitmapImageRep"];
    }

  imageData = NULL;
  imagePlanes = NULL;
  _pixelsWide = width;
  _pixelsHigh = height;
  size.width  = width;
  size.height = height;
  bitsPerSample = bps;
  samplesPerPixel = spp;
  hasAlpha   = alpha;  
  _isPlanar   = isPlanar;
  _colorSpace = [colorSpaceName retain];
  if (!pixelBits)
    pixelBits = bps * ((_isPlanar) ? 1 : spp);
  bitsPerPixel            = pixelBits;
  if (!rowBytes) 
    rowBytes = ceil((float)width * bitsPerPixel / 8);
  bytesPerRow            = rowBytes;
  printf("bytesPerRow:%ld\n",(long)bytesPerRow);

  if (planes) 
    {
      int i;
      OBJC_MALLOC(imagePlanes, unsigned char*, MAX_PLANES);
      if(!imagePlanes)
      {
        [NSException raise: NSAppKitVirtualMemoryException
          format: @"Out of virtual memory for initialising NSBitmapImageRep"];
      }
      for (i = 0; i < MAX_PLANES; i++)
 	imagePlanes[i] = NULL;
      for (i = 0; i < ((_isPlanar) ? samplesPerPixel : 1); i++)
 	imagePlanes[i] = planes[i];
    }
  
  return self;
}

- (void) dealloc
{
  if(_colorSpace)
    [_colorSpace autorelease];
  if(imagePlanes)
    OBJC_FREE(imagePlanes);
  if(imageData)
    [imageData release];
  [super dealloc];
}

+ (BOOL) canInitWithData: (NSData *)data
{
  TIFF *image = NULL;
  image = NSTiffOpenData((char *)[data bytes], [data length], "r", NULL);
  if(image)
    NSTiffClose(image);

  return (image) ? YES : NO;
}

+ (BOOL) canInitWithPasteboard: (NSPasteboard *)pasteboard
{
  [self notImplemented: _cmd];
  return NO;
}

+ (NSArray *) imageFileTypes
{
  return [self imageUnfilteredFileTypes];
}

+ (NSArray *) imagePasteboardTypes
{
  return [self imageUnfilteredPasteboardTypes];
}

+ (NSArray *) imageUnfilteredFileTypes
{
  return [[NSArray arrayWithObjects: @"tiff", @"tif", nil] autorelease];
}

+ (NSArray *) imageUnfilteredPasteboardTypes
{
  return [[NSArray arrayWithObjects: NSTIFFPboardType, nil] autorelease];
}

// Getting Information about the Image 
- (int) bitsPerPixel
{
  return bitsPerPixel;
}

- (int) samplesPerPixel
{
  return samplesPerPixel;
}

- (BOOL) isPlanar
{
  return _isPlanar;
}

- (int) numberOfPlanes
{
  return (_isPlanar) ? samplesPerPixel : 1;
}

- (int) bytesPerPlane
{
  return bytesPerRow*_pixelsHigh;
}

- (int) bytesPerRow
{
  return bytesPerRow;
}

// Getting Image Data 
- (unsigned char *) bitmapData
{
  unsigned char *planes[MAX_PLANES];
  [self getBitmapDataPlanes: planes];
  return (unsigned char *)[imageData mutableBytes];
}

- (void) getBitmapDataPlanes: (unsigned char **)data
{
  int i;
  BOOL initPlanes = NO;
  
  if (!imageData)
  {
    long length;
      
    length = (long)bytesPerRow * _pixelsHigh * sizeof(unsigned char);
    imageData = [[NSMutableData dataWithLength: length] retain];
      // FIXME: will this raise an exception if out of memory?
    initPlanes = YES;
  }
  if(!imagePlanes)
  {
    OBJC_MALLOC(imagePlanes, unsigned char*, MAX_PLANES);
    if(!imagePlanes)
    {
      [NSException raise: NSAppKitVirtualMemoryException
        format: @"Out of virtual memory for initialising NSBitmapImageRep"];
    }
    initPlanes = YES;
  }
  if(initPlanes)
  {
    unsigned char* bits;
    
    bits = [imageData mutableBytes];
    if (_isPlanar)
    {
      for (i=0; i < samplesPerPixel; i++)
        imagePlanes[i] = bits + (i*bytesPerRow * _pixelsHigh);
      for (i= samplesPerPixel; i < MAX_PLANES; i++)
        imagePlanes[i] = NULL;
    }
    else
    {
      imagePlanes[0] = bits;
      for (i= 1; i < MAX_PLANES; i++) 
        imagePlanes[i] = NULL;
    }
  }
  if (data)
    for (i=0; i < samplesPerPixel; i++)
      data[i] = imagePlanes[i];
}

- (BOOL) draw
{
  unsigned char *planes[MAX_PLANES];
  NSRect rect;
  
  [self getBitmapDataPlanes: planes];
  printf("imageData length:%ld\n",(long)[imageData length]);
  rect.origin = pxkGetCurrentPos();
  rect.size = size;
  NSDrawBitmap(rect,_pixelsWide,_pixelsHigh,bitsPerSample,
    samplesPerPixel,bitsPerPixel,bytesPerRow,_isPlanar,
    hasAlpha,_colorSpace,(const unsigned char **)planes);
  return YES;// FIXME: NSDrawBitmap does not return a boolean...
  //return pxkBitmapImageRepDraw(dataPtr,
  //                             bytesPerRow,
  //                             samplesPerPixel,
  //                             bitsPerSample,
  //                             bitsPerPixel,
  //                             _isPlanar,
  //                             size);
}

// Producing a TIFF Representation of the Image 
+ (NSData *) TIFFRepresentationOfImageRepsInArray: (NSArray *)anArray
{
  [self notImplemented: _cmd];
  return nil;
}

+ (NSData *) TIFFRepresentationOfImageRepsInArray: (NSArray *)anArray
		usingCompression: (NSTIFFCompression)compressionType
		factor: (float)factor
{
  [self notImplemented: _cmd];
  return nil;
}

- (NSData *) TIFFRepresentation
{
  [self notImplemented: _cmd];
  return nil;
}

- (NSData *) TIFFRepresentationUsingCompression: (NSTIFFCompression)compressionType
		factor: (float)factor
{
  [self notImplemented: _cmd];
  return nil;
}


// Setting and Checking Compression Types 
+ (void) getTIFFCompressionTypes: (const NSTIFFCompression **)list
		count: (int *)numTypes
{
  [self notImplemented: _cmd];
}

+ (NSString *) localizedNameForTIFFCompressionType: (NSTIFFCompression)compression
{
  [self notImplemented: _cmd];
  return nil;
}

- (BOOL) canBeCompressedUsing: (NSTIFFCompression)compression
{
  [self notImplemented: _cmd];
  return NO;
}

- (void) getCompression: (NSTIFFCompression *)compression
		factor: (float *)factor
{
  [self notImplemented: _cmd];
}

- (void) setCompression: (NSTIFFCompression)compression
		factor: (float)factor
{
  [self notImplemented: _cmd];
}


// NSCoding protocol
- (void) encodeWithCoder: aCoder
{
  [self notImplemented: _cmd];
}

- initWithCoder: aDecoder
{
  [self notImplemented: _cmd];
  return nil;
}

@end
