/*
 * Filename: NSButton.m
 * Project:  Pascal's X kit
 * Module:   Basic application components
 *
 * Function: This is the Button class.  See header file for a description.
 *
 * Author:   Pascal Forget
 * Date:     February 1997
 *
 * Copyright (C) 1997 Pascal Forget
 */

#include "NSButton.h"
#include "NSButtonCell.h"
#include "PXKDrawing.h"
#include "PXKGeometry.h"
#include "misc.h"

Class cellClass;

@interface NSButton (Private)

- (void)drawBorder;
- (void)drawTitle;
- (float)widthOfCell;
@end

@implementation NSButton (Private)

- (void)drawBorder;
{
#if 0    
    NSRect rect = [self frame]; // frame is an NSView's instance variable
    NSRect clipRect; 
    
    rect.origin.x = 0.0;
    rect.origin.y = 0.0;
    
    clipRect = rect;
    
    rect.origin.x = 1.0;    
    rect.size.width -= 2.0;
    rect.size.height -= 1.0;
    
    if ((cell != nil) && (titlePosition == NSAtTop))
    {
	rect.origin.y = 5.0;
	rect.size.height -= 5.0;
    }
    
    NSDrawGroove(rect, clipRect);
#endif    
}

- (void)drawTitle;
{
#if 0    
    NSRect rect = {{0.0, 0.0}, {80.0, 16.0}};
    
    rect.size.width = [self widthOfCell] + 2.0;

    // Center the title horizontally
    rect.origin.x = (frame.size.width - rect.size.width) / 2.0;
    
    if ((titleCell != nil) && (titlePosition == NSAtTop))
    {
	[titleCell drawWithFrame: rect inView: self];
    }
#endif    
}

- (float)widthOfCell;
{
    return [[cell font] widthOfString: [cell stringValue]];
}

@end

@implementation NSButton

+ (Class)cellClass;
{
    return cellClass;
}

+ (void)setCellClass:(Class)classId;
{
    cellClass = classId;
}

- (void)display;
{
    NSRect rect = [self frame];

    rect.origin.x = 0.0;
    rect.origin.y = 0.0;

    [super display];
    
    [[NSColor lightGrayColor] set];

    [(NSView *)self lockFocus];
    if (isOpaque) {
      NSRectFill(rect);
    }
    [cell drawWithFrame: rect inView: self];

    [(NSView *)self unlockFocus];
}

- (id)initWithFrame:(NSRect)frameRect;
{
    [super initWithFrame:frameRect];
    isEnabled = YES;
    target = nil;
    state = 0;

    backgroundColor = [NSColor lightGrayColor];
    
    cell = [[NSButtonCell alloc] initTextCell: @"cell"];
    return self;
}

- (void)mouseDown:(NSEvent *)theEvent;
{
  BOOL continueTracking = YES;
  NSEvent *event;
  NSEventType event_type;
  NSDate *date = [NSDate dateWithTimeIntervalSinceNow:40];
  NSPoint aPoint;

  // HACK no. 2 -- NSControl resets the state to 0 !!
  BOOL oldState=[self state];
  BOOL lastState=[self state];

  // Not enabled, ignore the event
  if (!isEnabled) {
	return;
  }


  // NSControl is not finished enough..
  // Should much of this functionality be put into NSControl, or
  // at least into seperate methods?
  //
  //  [super mouseDown: theEvent];
  
  // Track the mouse continuously.  If it exits the button
  // without the mouse button being released, just set the
  // state of the button back to normal and don't send
  // a message to the target.
  
  // Only assume button was ON if this type of button
  // HACK no 1 modified NSButtonCell to have buttonType method,
  if (([cell buttonType] == NSMomentaryPushButton) ||
      ([cell buttonType] == NSMomentaryChangeButton)) {
    [self setState: 1];
  } else {
    [self setState: !oldState];
    [self highlight: YES];
  }
  
  [self display];
  
  while ((event = [NSApp nextEventMatchingMask: NSLeftMouseDraggedMask |
			 NSLeftMouseUpMask | NSMouseMovedMask |
			 NSMouseEnteredMask | NSMouseExitedMask
			 untilDate: date
			 inMode: @"NSEventTrackingRunLoopMode"
			 dequeue: YES]) != nil) {
    
    event_type = [event type];

    if (event_type != NSLeftMouseUp) {
      aPoint = [superview convertPoint: [event locationInWindow]
			  fromView: nil];      
      // Work out state
      if ([self hitTest: aPoint] != nil) {
	[self setState: !oldState];
	[self highlight: YES];
      } else {
	[self setState: oldState];
	[self highlight: NO];
      }
      // Only redraw if there's been a change 
      if ([self state] != lastState) {
     	lastState=[self state];
	[self display];
      }      
    } else if (event_type == NSLeftMouseUp) {
      [self mouseUp: event];
      return;
    }
  }
  
  // Timeout - change to old state ?
  [self setState: oldState];
  [self highlight: NO];
  [self display];
}

- (void)mouseUp:(NSEvent *)theEvent;
{
  NSPoint aPoint;

  // Not enabled, ignore the event
  if (!isEnabled) {
	return;
  }

  // NSCell too knackered to do anything
  // [super mouseUp: theEvent];

  // Behaviour dependant on type of button
    switch ([(NSButtonCell *)cell buttonType]) {
    case NSPushOnPushOffButton :
      [self highlight: NO];
      [self display];
      break;
    case NSMomentaryPushButton :
    case NSMomentaryChangeButton :
      // Reset state
      [self setState: 0];
      [self display];
      break;
    default:
    }

    // Check if we've activated the button
    aPoint = [superview convertPoint: [theEvent locationInWindow]
			fromView: nil];      
    if ([self hitTest: aPoint] != nil) {
      // send the Action
      [(NSControl *)self sendAction];
    }
}

- (void)setState:(int)value;
{
  if (cell == nil) { /* impossible condition: fatal error */
    /* log an error, raise an exception */
  }
  
  [cell setState:value];
}

- (void)setType:(int)aType;
{
  if (cell == nil) { /* impossible condition: fatal error */
    /* log an error, raise an exception */
  }
  
  [cell setType:aType];
}

- (int)state;
{
  return [cell state];
}

- (NSImage *)alternateImage;
{
  return [(NSButtonCell *)cell alternateImage];
}

- (NSString *)alternateTitle;
{
  return [(NSButtonCell *)cell alternateTitle];
}

- (void)getPeriodicDelay:(float *)delay interval:(float *)interval;
{
  [(NSButtonCell *)cell getPeriodicDelay:delay interval:interval];
}

- (void)highlight:(BOOL)flag;
{
  // This method not implemented
  [(NSButtonCell *)cell highlight:flag withFrame:[self frame] inView:self];
}

- (NSImage *)image;
{
  return [(NSCell *)cell image];
}

- (NSCellImagePosition)imagePosition;
{
  return [(NSButtonCell *)cell imagePosition];
}

- (BOOL)isBordered;
{
  return [(NSCell *)cell isBordered];
}

- (BOOL)isTransparent;
{
  return [(NSButtonCell *)cell isTransparent]; 
}

- (void)performClick:(id)sender;
{
  [(NSButtonCell *)cell performClick:sender];    
}

- (void)setAlternateImage:(NSImage *)anImage;
{
  [(NSButtonCell *)cell setAlternateImage:anImage];
}

- (void)setAlternateTitle:(NSString *)aString;
{
  [(NSButtonCell *)cell setAlternateTitle:aString];
}

- (void)setBordered:(BOOL)flag;
{
  [(NSCell *)cell setBordered:flag];
}

- (void)setImage:(NSImage *)anImage;
{
  [(NSCell *)cell setImage:anImage];
}

- (void)setImagePosition:(NSCellImagePosition)aPosition;
{
  [(NSButtonCell *)cell setImagePosition:aPosition];
}

- (void)setPeriodicDelay:(float)delay interval:(float)interval;
{
  [(NSButtonCell *)cell setPeriodicDelay:delay interval:interval];
}

- (void)setTitle:(NSString *)aString;
{
  [(NSButtonCell *)cell setTitle:aString];
}

- (void)setTransparent:(BOOL)flag;
{
  [(NSButtonCell *)cell setTransparent:flag];
}

- (NSString *)title;
{
  return [(NSButtonCell *)cell title];
}

@end
