/*
 * Filename: NSButtonCell.m
 * Project:  Pascal's X Kit
 * Module:   Basic components
 *
 * Function: See header file
 *
 * Author:   Pascal Forget
 * Date:     February 1997
 *
 * Copyright (C) 1997 Pascal Forget
 */

#include <Foundation/NSGeometry.h>
#include "NSButtonCell.h"
#include "NSColor.h"
#include "PXKDrawing.h"
#include "PXKGeometry.h"
#include "misc.h"

@implementation NSButtonCell

- (NSImage *)alternateImage;
{
  return altImage;
}

- (NSString *)alternateTitle;
{
  return altTitle;
}

+ (BOOL)prefersTrackingUntilMouseUp;  /* Subclasses may override */
{
  return NO; /* Tracking stops when the mouse leaves the NSCell */
}

- (SEL)action; /* Implemented by subclasses */
{
  return NULL;
}

- (NSTextAlignment)alignment;
{
  return cellAlignment;
}

- (void)calcDrawInfo:(NSRect)aRect;
{
  /* Implemented by subclasses as per OpenStep specification */
}

- (NSSize)cellSize;
{
  return cellSize;
}

- (NSSize)cellSizeForBounds:(NSRect)aRect;
{
  NSSize invalidSize;
  
  if ((aRect.size.width >= cellSize.width) &&
      (aRect.size.height >= cellSize.height)) {
    return aRect.size;
  } else {
    invalidSize.width = 0.0;
    invalidSize.height = 0.0;
    return invalidSize;
  }
}

- (NSComparisonResult)compare:(id)otherCell;
{
  if ((!otherCell) || ([otherCell isKindOfClass:[NSCell class]] == NO)) {
#if 0
    NSRaiseException(NSBadComparisonException); /* Right syntax? */
#else
    exit(-1); /* Crash the program */
#endif
  } else {
    return [[self stringValue] compare:[otherCell stringValue]];
  }
}

- (NSView *)controlView;
{
  return nil; /* Implemented by subclasses */
}

- (void)dealloc;
{
  if (altTitle != nil) {
    [altTitle release];
  }
  if (title != nil) {
    [title release];
  }

  if (altImage != nil) {
    [altImage release];
  }

  /* image handled by NSCell */
  
  [super dealloc];
}

/*
 * We need to overide this as, the Bevel can change dependant on the Button
 * State and *Type*.
 *
 * Should this not be part of NSCell, then
 * When we do a setType: we set how the Cell is highlighted, etc..
 *
 */

- (void)drawWithFrame:(NSRect)rect inView:(NSView *)aView;
{
  NSRect bounds = rect;
  NSRect aLine = rect;

  //  [super drawWithFrame:rect inView:aView];
  [aView lockFocus];
  
  bounds.origin.x+=2;	      // These should be defined as constant somewhere
  bounds.origin.y+=2;
  bounds.size.width -= 4.0;
  bounds.size.height -= 4.0;
  
  aLine.origin.x++;
  aLine.origin.y++;
  aLine.size.width-=2;
  aLine.size.height-=2;
    
  switch (buttonType) {
  case NSMomentaryPushButton:
  case NSMomentaryChangeButton:
  case NSToggleButton:
    if (![self state]) {
      dbug(msg("Cell state: 0\n"));
      NSDrawButton(aLine, rect);
    } else {
      dbug(msg("Cell state: 1\n"));
			NSDrawGrayBezel(aLine, rect);
    }
    break;
    
  case NSPushOnPushOffButton:
    if (isHighlighted) {
      NSDrawGrayBezel(aLine, rect);
    } else {
      NSDrawButton(aLine, rect);
    }
    break;
    
  case NSSwitchButton:
  case NSRadioButton:
    break;

  case NSOnOffButton:
    NSDrawButton(aLine, rect);	  
    break;
    
  }
  
  [aView unlockFocus];
  
  [self drawInteriorWithFrame: bounds inView: aView];
}

/*
 * NSButtonCell overrides NSCell's implementation of drawTextInRect:
 * because the text depends on the state of the cell.  If the state
 * equals 0, we draw the cell's title, otherwise we draw the alternate
 * title.
 */

- (void)drawTextInRect:(NSRect)frame; {
  int width;
  int font_height;
  int alignment=[self alignment];
  NSPoint location;
  NSString *aString;
  NSImage *theImage;
  NSRect imageRect, textRect=frame;

  if (cellType != NSImageCellType) {
    /*
     * Wipe out the contents of the window by filling its
     * area with the backGroundColor (defined in NSView.h)
     * or the White if the Button is highlighted
     * - should really do a NSHighlightRect() after drawing it all
     */    

    // NSButton does a RectFill ... 
    if (isOpaque) {
      [_controlView lockFocus];

      [[_controlView backgroundColor] set];
      
      switch (buttonType) {
      case NSRadioButton:
      case NSSwitchButton:
      case NSToggleButton:
      case NSMomentaryChangeButton:
	break;
      case NSPushOnPushOffButton:
      case NSMomentaryPushButton:
      case NSOnOffButton:
	if ([self state]) {
	  NSEraseRect(frame);
	}
	break;
      default:
      }
      
      [_controlView unlockFocus];
    }
  }
  
  if ((cellType == NSTextCellType) || 
      ((imagePosition != NSImageOnly) && (cellType == NSImageCellType))) {
  
    // Work out what string to display
    switch (buttonType) {
    case NSMomentaryPushButton :
    case NSOnOffButton :
    case NSMomentaryChangeButton :
    case NSRadioButton :
    case NSSwitchButton :
      aString = [self title];
      break;
    default:
      if (![self state]) {
	aString = [self title];
      } else {
	aString = [self alternateTitle];
      }
    }
    
    /* get string width for centering */
    width = [cellFont widthOfString: aString];
    
    font_height = [cellFont pointSize];
    
    /* Work Out where text should be */

    // Take account of Images

    theImage=[self image];
    if (theImage) {
      // NSImage knackered!
      // imageRect.size=[theImage size];
      imageRect.size.width=10.0;
      imageRect.size.height=10.0;

      // Relocate the textRect to make room for Image
      switch (imagePosition) {
      case NSImageLeft:
      case NSImageOnly:
	textRect.origin.x += imageRect.size.width + 2.0;
	textRect.size.width -= imageRect.size.width + 2.0;
	break;
      case NSImageRight:
	textRect.size.width -= imageRect.size.width - 2.0;
	break;
      case NSImageBelow:
	// Do we put the text Directly above the image?
	break;
      case NSImageAbove:
	textRect.origin.y += imageRect.size.height + 2.0;
	textRect.size.height -= imageRect.size.height + 2.0;
	break;
      case NSImageOverlaps:
	// ?
	break;
      }
    }


    /* centre text vertically */
    location.y = textRect.origin.y +
      (textRect.size.height + font_height) / 2.0 - 2.0;

    if((alignment == NSLeftTextAlignment) && (width < textRect.size.width)) {
      location.x = textRect.origin.x + 1.0;
      //	location.y = textRect.origin.y + font_height;
    } else if((alignment == NSRightTextAlignment) 
	      && ((width+5)<textRect.size.width)) {
      location.x = textRect.origin.x + textRect.size.width - 2 - width;
      //	location.y = textRect.origin.y + font_height;
    } else {
      if (width > (textRect.size.width-4)) {
	location.x = textRect.origin.x + textRect.size.width - width-2;
	//	location.y = textRect.origin.y + font_height;
      } else {
	/* output text, centered on each line */
	location.x = textRect.origin.x + (textRect.size.width - width-2)/2;
	//	location.y = textRect.origin.y + font_height;
      }
    }

    
    // Draw the appropriate Text
    // Tests if CELL is Enables, not Control...
    // Why the difference?
    if (!isEnabled) {
      [[NSColor darkGrayColor] set];
      msg("setting colour of disabled cell\n");

    } else {
      [[NSColor blackColor] set];
    }
    [_controlView lockFocus];
    [cellFont set];
    pxkDrawString([aString cString], location);
    [_controlView unlockFocus];
  }
}

/*
 * NSButtonCell overrides NSCell's implementation of drawImage:
 * because the Image depends on the state of the cell and the type of
 * button.  If the state
 * equals 0, we draw the cell's image, otherwise we draw the alternate
 * image.
 * (Can this be generalied to NSCell??)
 *
 * Stuart 10/4/97
 */
- (void)drawImage; {
  float width, height;
  NSPoint location;
  NSImage *theImage;
  NSRect aRect;
  NSRect frame = [_controlView frame];

  /* Work out what image to display */
  switch (buttonType) {
  case NSSwitchButton :
  case NSRadioButton :
    if ([self state]) {
      theImage = [self alternateImage];
      break;
    }
    
  default:
    theImage = [self image];
  }
  
  // Can't use any NSImage methods as class is knackered!
  // size = [theImage size];
  // width = size.width;
  // height = size.height;

  // Dummy thing
  width = 10.0 ;
  height = 10.0 ;

  /* Work out where Image should be */
  //  if (imagePosition != NSNoImage) {
    switch (imagePosition) {
    case NSImageLeft:
      location.x = 2.0;
      location.y = (frame.size.height - height) / 2.0;
      break;
    case NSImageRight:
      location.x = frame.size.width - width - 4.0;
      location.y = (frame.size.height - height) / 2.0;
      break;
    case NSImageBelow:
      location.x = (frame.size.width - width) / 2.0;
      location.y = frame.size.height - height - 2.0 ;
      break;
    case NSImageAbove:
      location.x = (frame.size.width - width) / 2.0;
      location.y = 1.0;
      break;
    case NSImageOnly:
      location.x = (frame.size.width - width) / 2.0;
      location.y = (frame.size.height - height) / 2.0;
      break;
    case NSImageOverlaps:
      // ?
      break;
    default:
      location.x = 1.0;
      location.y = 1.0;
    }
    /* Draw the Image */

    /* can't yet, draw a Rect instead */

    aRect.origin.x=location.x;
    aRect.origin.y=location.y;
    aRect.size.width=width;
    aRect.size.height=height;

    switch (buttonType) {
    case NSRadioButton :
    case NSSwitchButton :
      if ([self state]) {
	[[NSColor whiteColor] set];
	break;
      }
    default:
      [[NSColor blackColor] set];
    }

    [_controlView lockFocus];

    NSRectFill(aRect);

    [_controlView unlockFocus];    

}

- (void)getPeriodicDelay:(float *)delay interval:(float *)interval;
{
    *delay = periodicDelay;        /* I'm not sure I'm doing this right */
    *interval = periodicInterval;
}

- (int)highlightsBy;
{
    return highlightsBy;
}

- (NSCellImagePosition)imagePosition;
{
    return imagePosition;
}

- init;
{
    [super init];
    stringValue = nil;
    isOpaque = YES;
    isFloatAutoRange = YES;
    isEnabled=YES;
    state = 0;
    cellAlignment = NSLeftTextAlignment;
    cellType = NSNullCellType;
    entryType = NSAnyType;
    
    return self;
}

- (BOOL)isTransparent;
{
    if ([self isOpaque] == NO)
    {
	return YES;
    }
    else
    {
	return NO;
    }
}

- (NSString *)keyEquivalent;
{
    return keyEquivalent;
}

- (NSFont *)keyEquivalentFont;
{
    return keyEquivalentFont;
}

- (unsigned int)keyEquivalentModifierMask;
{
    return keyEquivalentModifierMask;
}

- (void)performClick:(id)sender;
{
    id aTarget = [self target];
    SEL aSel = [self action];
    
    /*
     * Should depress the button, or something like that,
     * according to how the cell wants to look like when
     * its state changes, then bring back the original state.
     */
       

    /* Cells send their action to their target when they are clicked. */
    if ((aTarget != nil) 
	&& (aSel != 0) 
	&& ([aTarget respondsToSelector:aSel] == YES))
    {
	[(NSObject *)aTarget perform:aSel];
    }
}

- (void)setAlternateImage:(NSImage *)anImage;
{
    altImage = anImage;
}

- (void)setAlternateTitle:(NSString *)aString;
{    
    if (altTitle != nil)
    {
	[altTitle release];
    }

    if (aString != nil)
    {
	altTitle = [[NSString alloc] initWithString:aString];
    }
}

- (void)setHighlightsBy:(int)aType;
{
    highlightsBy = aType;
}

- (void)setImagePosition:(NSCellImagePosition)aPosition;
{
    imagePosition = aPosition;
}

- (void)setKeyEquivalent:(NSString *)aKeyEquivalent;
{
    if (keyEquivalent != nil)
    {
	[keyEquivalent release];
    }

    if (aKeyEquivalent != nil)
    {
	keyEquivalent = [[NSString alloc] initWithString:aKeyEquivalent];
    }
}

/*
 * In setKeyEquivalentFont, we do not retain the Font object that we
 * get because Font objects are never freed or released during the
 * life of the application.
 */
- (void)setKeyEquivalentFont:(NSFont *)fontObj;
{
    keyEquivalentFont = fontObj; 
}

- (void)setKeyEquivalentFont:(NSString *)fontName size:(float)fontSize;
{
    keyEquivalentFont = [NSFont fontWithName:fontName size:fontSize];
}

- (void)setKeyEquivalentModifierMask:(unsigned int)mask;
{
    keyEquivalentModifierMask = mask;
}

- (void)setPeriodicDelay:(float)delay interval:(float)interval;
{
    periodicDelay = delay;
    periodicInterval = interval;
}

- (void)setShowsStateBy:(int)aType;
{
    showsStateBy = aType;
}

- (void)setState:(int)anInt;
{
    if (anInt == 0)
    {
	[self setIntValue:0];
    }
    else
    {
	[self setIntValue:1];
    }
}

- (void)setTitle:(NSString *)aString;
{   
    if (title != nil)
    {
	[title release];
    }

    if (aString != nil)
    {
	title = [[NSString alloc] initWithString:aString];
    }
}

- (void)setTransparent:(BOOL)flag;
{
    if (flag == YES)
    {
	isOpaque = NO;
    }
    else
    {
	isOpaque = YES;
    }
}

- (void)setType:(NSButtonType)aType; {
  buttonType = aType;

  // Initialise images for Radio and Switch buttons
  // Not very efficient...
  switch (aType) {
  case NSRadioButton:
    msg("initialising RadioButton\n");
    if (!image) {
      [self setImage: [NSImage alloc]];
    }
    if (!altImage) {
      [self setAlternateImage: [NSImage alloc]];
    }
    [self setImagePosition: NSImageLeft];
    break;
  case NSSwitchButton:
    msg("initialising SwitchButton\n");
    if (!image) {
      [self setImage: [NSImage alloc]];
      //	[image buildImageFromString:_NSSwitchON];
    }
    if (!altImage) {
      [self setAlternateImage: [NSImage alloc]];
    }
    [self setImagePosition: NSImageLeft];
    break;
  default:
  }
}

- (int)showsStateBy;
{
    return showsStateBy;
}

- (int)state;
{
    if ([self intValue] == 0)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}

- (NSString *)title;
{
    return title;
}

- (NSRect)titleRectForBounds:(NSRect)theRect;
{
#if 0
    fprintf(stdout, "Cell titleRectForBounds: %d %d %d %d\n",
	    theRect.origin.x, theRect.origin.y,
	    theRect.size.width, theRect.size.height);
#endif        
    /* Goofy implementation for testing only */
    return theRect;
}

- (NSCellType)type;
{
    return cellType;
}

- (NSButtonType)buttonType;
{
  return buttonType;
}

@end
