/*
 * Filename: NSCell.m
 * Project:  Pascal's X Kit
 *
 * Module:   Basic components
 *
 * Function: See header file
 *
 * Author:   Pascal Forget
 * Date:     January 1997
 *
 * Copyright (C) 1997 Pascal Forget.  All Rights Reserved.
 */

#include "NSCell.h"
#include "PXKList.h"
#include "PXKDrawing.h"
#include "NSControl.h"
#include "misc.h"

PXKList *_allCells; /* Contains all instances of the Cell class */
static NSView *_controlView;

@implementation NSCell

+ (void)initialize;
{
    if (!_allCells) {
	_allCells = [[PXKList alloc] init];
    }
}

+ (BOOL)prefersTrackingUntilMouseUp;  /* Subclasses may override */
{
    return NO; /* Tracking stops when the mouse leaves the NSCell */
}

- (SEL)action; /* Implemented by subclasses */
{
    return NULL;
}

- (NSTextAlignment)alignment;
{
    return cellAlignment;
}

- (void)calcDrawInfo:(NSRect)aRect;
{
    /* Implemented by subclasses as per OStep specification */
}

- (NSSize)cellSize;
{
    return cellSize;
}

- (NSSize)cellSizeForBounds:(NSRect)aRect;
{
    NSSize invalidSize;
    
    if ((aRect.size.width >= cellSize.width) &&
	(aRect.size.height >= cellSize.height)) {
	return aRect.size;
    } else {
	invalidSize.width = 0.0;
	invalidSize.height = 0.0;
	return invalidSize;
    }
}

- (NSComparisonResult)compare:(id)otherCell;
{
    if ((!otherCell) || ([otherCell isKindOfClass:[NSCell class]] == NO)) {
#if 0
	NSRaiseException(NSBadComparisonException); /* Right syntax? */
#else
	exit(-1); /* Crash the program */
#endif
    } else {
	return [[self stringValue] compare:[otherCell stringValue]];
    }
}

- (NSView *)controlView;
{
    return nil; /* Implemented by subclasses */
}

- (void)dealloc;
{
    if (stringValue != nil) {
	[stringValue release];
    }
    
    [_allCells removeObject:self];
    [super dealloc];
}

- (double)doubleValue;
{
    return [stringValue doubleValue];
}

- (NSRect)drawingRectForBounds:(NSRect)theRect;
{
    NSRect aRect;
    
    if (theRect.size.width >= cellSize.width) {
	aRect.size.width = theRect.size.width;
    } else {
	aRect.size.width = cellSize.width;
    }

    if (theRect.size.height >= cellSize.height) {
	aRect.size.height = theRect.size.height;
    } else {
	aRect.size.height = cellSize.height;
    }

    aRect.origin.x = theRect.origin.x;
    aRect.origin.y = theRect.origin.y;
    
    return aRect;
}

- (void)drawBorder;
{
    /* Not implemented yet */
}

- (void)drawImage;
{
    /* Not implemented yet */
}

- (void)drawTextInRect:(NSRect)frame;
{
    /* Goofy implementation for testing only */
    
    int width;
    int font_height;
    int alignment = 0;

    NSPoint location;
    

    dbug(msg("Cell text frame: %3.1f %3.1f %3.1f %3.1f\n",
	    frame.origin.x, frame.origin.y,
	    frame.size.width, frame.size.height));
    
    /* get string width for centering */
    width = [cellFont widthOfString: stringValue];

    font_height = [cellFont pointSize];

    /*
     * Wipe out the contents of the window by filling its
     * area with the backGroundColor (defined in NSView.h)
     */

    if (isOpaque) {
	[_controlView lockFocus];
	[[_controlView backgroundColor] set];
	NSRectFill(frame);
	[_controlView unlockFocus];
    }

#if 0
    if((alignment == NSLeftTextAlignment) && (width < frame.size.width)) {
	XDrawString(NSDisplay, [_controlView xWindow],
		    NSBlackGC,
		    frame.origin.x + 1,
		    frame.origin.y + font_height,
		    [stringValue cString], len1);
    } else if((alignment == NSRightTextAlignment) 
	      && ((width+5)<frame.size.width))
    {
	XDrawString(NSDisplay, [_controlView xWindow],
		    NSBlackGC,
		    frame.origin.x + frame.size.width - 5 - width,
		    frame.origin.y + font_height,
		    [stringValue cString], len1);
    } else {

	if (width > (frame.size.width-4)) {
	    XDrawString(NSDisplay, [_controlView xWindow],
			NSBlackGC,
			frame.origin.x + frame.size.width - width-5, 
			frame.origin.y + font_height,
			[stringValue cString], len1);	    
	} else {
	    /* output text, centered on each line */
	    XDrawString(NSDisplay, [_controlView xWindow],
			NSBlackGC,
			frame.origin.x + (frame.size.width - width-6)/2, 
			frame.origin.y + font_height,
			[stringValue cString], len1);
	}
    }
#else
    if((alignment == NSLeftTextAlignment) && (width < frame.size.width)) {
	location.x = frame.origin.x + 1.0;
	location.y = frame.origin.y + font_height;
	
    } else if((alignment == NSRightTextAlignment) 
	      && ((width+5)<frame.size.width))
    {
	location.x = frame.origin.x + frame.size.width - 2 - width;
	location.y = frame.origin.y + font_height;
    } else {

	if (width > (frame.size.width-4)) {
	    location.x = frame.origin.x + frame.size.width - width-2;
	    location.y = frame.origin.y + font_height;
	} else {
	    /* output text, centered on each line */
	    location.x = frame.origin.x + (frame.size.width - width-2)/2;
	    location.y = frame.origin.y + font_height;
	}
    }

#endif
    
    [[NSColor blackColor] set];
    [_controlView lockFocus];
    [cellFont set];
    pxkDrawString([stringValue cString], location);
    [_controlView unlockFocus];
}

- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView *)controlView;
{
    _controlView = controlView;
    
    if (image) {
	[self drawImage];
    }

    dbug(msg("Cell interior frame: %3.1f %3.1f %3.1f %3.1f\n",
	    cellFrame.origin.x, cellFrame.origin.y,
	    cellFrame.size.width, cellFrame.size.height));
    
    
    if (stringValue) {
#if 0	
	[self drawTextInRect:[self titleRectForBounds:cellFrame]];
#else
	[self drawTextInRect: cellFrame];
#endif
    }
}

- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView;
{
    [self drawBorder];
    [self drawInteriorWithFrame:cellFrame inView:controlView];
}

- (int)entryType;
{
    return entryType;
}

- (float)floatValue;
{
    return [stringValue floatValue];
}

- (NSFont *)font;
{
    return cellFont;
}

- (void)highlight:(BOOL)lit
	withFrame:(NSRect)cellFrame
	   inView:(NSView *)controlView;
{
    isHighlighted = lit;
    [self drawWithFrame:cellFrame inView:controlView];
}

- (NSImage *)image;
{
    return image;
}

- (NSRect)imageRectForBounds:(NSRect)theRect;
{
    NSRect imageRect;

    if ([self isBordered] == NO) {
	imageRect.origin.x = theRect.origin.x;
	imageRect.origin.y = theRect.origin.y;
	imageRect.size.width = NSX_MAX(cellSize.width,theRect.size.width);
	imageRect.size.height = NSX_MAX(cellSize.width,theRect.size.height);
    } else {
	imageRect.origin.x = theRect.origin.x + 1.0;
	imageRect.origin.y = theRect.origin.y + 1.0;
	imageRect.size.width = NSX_MAX(cellSize.width - 2.0,
				       theRect.size.width - 2.0);
	imageRect.size.height = NSX_MAX(cellSize.width - 2.0,
					theRect.size.height - 2.0);
    }
    return imageRect;
}

- init;
{
    [super init];
    stringValue = nil;
    isOpaque = YES;
    isFloatAutoRange = YES;
    state = 0;
    cellAlignment = NSLeftTextAlignment;
    cellType = NSNullCellType;
    entryType = NSAnyType;
    cellFont = [NSFont fontWithName:@"helvetica" size:12.0];
    [_allCells addObject:self];
    
    return self;
}

- (id)initImageCell:(NSImage *)anImage;
{
    [self init];

    if (anImage != nil) {
	image = [anImage retain];
    }

    cellType = NSImageCellType;
    
    return self;
}

- (id)initTextCell:(NSString *)aString;
{
    [self init];
    
    if (aString != nil) {
	stringValue = [aString retain];
    }

    cellType = NSTextCellType;
    
    return self;
}

- (int)intValue;
{
    return [stringValue intValue];
}

- (BOOL)isBezeled;
{
    return isBezeled;
}

- (BOOL)isBordered;
{
    return isBordered;
}

- (BOOL)isContinuous;
{
    return isContinuous;
}

- (BOOL)isEditable;
{
    return isEditable;
}

- (BOOL)isEnabled;
{
    return isEnabled;
}

- (BOOL)isEntryAcceptable:(NSString *)aString;
{
    BOOL isNumeric = YES;   /* Int, float or double */
    BOOL isPositive = YES;  /* Numeric > 0 */
    BOOL isInt = YES;
    int i=0, count;
    const char *str;
    char c;
    int numberOfDots = 0;
    
    if ((entryType == NSAnyType) || (aString == nil)) {
	return YES;
    } else {

	count = [aString length];

	str = [aString cString];

	if (entryType == NSDateType) { /* VERY primitive date validation */
	    
	    for (i=0; i<count; i++) {
		c = str[i];
		if ((c=='k') ||
		    (c=='w') ||
		    (c=='x') ||
		    (c=='z') ||
		    (c<32) ||
		    (c>122))
		{
		    return NO;
		}
	    }
	    return YES;
	    
	} else {   /* Entry is of one of the numeric types */

	    /* Ignore leading blanks and TABS */

	    while (((c=str[i]) == ' ') || (c=='\t')) {
		i++;
	    }

	    /* See if the current char is a minus sign */

	    if (c == '-') {
		isPositive = NO;
	    }

	    for (; i<count; i++) {
		c = str[i];

		if (c == '.') {
		    numberOfDots++;
		    
		    if (numberOfDots > 1) {
			isNumeric = NO;
		    }
		} else {
		    
		    if ((c>'9') || (c < '0')) {
			isNumeric = NO;
		    }
		}
	    }

	    if (isNumeric &&
		isInt &&
		(((float)[aString intValue]) == [aString floatValue]))
	    {
		isInt = YES;
	    } else {
		isInt = NO;
	    }

	    if (entryType == NSIntType) {
		if (isNumeric && isInt) {
		    return YES;
		} else {
		    return NO;
		}
	    } else if (entryType == NSPositiveIntType) {
		if (isNumeric && isInt && isPositive) {
		    return YES;
		} else {
		    return NO;
		}
	    } else if (entryType == NSFloatType) {
		if (isNumeric && (isInt == NO)) {
		    return YES;
		} else {
		    return NO;
		}
	    } else if (entryType == NSPositiveFloatType) {
		if (isNumeric && (isInt == NO) && isPositive) {
		    return YES;
		} else {
		    return NO;
		}		
	    } else if (entryType == NSDoubleType) {
		if (isNumeric && (isInt == NO)) {
		    return YES;
		} else {
		    return NO;
		}
	    } else if (entryType == NSPositiveDoubleType) {
		if (isNumeric && (isInt == NO) && isPositive) {
		    return YES;
		} else {
		    return NO;
		}		
	    }	   
	}
    }
    return YES;
}

- (BOOL)isHighlighted;
{
    return isHighlighted;
}

- (BOOL)isOpaque;
{
    return isOpaque;
}

- (BOOL)isScrollable;
{
    return isScrollable;
}

- (BOOL)isSelectable;
{
    return isSelectable;
}

- (NSString *)keyEquivalent;
{
    return nil; /* Implemented by subclasses. */
}

- (id)representedObject;
{
    return representedObject;
}

- (void)setAction:(SEL)aSelector;
{
    /* 
     *  implemented by subclasses.  Suggestion:
     *     action = aSelector;
     */
}

- (void)setAlignment:(NSTextAlignment)mode;
{
    cellAlignment = mode;
}

- (void)setBezeled:(BOOL)flag;
{
    isBezeled = flag;
}

- (void)setBordered:(BOOL)flag;
{
    isBordered = flag;
}

- (void)setContinuous:(BOOL)flag;
{
    isContinuous = flag;
}

- (void)setDoubleValue:(double)aDouble;
{
    char buf[128];
    sprintf(buf, "%f", aDouble);
    [self setStringValue:[NSString stringWithCString:buf]];
}

- (void)setEditable:(BOOL)flag;
{
    isEditable = flag;
}

- (void)setEnabled:(BOOL)flag;
{
    isEnabled = flag;
}

- (void)setEntryType:(int)aType;
{
    entryType = aType;
}

- (void)setFloatingPointFormat:(BOOL)autoRange
			  left:(unsigned int)leftDigits
			 right:(unsigned int)rightDigits;
{
    isFloatAutoRange = autoRange;
    floatLeftDigits = leftDigits;
    floatRightDigits = rightDigits;
}

- (void)setFloatValue:(float)aFloat;
{
    char buf[128];
    sprintf(buf, "%f", aFloat);
    [self setStringValue:[NSString stringWithCString:buf]];
}

- (void)setFont:(NSFont *)fontObject;
{
    cellFont = fontObject;
}

- (void)setImage:(NSImage *)anImage;
{
    if (image != anImage) {
	if (image != nil) {
	    [image release];
	}

	if (anImage != nil) {
	    image = [anImage retain];
	} else {
	    image = nil;
	}
    }
}

- (void)setIntValue:(int)anInt;
{
    char buf[16];
    NSString *str;

    sprintf(buf, "%i", anInt);
    
    str = [[NSString alloc] initWithCString:buf];
    
    [self setStringValue:str];
    
    [str release];
}

- (void)setOpaque:(BOOL)flag;
{
    isOpaque = flag;
}

- (void)setRepresentedObject:(id)anObject;
{
    int i, count;
    NSCell *aCell;
    
    if (anObject == nil) {
	representedObject = nil; /* Set the instance variable, we're done. */
    } else {
	
	representedObject = anObject; /* Set the instance variable */
	
	count = [_allCells count];

	/*
	 * Look for an existing association between another cell
	 * and anObject
	 */
	
	for (i=0; i<count; i++) {
	    aCell = [_allCells objectAtIndex:i];
	
	    if ((aCell != self) && ([aCell representedObject] == anObject)) {
		[aCell setRepresentedObject:nil];
		return;
	    }
	}
    }
}

- (void)setScrollable:(BOOL)flag;
{
    isScrollable = flag;
}

- (void)setSelectable:(BOOL)flag;
{
    isSelectable = flag;
}

- (void)setState:(int)value;
{
    if ((value == 0) || (value == 1)) {
	state = value;
    }
}

- (void)setStringValue:(NSString *)aString;
{
    if (aString != stringValue) {

	if (stringValue != nil) {
	    [stringValue release];
	}

	if (aString != nil) {
	    stringValue = [aString copyWithZone:[self zone]];
	} else {
	    stringValue = nil; /* We don't send the retain message to nil */
	}
    }
}

- (void)setTag:(int)anInt;
{
    /* implemented by subclasses */
}

- (void)setTarget:(id)anObject;
{
    /* implemented by subclasses */
}

- (void)setType:(NSCellType)aType;
{
    cellType = aType;
}

- (void)setWraps:(BOOL)flag;
{
    wraps = flag;
}

- (int)state;
{
    return state;
}

- (NSString *)stringValue;
{
    return stringValue;
}

- (int)tag;
{
    return 0; /* implemented by subclasses */
}

- (void)takeDoubleValueFrom:(id)sender;
{
    [self setDoubleValue:[(NSCell *)sender doubleValue]];
}

- (void)takeFloatValueFrom:(id)sender;
{
    [self setFloatValue:[(NSCell *)sender floatValue]];
}

- (void)takeIntValueFrom:(id)sender;
{
    [self setIntValue:[(NSCell *)sender intValue]];
}

- (void)takeStringValueFrom:(id)sender;
{
    [self setStringValue:[(NSCell *)sender stringValue]];
}

- (id)target;
{
    return nil;
}

- (NSRect)titleRectForBounds:(NSRect)theRect;
{
    dbug(msg("Cell titleRectForBounds: %d %d %d %d\n",
	    theRect.origin.x, theRect.origin.y,
	    theRect.size.width, theRect.size.height));
    /* Goofy implementation for testing only */
    return theRect;
}

- (BOOL)trackMouse:(NSEvent *)theEvent
	    inRect:(NSRect)cellFrame
	    ofView:(NSView *)controlView
      untilMouseUp:(BOOL)flag;
{
    /*
     * while event.location is inside my frame and
     * the mouse button is still down
     */
    
    while (NO) {
	/* not implemented yet */
    }
    
    /*
     * if mouse is up while still in frame return YES, otherwise return NO
     */
    
    return NO;
}

- (NSCellType)type;
{
    return cellType;
}

- (BOOL)wraps;
{
    return wraps;
}

@end
