/*
 * Filename: NSColor.m
 * Project:  Pascal's X kit
 * Module:   Basic application components
 *
 * Function: This is the Color class in its second incarnation.
 *           Backwards compatible with first release, but much more
 *           complete.
 *
 * Author:   Pascal Forget
 * Date:     January 1997
 *
 * Copyright (C) 1997 Pascal Forget. All Rights Reserved.
 */

#include "NSColor.h"
#include "PXKDrawing.h"

static BOOL ignoresAlpha;

#ifndef PXK_MAX
#define PXK_MAX(x,y) (x>y)?x:y
#endif
#ifndef PXK_MIN
#define PXK_MIN(x,y) (x<y)?x:y
#endif

@interface NSColor (Private)
- initWithCalibratedWhite:(float)white alpha:(float)alpha;
- initWithCalibratedRed:(float)red
		  green:(float)green 
                   blue:(float)blue
		  alpha:(float)alpha;
@end

@implementation NSColor

+ (NSColor *)blackColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:0.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)blueColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:0.0 green:0.0 blue:1.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)brownColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:0.6 green:0.4 blue:0.2 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)clearColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:0.0 alpha:0.0];
    }
    
    return color;
}

+ (NSColor *)colorWithCalibratedRed:(float)red
			      green:(float)green 
                               blue:(float)blue
			      alpha:(float)alpha;
{
    float alphaComp, redComp, greenComp, blueComp;

    // All values are legal, but values less than 0.0 are set to 0.0,
    //   and values greater than 1.0 are set to 1.0.
    
    alphaComp = PXK_MAX(alpha,0.0);
    alphaComp = PXK_MIN(alphaComp,1.0);    
    redComp = PXK_MAX(red,0.0);
    redComp = PXK_MIN(redComp,1.0);
    greenComp = PXK_MAX(green,0.0);
    greenComp = PXK_MIN(greenComp,1.0);
    blueComp = PXK_MAX(blue,0.0);
    blueComp = PXK_MIN(blueComp,1.0);

    return [[NSColor alloc] initWithCalibratedRed: redComp
			    green: greenComp
			    blue: blueComp
			    alpha: alphaComp];
}

+ (NSColor *)colorWithCalibratedWhite:(float)white alpha:(float)alpha;
{
    float alphaComp = alpha;
    float whiteComp = white; 
    
    if (alphaComp < 0.0) {
	alphaComp = 0.0;
    }

    if (alphaComp > 1.0) {
	alphaComp = 1.0;
    }

    if (whiteComp < 0.0) {
	whiteComp = 0.0;
    }

    if (whiteComp < 0.0) {
	whiteComp = 0.0;
    }

    return [[NSColor alloc] initWithCalibratedWhite:whiteComp
			    alpha:alphaComp];
}

+ (NSColor *)cyanColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:0.0 green:1.0 blue:1.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)darkGrayColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:0.3333333333333 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)grayColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:0.5 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)greenColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:0.0 green:1.0 blue:0.0 alpha:1.0];
    }
    
    return color;
}

+ (BOOL)ignoresAlpha;
{
    return ignoresAlpha;
}

+ (void)initialize;
{
    if (self == [NSColor class]) {
	ignoresAlpha = YES; /* Default value */
    }
}

+ (NSColor *)lightGrayColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:0.666666666666 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)magentaColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:1.0 green:0.0 blue:1.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)orangeColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:1.0 green:0.5 blue:0.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)purpleColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:0.5 green:0.0 blue:0.5 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)redColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:1.0 green:0.0 blue:0.0 alpha:1.0];
    }
    
    return color;
}

+ (void)setIgnoresAlpha:(BOOL)flag;
{
    ignoresAlpha = flag;
}

+ (NSColor *)whiteColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedWhite:1.0 alpha:1.0];
    }
    
    return color;
}

+ (NSColor *)yellowColor;
{
    static NSColor *color;

    if (color == nil) {
	color = [self colorWithCalibratedRed:1.0 green:1.0 blue:0.0 alpha:1.0];
    }
    
    return color;
}

/* Instance methods */

- initWithCalibratedRed:(float)red
		  green:(float)green 
                   blue:(float)blue
		  alpha:(float)alpha;
{
    [self init];
    alphaComponent = alpha;
    redComponent = red;
    greenComponent = green;
    blueComponent = blue;
    colorSpaceName = [[NSString alloc] initWithCString:
				       "NSCalibratedRGBColorSpace"];
    return self;
}

- initWithCalibratedWhite:(float)white alpha:(float)alpha;
{
    [self init];
    alphaComponent = alpha;
    whiteComponent = white;
    colorSpaceName = [[NSString alloc] initWithCString:
				       "NSCalibratedWhiteColorSpace"];
    return self;
}

- (float)alphaComponent;
{
    return alphaComponent;
}

- (float)blueComponent;
{
    return blueComponent;
}

- (NSString *)colorSpaceName;
{
    return colorSpaceName;
}

- (void)getRed:(float *)red
	 green:(float *)green 
          blue:(float *)blue 
	 alpha:(float *)alpha;
{
    if (red != NULL) *red = redComponent;
    if (green != NULL) *green = greenComponent;
    if (blue != NULL) *blue = blueComponent;
    if (alpha != NULL) *alpha = alphaComponent;
}

- (void)getWhite:(float *)white 
	 alpha:(float *)alpha;
{
    if (white != NULL) *white = whiteComponent;
    if (alpha != NULL) *alpha = alphaComponent;    
}

- (float)greenComponent;
{
    return greenComponent;
}

- (float)redComponent;
{
    return redComponent;
}

- (void)set;
{
    setCurrentColor(self);
}

- (float)whiteComponent;
{
    return whiteComponent;
}

@end
