/*
 * Filename: NSControl.m
 * Project:  Pascal's X kit
 * Module:   Basic application components
 *
 * Function: This is the Control class.  See header file for a more
 *           complete description
 *
 * Author:   Pascal Forget
 * Date:     March 1996
 *
 * Copyright (C) 1996 Free Software Foundation
 */

#include "NSControl.h"

@interface NSControl(NonOpenStep)
- (NSRect)cellframe:(NSCell *)aCell;
@end

@implementation NSControl

+ (Class)cellClass;
{
    return nil; /* Overridden by subclasses */
}

- (SEL)action;
{
    return action;
}

- (NSTextAlignment)alignment;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell alignment];
    } else {
	return NSLeftTextAlignment;
    }
}   

- (id)cell;
{
    return cell;
}

- (double)doubleValue;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell doubleValue];
    } else {
	return 0.0;
    }
}   
    
- (float)floatValue;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell floatValue];
    } else {
	return 0.0;
    }
}

- (NSFont *)font;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell font];
    } else {
	return nil;
    }
}

- (BOOL)ignoresMultiClick;
{
    return ignoresMultiClick;
}

- (id)initWithFrame:(NSRect)frameRect;
{
    [super initWithFrame:frameRect];
    isEnabled = YES;
    target = nil;
    state = 0;    
    return self;
}

- (int)intValue;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell intValue];
    } else {
	return 0;
    }
}

- (BOOL)isContinuous;
{
    return isContinuous;
}

- (BOOL)isEnabled;
{
    return isEnabled;
}

- (void)mouseDown:(NSEvent *)theEvent;
{
    NSCell *aCell;
    NSRect aRect = [self frame];
    
    /* self update selected cell based on event location */
    
    /*
      ...
     
     */
    
    /* Get the selected cell */
    aCell= [self selectedCell];
    
    if (aCell != nil) {

	/* highlight cell */
	[aCell setState:1];
	
	aRect.size = [aCell cellSizeForBounds:aRect];
	
	/* instruct the cell to keep track of the mouse cursor */
	[aCell trackMouse:theEvent
	       inRect:aRect
	       ofView:self
	       untilMouseUp:YES]; /* untilMouseUp, how to determine */
	       
	/* unhighlight cell */
	[aCell setState:0];

    }
}

- (id)selectedCell;
{
    /*
     * overridden by some classes such as NSMatrix
     */    
    return cell;
}

- (int)selectedTag;
{
    /*
     * overridden by some classes such as NSMatrix
     */
    return [(NSCell *)cell tag];
}

- sendAction;
{
    if (action && target) {
	return [target perform:action];
    }
    return nil;
}

- (void)setAction:(SEL)aSelector;
{
    action = aSelector;
}

- (void)setAlignment:(NSTextAlignment)mode;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setAlignment:mode];
    }
}

- (void)setCell:(NSCell *)aCell;
{
    cell = aCell;
}

- (void)setContinuous:(BOOL)flag;
{
    isContinuous = flag;
}

- (void)setDoubleValue:(double)aDouble;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setDoubleValue:aDouble];
    }
}
    
- (void)setEnabled:(BOOL)flag;
{
    isEnabled = flag;
}

- (void)setFloatValue:(float)aFloat;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setFloatValue:aFloat];
    }
}

- (void)setFloatingPointFormat:(BOOL)autoRange
			  left:(unsigned int)leftDigits
			 right:(unsigned int)rightDigits;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setFloatingPointFormat:autoRange
	       left:leftDigits
	       right:rightDigits];
    }
}

- (void)setFont:(NSFont *)fontObject;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setFont:fontObject];
    }
}

- (void)setIgnoresMultiClick:(BOOL)flag;
{
    ignoresMultiClick = flag;
}

- (void)setIntValue:(int)anInt;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setIntValue:anInt];
    }
}
    
- setState:(int)newState;
{
    state = newState;
    return self;
}

- (void)setStringValue:(NSString *)aString;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setStringValue:aString];
    }
}

- (void)setTag:(int)anInt;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setTag:anInt];
    }
}
    

- (void)setTarget:(id)anObject;
{
    target = anObject;
}

- (int)state;
{
    return state;
}

- (NSString *)stringValue;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell stringValue];
    } else {
	return nil;
    }
}

- (int)tag;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	return [aCell tag];
    } else {
	return 0;
    }
}

/* Interacting with Other Controls */

- (void)takeDoubleValueFrom:(id)sender;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setDoubleValue:[(NSControl *)sender doubleValue]];
    }
}

- (void)takeFloatValueFrom:(id)sender;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setFloatValue:[(NSControl *)sender floatValue]];
    }
}

- (void)takeIntValueFrom:(id)sender;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setIntValue:[(NSControl *)sender intValue]];
    }
}

- (void)takeStringValueFrom:(id)sender;
{
    NSCell *aCell = [self selectedCell];
    
    if (aCell != nil) {
	[aCell setStringValue:[(NSControl *)sender stringValue]];
    }
}

- (id)target;
{
    return target;
}

- (int)toggleState; /* returns the new state */
{
    if (state == 0) {
	state = 1;
    } else {
	state = 0;
    }
    return state;
}

// subclasses should override as needed
- (void)updateCell:(NSCell *)aCell;
{
    [cell drawWithFrame: [self frame] inView: self];
}

// subclasses should override as needed
- (void)updateCellInside:(NSCell *)aCell;
{
    [cell drawInteriorWithFrame: [self frame] inView: self];    
}

@end

@implementation NSControl(NonOpenStep)

/*
 * Default implementation.  Will be overridden for controls that
 * contain more than one cell, such as NSMatrix, where the cell's
 * frame is different than the control's frame
 */
- (NSRect)cellframe:(NSCell *)aCell;
{
    return [self frame];
}

@end
