/*
 * Filename: NSMatrix.m
 * Project:  Pascal's X kit
 * Module:   Basic application components
 *
 * Function: See header file for a description
 *
 * Author:   Pascal Forget
 * Date:     March 1996
 *
 * Copyright (C) 1996 Free Software Foundation, Inc.
 */

#include "NSMatrix.h"
#include "NSActionCell.h"
#include "PXKShortInt.h"
#include "misc.h"

#define PXK_DEFAULT_CELL_HEIGHT 20

Class defaultCellClass;

@implementation NSMatrix

+ (Class)cellClass;
{
    return defaultCellClass;
}

+ (void)setCellClass:(Class)classId;
{
    defaultCellClass = classId;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent;
{
    if (mode == NSListModeMatrix)
    {
	return NO;
    }
    return YES;
}

- (void)addColumn;
{
    int i;
    PXKShortInt *anInt;
    NSMutableArray *aRow;

    if (numRows <= 0)
    {
	[rows addObject:[[NSMutableArray alloc] init]];
	numRows = 1;

	[rowHeights removeAllObjects];
	anInt = [[PXKShortInt alloc] init];
	[anInt setIntValue:PXK_DEFAULT_CELL_HEIGHT];
	[rowHeights addObject:anInt];
    }
    
    for (i=0; i<numRows; i++)
    {
	aRow = (NSMutableArray *)[rows objectAtIndex:i];

	if (cellPrototype != nil)
	{
	    [aRow addObject:[cellPrototype copy]];
	}
	else
	{
	    [aRow addObject:[[cellClass alloc] init]];
	}

	anInt = [[PXKShortInt alloc] init];
	[anInt setIntValue:frame.size.width / (float)(numCols+1)];
	[colWidths addObject:anInt];
	
    }
    numCols++;
}

- (void)addColumnWithCells:(NSArray *)cellArray;
{
    [self insertColumn:numCols withCells:cellArray];
}

- (void)addRow;
{
    int i;
    PXKShortInt *anInt;
    NSMutableArray *newRow = [[NSMutableArray alloc] init];

    if (cellPrototype != nil)
    {
	[newRow addObject:[cellPrototype copy]];
    }
    else
    {
	[newRow addObject:[[cellClass alloc] init]];
    }
    
    if (numCols > 0)
    {
	for (i=1; i<numCols; i++)
	{
	    if (cellPrototype != nil)
	    {
		[newRow addObject:[cellPrototype copy]];
	    }
	    else
	    {
		[newRow addObject:[[cellClass alloc] init]];
	    }
	}
	
	anInt = [[PXKShortInt alloc] init];
	[anInt setIntValue:PXK_DEFAULT_CELL_HEIGHT];
	[rowHeights addObject:anInt];
    
	[rows addObject:newRow];
	numRows++;

    }
    else
    {
	[self addColumn];
    }
}

- (void)addRowWithCells:(NSArray *)cellArray;
{
    [self insertRow:numRows withCells:cellArray];
}

- (BOOL)allowsEmptySelection;
{
    return allowsEmptySelection;
}

- (BOOL)autosizesCells;
{
    return autosize;
}

- (id)cellAtRow:(int)row column:(int)column;
{
    if ((row >= numRows) || (row < 0))
    {
	fprintf(stderr, "NSMatrix cellAt:: invalid row (%d)\n", row);
	return nil;
    }

    if ((column >= numCols) || (column < 0))
    {
	fprintf(stderr, "NSMatrix cellAt:: invalid column (%d)\n", column);
	return nil;
    }

    return [(NSArray *)[rows objectAtIndex:row] objectAtIndex:column];
}

- (Class)cellClass;
{
    return cellClass;
}

- (NSRect)cellFrameAtRow:(int)row column:(int)column;
{
    NSRect r;
    int i;

    /* Initialize the size to all zeroes */
    r.origin.x = 0;
    r.origin.y = 0;
    r.size.width = 0;
    r.size.height = 0;

    /* Validate arguments */
    if ((row >= numRows) || (row < 0))
    {
	return r;
    }

    if ((column >= numCols) || (column < 0))
    {
	return r;
    }    

    /* Compute the x origin */
    for (i=0; i<column; i++)
    {
	r.origin.x += [(PXKShortInt *)[colWidths objectAtIndex:i] intValue];
	r.origin.x += interCell.width;
    }

    /* Get the column width */
    r.size.width = [(PXKShortInt *)[colWidths objectAtIndex:i] intValue];

    /* Compute the y origin */
    for (i=0; i<row; i++)
    {
	r.origin.y += [(PXKShortInt *)[rowHeights objectAtIndex:i] intValue];
	r.origin.y += interCell.height;
    }

    /* Get the height */
    r.size.height = [(PXKShortInt *)[rowHeights objectAtIndex:i] intValue];

    dbug(msg("Matrix cellFrameAtRow: %d column:%d"
	    " is: x %d y %d w %d h %d\n",
	    row, column,
	    r.origin.x, r.origin.y, r.size.width, r.size.height));
    
    return r;
}

- (NSArray *)cells;    /* return an array of the matrix' cells. */
{
    NSMutableArray *cells;
    int i;
    
    if ((numCols < 1) || (numRows < 1))
    {
	return nil;
    }
    else
    {
	cells = [[NSMutableArray alloc] initWithCapacity:(numRows * numCols)];

	for (i=0; i<numRows; i++)
	{
	    [cells addObjectsFromArray:(NSArray *)[rows objectAtIndex:i]];
	}

	return cells;
    }
}

- (NSSize)cellSize;
{
    /*****************************************
     * This method is for the OpenStep mode, *
     * where all rows have the same size.    *                                
     *****************************************/

    NSSize s;

    if ((numCols<1) || (numRows<1))
    {
	return s;
    }

    s.width = [[colWidths objectAtIndex:0] intValue];
    s.height= [[rowHeights objectAtIndex:0] intValue];

    return s;
}

- (id)cellWithTag:(int)anInt;
{
    int i, j;
    NSMutableArray *aRow;
    NSCell *aCell;
    
    for (i=0; i<numRows; i++)
    {
	aRow = (NSMutableArray *)[rows objectAtIndex:i];

	for (j=0; j<numCols; j++)
	{
	    aCell = [aRow objectAtIndex:j];
	    
	    if ([aCell tag] == anInt)
	    {
		return (id)aCell;
	    }
	}
    }
    return nil;
}

- (void)deselectAllCells;
{
    [selectedCells removeAllObjects];
}

- (void)deselectSelectedCell;
{
    [selectedCells removeLastObject];
}

- (void)display;
{
    int row,col;
    NSMutableArray *aRow;

    dbug(msg("Matrix display\n"));
    
    [super display];

    for (row=0; row<numRows; row++)
    {
	aRow = [rows objectAtIndex:row];
	
	for (col=0; col<numCols; col++)
	{
	    [[aRow objectAtIndex:col] drawWithFrame:
				      [self cellFrameAtRow:row column:col]
				      inView:self];
	}
    }
}

- (SEL)doubleAction;
{
    return doubleAction;
}

- (BOOL)drawsBackground;
{
    return drawsBackground;
}

- (BOOL)drawsCellBackground;
{
    return drawsCellBackground;
}

- empty;
{
    [rows removeAllObjects];
    [selectedCells removeAllObjects];
    numRows = 0;
    numCols = 0;
    return self;
}

- (SEL)errorAction;
{
    return errorAction;
}

- (void)getNumberOfRows:(int *)rowCount columns:(int *)columnCount;
{
    *rowCount = numRows;
    *columnCount = numCols;
}

/*
 * The next function returns NO if the point is located inside
 * the matrix but also in an intercell gap, which may or may
 * not be the behavior God intended...
 */

- (BOOL)getRow:(int *)row column:(int *)column forPoint:(NSPoint)aPoint;
{
    NSRect myFrame = [self frame];
    NSRect cFrame;
    int i, j;

    /* Trivial rejection if the point is not in the Matrix's frame rect */

    if ((aPoint.x < myFrame.origin.x)
	|| (aPoint.x > (myFrame.origin.x + myFrame.size.width))
	|| (aPoint.y < myFrame.origin.y)
	|| (aPoint.y > (myFrame.origin.y + myFrame.size.height)))
    {
	return NO;
    }
    else
    {
	/* Here an optimized algo could be used at the expense of clarity */

	for (i=0; i<numRows; i++)
	{
	    for (j=0; j<numCols; j++)
	    {
		cFrame = [self cellFrameAtRow:i column:j];

		if ((aPoint.x >= cFrame.origin.x)
		    && (aPoint.x <= (cFrame.origin.x + cFrame.size.width))
		    && (aPoint.y >= cFrame.origin.y)
		    && (aPoint.y <= (cFrame.origin.y + cFrame.size.height)))
		{
		    *row = i;
		    *column = j;
		    return YES;
		}
	    }
	}
	return NO;
    }

}

- (BOOL)getRow:(int *)row column:(int *)column ofCell:(NSCell *)aCell;
{
    int i, j;

    for (i=0; i<numRows; i++)
    {
	for (j=0; j<numCols; j++)
	{
	    if ((NSCell *)[self cellAtRow:i column:j] == aCell)
	    {
		*row = i;
		*column = j;
		return YES;
	    }
	}
    }
    return NO;
}

- (void)highlightCell:(BOOL)flag atRow:(int)row column:(int)column;
{
    NSCell *aCell = [self cellAtRow:row column:column];
    NSRect cellFrame;

    if (aCell != nil)
    {
	cellFrame = [self cellFrameAtRow:row column:column];
	[aCell highlight:flag withFrame:cellFrame inView:self];
    }

}

- (id)initWithFrame:(NSRect)frameRect;
{
    return [self initWithFrame:frameRect
		 mode:NSTrackModeMatrix
		 cellClass:[NSCell class]
		 numberOfRows:0
		 numberOfColumns:0];
}

- (id)initWithFrame:(NSRect)frameRect
	       mode:(int)aMode
	  cellClass:(Class)classId
       numberOfRows:(int)rowsHigh
    numberOfColumns:(int)colsWide
{
    [super initWithFrame:frameRect];


    if (rowsHigh < 0)
    {
	fprintf(stderr, "NSMatrix initWithFrame:mode:... number of "
		"rows has to be >= 0.  Will create matrix with 0 rows.\n");

	numRows = 0;
    }
    else
    {
	numRows = rowsHigh;
    }

    if (colsWide < 0)
    {
	fprintf(stderr, "NSMatrix initWithFrame:mode:... number of "
		"columns has to be >= 0.  "
		"Will create matrix with 0 columns.\n");

	numCols = 0;
    }
    else
    {
	numCols = colsWide;
    }

    rows = [[NSMutableArray alloc] init];    
    rowHeights = [[NSMutableArray alloc] init];
    colWidths = [[NSMutableArray alloc] init];
    selectedCells = [[NSMutableArray alloc] init];
    interCell.width = interCell.height = 0;
    cellPrototype = nil;

    return self;
}

- (id)initWithFrame:(NSRect)frameRect
	       mode:(int)aMode
	  prototype:(NSCell *)aCell
       numberOfRows:(int)rowsHigh
    numberOfColumns:(int)colsWide
{
    [self initWithFrame:frameRect
	  mode:NSTrackModeMatrix
	  cellClass:[NSCell class]
	  numberOfRows:0
	  numberOfColumns:0];

    if (aCell == nil)
    {
	fprintf(stderr, "NSMatrix "
		"initWithFrame:mode:prototype:numberOfRows:numberOfColumns: "
		"prototype can't be nil. "
		" Using NSCell as the default class.\n");
	
	cellPrototype = nil;

	cellClass = [NSCell class];
    }
    else
    {
	cellPrototype = [aCell retain];
    }
    
    return self;
}

- (void)insertColumn:(int)column;
{
    NSMutableArray *currentRow;
    id newCell;
    int i, j;

    i = column - numCols;

    while (i--)
    {
	[self addColumn];
    }

    if (numCols > 1)
    {
	for (i=0; i<numRows; i++)
	{
	    currentRow = [rows objectAtIndex:i];
	    newCell = [currentRow lastObject];

	    for (j=numCols-1; j>column; j--)
	    {
		[currentRow replaceObjectAtIndex:j withObject:
			    [currentRow objectAtIndex:i-1]];
	    }
	    [currentRow replaceObjectAtIndex:column withObject:newCell];
	}
    }
}

- (void)insertColumn:(int)column withCells:(NSArray *)cellArray;
{
    NSCell *newCell;
    int i, count;

    [self insertColumn:column];

    if (cellArray != nil)
    {
	count = [cellArray count];
    }
    else
    {
	count = 0;
    }

    for (i=0; i<numRows; i++)
    {
	if (i < count)
	{
	    newCell = (NSCell *)[cellArray objectAtIndex:i];

	    [self putCell:newCell atRow:i  column:column];
	}
	else
	{
	    break;
	}
    }
}

- (void)insertRow:(int)row;
{
    NSMutableArray *aRow;
    int i;

    i = row - numRows;

    while (i--)
    {
	[self addRow];
    }

    if (numRows > 1)
    {
	aRow = [rows lastObject];

	for (i=numRows-1; i> row; i--)
	{
	    [rows replaceObjectAtIndex:i withObject:[rows objectAtIndex:i-1]];
	}
	[rows replaceObjectAtIndex:row withObject:aRow];
    }
}

- (void)insertRow:(int)row withCells:(NSArray *)cellArray;
{
    NSCell *newCell;
    int i, count;

    [self insertRow:row];

    if (cellArray != nil)
    {
	count = [cellArray count];
    }
    else
    {
	count = 0;
    }

    for (i=0; i<numCols; i++)
    {
	if (i < count)
	{
	    newCell = (NSCell *)[cellArray objectAtIndex:i];

	    [self putCell:newCell atRow:row  column:i];
	}
	else
	{
	    break;
	}
    }
}

- (NSSize)intercellSpacing;
{
    return interCell;
}

- (BOOL)isAutoscroll;
{
    return autoscroll;
}

- (BOOL)isSelectionByRect;
{
    return selectionByRect;
}

- (id)makeCellAtRow:(int)row column:(int)column;
{
    id newCell;

    if ((row<numRows)
	&& (row>-1)
	&& (column<numCols)
	&& (column>-1))
    {
	newCell = [[cellClass alloc] init];
	[[rows objectAtIndex:row] replaceObjectAtIndex:column with:newCell];
	return newCell;
    }
    else
    {
	return nil;
    }
}

- (NSMatrixMode)mode;
{
    return mode;
}

- (void)mouseDown:(NSEvent *)theEvent;
{
    NSView *view = [[self window] contentView];
    NSPoint location = [theEvent locationInWindow];
    NSPoint point = [view convertPoint:location toView:self];
    NSCell *aCell;
    int aRow, aCol;

    [self getRow:&aRow column:&aCol forPoint:point];

    aCell = [self cellAtRow:aRow column:aCol];

    if (aCell != nil)
    {
	[self selectCellAtRow:aRow column:aCol];
    }

    if ([theEvent clickCount] == 1)
    {
	[self sendAction];
    }
    else
    {
	[self sendDoubleAction];
    }
}

- (int)numCols;
{
    return numCols;
}

- (int)numRows;
{
    return numRows;
}

- (id)prototype;
{
    return cellPrototype;
}

- (void)putCell:(id)newCell atRow:(int)row column:(int)column;
{
    NSMutableArray *aRow;

    if ((row>-1)
	&& (row<numRows)
	&& (column>-1)
	&& (column<numCols))
    {
	aRow = (NSMutableArray *)[rows objectAtIndex:row];
	[aRow replaceObjectAtIndex:column withObject:newCell];
    }
}
- (void)removeColumn:(int)column;
{
    int i;
    NSMutableArray *aRow;
    
    if ((column > -1) && (column < numCols))
    {
	for (i=0; i<numRows; i++)
	{
	    aRow = (NSMutableArray *)[rows objectAtIndex:i];
	    [aRow removeObjectAtIndex:column];
	}
	[colWidths removeObjectAtIndex:column];
	numCols--;
    }
}

- (void)removeRow:(int)row;
{
    if ((row > -1) && (row < numRows))
    {
	[rows removeObjectAtIndex:row];
	
	[rowHeights removeObjectAtIndex:row];
	numRows--;
    }
}

- (void)selectAll:(id)sender;
{
    [selectedCells removeAllObjects];
    [selectedCells addObjectsFromArray:[self cells]];
}

- (void)selectCellAtRow:(int)row column:(int)column;
{
    int index;
    id aCell = [self cellAtRow:row column:column];
    
    if (aCell != nil)
    {
	index = [selectedCells indexOfObject:aCell];

	if ((index < 0) || (index > [selectedCells count]))
	{
	    [selectedCells addObject:aCell];
	}
    }
}

- (BOOL)selectCellWithTag:(int)anInt;
{
    id aCell = [self cellWithTag:anInt];
    int index;
    
    if (cell != nil)
    {
	index = [selectedCells indexOfObject:aCell];

	if ((index < 0) || (index > [selectedCells count]))
	{
	    [selectedCells addObject:aCell];
	}
	return YES;
    } else {
	return NO;
    }
}

- (id)selectedCell;
{
    return [selectedCells lastObject];
}

/*
 * returns an array of the selected cells
 */
- (NSArray *)selectedCells;
{
    return selectedCells;
}

- (int)selectedColumn;
{
    int row, col;
    NSMutableArray *aRow, *aCol;
    id aCell;
    
    if ([selectedCells count]) {

	aCell = [selectedCells lastObject];
	
	for (row=0; row<numRows; row++)
	{
	    aRow = [rows objectAtIndex:row];

	    for (col=0; col<numCols; col++)
	    {
		aCol = [aRow objectAtIndex:col];
		      
		if ([aCol indexOfObject:aCell] < numCols)
		{
		    return col;
		}
	    }
	}
    }

    return -1; /* not found */
}

- (int)selectedRow;

{
    int row, col;
    NSMutableArray *aRow, *aCol;
    id aCell;
    
    if ([selectedCells count] > 0)
    {
	aCell = [selectedCells lastObject];
	
	for (row=0; row<numRows; row++)
	{
	    aRow = [rows objectAtIndex:row];

	    for (col=0; col<numCols; col++)
	    {
		aCol = [aRow objectAtIndex:col];
		      
		if ([aCol indexOfObject:aCell] < numCols)
		{
		    return row;
		}
	    }
	}
    }

    return -1; /* not found */
}

- (BOOL)sendAction;
{
    NSObject *cellTarget, *myTarget;
    SEL cellAction, myAction;
    NSCell *aCell = [self selectedCell];
    NSRect cellFrame;
    int aRow = [self selectedRow];
    int aCol = [self selectedColumn];

    if (aCell != nil)
    {
	[aCell setState:1];
	cellFrame = [self cellFrameAtRow:aRow column:aCol];
	[aCell highlight:YES withFrame:cellFrame inView:self];

	if ([aCell isKindOfClass:[NSActionCell class]] == YES)
	{
	    cellTarget = [aCell target];
	    cellAction = [aCell action];

	    if ((cellTarget != nil) 
		&& (cellAction != 0) 
		&& ([cellTarget respondsToSelector:cellAction] == YES))
	    {
		[(NSObject *)cellTarget perform:cellAction];
	    }
	    else
	    {
		myTarget = [(NSControl *)self target];
		myAction = [(NSControl *)self action];

		if ((myTarget != nil)
		    && (myAction != 0)
		    && ([myTarget respondsToSelector:myAction] == YES))
		{
		    [myTarget perform:myAction];
		}
	    }
	}
    }
    return YES; /* ? */
}

- (BOOL)sendDoubleAction;
{
    NSObject *myTarget = [self target];
    SEL myDoubleAction = [self doubleAction];

    if ((myTarget != nil)
	&& (myDoubleAction != 0)
	&& ([(NSObject *)myTarget respondsToSelector:myDoubleAction] == YES))
    {
	[(NSObject *)myTarget perform:myDoubleAction];
    }

    return YES; /* ? */
}

- (void)setAllowsEmptySelection:(BOOL)flag;
{
    allowsEmptySelection = flag;
}

- (void)setAutoscroll:(BOOL)flag;
{
    autoscroll = flag;
}

- (void)setAutosizesCells:(BOOL)flag;
{
    autosize = flag;
}

- (void)setCellClass:(Class)classId;
{
    cellClass = classId;
}

- (void)setCellSize:(NSSize)aSize;
{
    int i;

    for (i=0; i<numCols; i++)
    {
	[(PXKShortInt *)[colWidths objectAtIndex:i] setIntValue:aSize.width];
    }

    for (i=0; i<numRows; i++)
    {
	[(PXKShortInt *)[rowHeights objectAtIndex:i] setIntValue:aSize.height];
    }
}

- (void)setColumnWidth:(int)width at:(int)column;
{
    if ((column > -1) && (column < numCols))
    {
	[(PXKShortInt *)[colWidths objectAtIndex:column] setIntValue:width];
    }
}

- (void)setDoubleAction:(SEL)aSelector;
{
    doubleAction = aSelector;
}

- (void)setDrawsBackground:(BOOL)flag;
{
    drawsBackground = flag;
}

- (void)setDrawsCellBackground:(BOOL)flag;
{
    drawsCellBackground = flag;
}

- (void)setErrorAction:(SEL)aSelector;
{
    errorAction = aSelector;
}

- (void)setIntercellSpacing:(NSSize)aSize;
{
    interCell = aSize;
}

- (void)setMode:(NSMatrixMode)aMode;   // sets the selection mode of the matrix
{
    mode = aMode;
}

- (void)setPrototype:(NSCell *)aCell;
{
    cellPrototype = [aCell retain];
}

- (void)setScrollable:(BOOL)flag;
{
    scrollable = flag;
}

- (void)setState:(int)value atRow:(int)row column:(int)column;
{
    NSMutableArray *aRow;
    NSCell *aCell;

    if ((row>-1)
	&& (row<numRows)
	&& (column>-1)
	&& (column<numCols))
    {
	aRow = (NSMutableArray *)[rows objectAtIndex:row];
	aCell = (NSCell *)[aRow objectAtIndex:column];
	[aCell setState:value];
    }
}

- setColWidth:(int)w at:(int)col;
{
    return self;
}

- setRowHeight:(int)h at:(int)row;
{
    return self;
}

- (void)setSelectionByRect:(BOOL)flag;
{
    selectionByRect = flag;
}

- (void)sizeToCells;
{
    NSSize newSize;
    PXKShortInt *value;
    int row, col;

    newSize.width = 0;
    newSize.height = 0;

    fprintf(stderr, "intercell height: %4.0f\n", interCell.height);
    
    for (row=0; row<numRows; row++)
    {
	value = [rowHeights objectAtIndex:row];

	newSize.height += [value intValue];

	if ((row>0) && (row<(numRows-1)))
	{
	    newSize.height += interCell.height;
	}
    }

    for (col=0; col<numCols; col++)
    {
	value = [colWidths objectAtIndex:col];

	newSize.width += [value intValue];

	if ((col>0) && (col<(numCols-1)))
	{
	    newSize.width += interCell.width;
	}
    }

    dbug(msg(stdout, "Matrix size: %4.0f %4.0f\n",newSize.width,newSize.height));
    
    [(NSView *)self setFrameSize:newSize];
}

@end
