/* 
   NSText.m

   The RTFD text class

   Copyright (C) 1996 Free Software Foundation, Inc.

   Author:  Scott Christley <scottc@net-community.com>
   Date: 1996
   Modified by: ANOQ of the Sun <anoq@vip.cybercity.dk>
   Date: November 1997
   
   This file is part of the NSXKit.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include "NSText.h"
#include "NSApplication.h"
#include "NSWindow.h"
#include "NSFontPanel.h"
#include "PXKDrawing.h"
#include <Foundation/NSCharacterSet.h>
#include <Foundation/NSException.h>
#include "NSTextPrivate.h"

// NSText notifications
NSString *NSTextDidBeginEditingNotification = @"TextDidBeginEditingNotification";
NSString *NSTextDidEndEditingNotification = @"TextDidEndEditingNotification";
NSString *NSTextDidChangeNotification = @"TextDidChangeNotification";

  //Copies the array pc to the array dpc and returns a pointer to the copy
NSTextChunk *NSChunkCopy(NSTextChunk *pc,NSTextChunk *dpc)
{
  NSCParameterAssert(pc!=NULL);
  NSCParameterAssert(dpc!=NULL);
  
  dpc = NSChunkGrow(dpc,pc->used);
  memmove((char *)dpc+sizeof(NSTextChunk),(char *)pc+sizeof(NSTextChunk),pc->used);
  return dpc;
}

  //Increases the array identified by the pointer pc to a size of newUsed bytes
NSTextChunk *NSChunkGrow(NSTextChunk *pc,int newUsed)
{
  return NSChunkZoneGrow(pc,newUsed,NSDefaultMallocZone());
}

  //Allocates initial memory for a structure whose first field is an NSTextChunk structure and
  //subsequent field is a variable-sized array. The amount of memory allocated is equal to 
  //initUsed. If initUsed is 0, growBy bytes are allocated. growBy specifies how much memory
  //should be allocated when the chunk grows
NSTextChunk *NSChunkMalloc(int growBy,int initUsed)
{
  return NSChunkZoneMalloc(growBy,initUsed,NSDefaultMallocZone());
}

  //Increases the amount of memory available for the array identified by the pointer pc, as
  //determined by the array's NSTextChunk.
NSTextChunk *NSChunkRealloc(NSTextChunk *pc)
{
  return NSChunkZoneRealloc(pc,NSDefaultMallocZone());
}

  //Like NSChunkCopy, but uses the specified zone of memory
NSTextChunk *NSChunkZoneCopy(NSTextChunk *pc,NSTextChunk *dpc,NSZone *zone)
{
  return NULL;//FIXME: Not implemented
}

  //Like NSChunkGrow, but uses the specified zone of memory
NSTextChunk *NSChunkZoneGrow(NSTextChunk *pc,int newUsed,NSZone *zone)
{
  int newAlloc;
  
  NSCParameterAssert(pc!=NULL);
  NSCParameterAssert(newUsed>=0);

  if(newUsed>pc->allocated)
  {
    if(newUsed > pc->allocated+pc->growby)//Is there no NSMax(a,b) macro?
      newAlloc = pc->allocated+pc->growby;
    else
      newAlloc = newUsed;
    pc = NSZoneRealloc(zone,pc,sizeof(NSTextChunk)+newAlloc);//FIXME:does this raise an exception automatically?
    pc->allocated = newAlloc;
  }
  pc->used = newUsed;
  return pc;
}

  //Like NSChunkMalloc, but uses the specified zone of memory
NSTextChunk *NSChunkZoneMalloc(int growBy,int initUsed,NSZone *zone)
{
  NSTextChunk *pc;
  int initAlloc;
  
  NSCParameterAssert(growBy>0);
  NSCParameterAssert(initUsed>=0);
  
  if(initUsed==0)
    initAlloc = growBy;
  else
    initAlloc = initUsed;
  pc = NSZoneMalloc(zone,sizeof(NSTextChunk)+initAlloc);//FIXME:does this raise an exception automatically?
  pc->used = initUsed;
  pc->allocated = initAlloc;
  pc->growby = growBy;

  return pc;
}

  //Like NSChunkRealloc, but uses the specified zone of memory
NSTextChunk *NSChunkZoneRealloc(NSTextChunk *pc,NSZone *zone)
{
  NSCParameterAssert(pc!=NULL);

  pc = NSZoneRealloc(zone,pc,sizeof(NSTextChunk)+pc->allocated+pc->growby);//FIXME:does this raise an exception automatically?
  pc->allocated += pc->growby;
  return pc;
}

  //Determines the placement of characters in a line of text. Refers to the NSText object calling
  //the function, and *layInfo is an NSLayInfo struct. The function returns 1 if a word's length
  //exceeds the width of a line and the NSText's charWrap instance varaible is NO. Otherwise, it
  //returns 0.
int NSScanALine(id self,NSLayInfo *layInfo)
{
  //FIXME: Not implemented
  return 0;
}

  //Draws a line of text, using the global variables set by NSScanALine(). The return value has
  //no signifigance.
int NSDrawALine(id self,NSLayInfo *layInfo)
{
  //FIXME: Not implemented
  return 0;
}

//
// NSText implementation
//
@implementation NSText

//
// Class methods
//
+ (void)initialize
{
  if (self == [NSText class])
    {
      // Initial version
      [self setVersion:1];
    }
}

//
// Instance methods
//
//
// Initialization
//

- initWithFrame:(NSRect)frameRect
{
  [super initWithFrame:frameRect];

  //Alloc first text block
  _textStorage = [[NSTextStorage alloc] init];
  
  breakArray = (NSBreakArray *)NSChunkMalloc(sizeof(NSLineDesc)*32,0);
  
  alignment = NSLeftTextAlignment;
  isEditable = YES;
  isRichText = NO;
  isSelectable = YES;
  importsGraphics = NO;
  usesFontPanel = YES;
  isHorizontallyResizable = YES;
  isVerticallyResizable = YES;
  isRulerVisible = NO;
  isFieldEditor = NO;
  textBackgroundColor = [[NSColor whiteColor] retain];
  textColor = [[NSColor blackColor] retain];
  defaultFont = [[NSFont userFontOfSize:12] retain];
  selectedRange.location = 0;
  selectedRange.length = 0;
  maxSize = NSZeroSize;
  minSize = NSZeroSize;
  return self;
}

- (void)dealloc
{
  if(_textStorage)
    [_textStorage release];
  NSChunkFree((NSTextChunk *)breakArray);
  [textBackgroundColor release];
  [textColor release];
  [defaultFont release];
  
  [super dealloc];
}

//
// Getting and Setting Contents 
//
- (void)replaceRange:(NSRange)range
	     withRTF:(NSData *)rtfData
{
  [self notImplemented:_cmd];//FIXME
}

- (void)replaceRange:(NSRange)range
	    withRTFD:(NSData *)rtfdData
{
  [self notImplemented:_cmd];//FIXME
}

- (NSData *)RTFDFromRange:(NSRange)range
{
  [self notImplemented:_cmd];//FIXME
  return nil;
}

- (NSData *)RTFFromRange:(NSRange)range
{
  [self notImplemented:_cmd];//FIXME
  return nil;
}

- (void)setText:(NSString *)string
{
  NSParameterAssert(string!=nil);
  
  [_textStorage
    replaceCharactersInRange:NSMakeRange(0,[_textStorage length])
    withString:string];
}

- (void)setText:(NSString *)string
	  range:(NSRange)range
{
  NSParameterAssert(string!=nil);
  
  [_textStorage
    replaceCharactersInRange:range
    withString:string];
}

- (NSString *)text
{
  return [_textStorage string];
}

//
// Managing Global Characteristics
//
- (NSTextAlignment)alignment
{
  return alignment;
}

- (BOOL)drawsBackground
{
  return drawsBackground;
}

- (BOOL)importsGraphics
{
  return importsGraphics;
}

- (BOOL)isEditable
{
  return isEditable;
}

- (BOOL)isRichText
{
  return isRichText;
}

- (BOOL)isSelectable
{
  return isSelectable;
}

- (BOOL)isFlipped
{
  return YES;//FIXME: is this correct?
}

- (void)setAlignment:(NSTextAlignment)mode
{
  alignment = mode;
}

- (void)setDrawsBackground:(BOOL)flag
{
  drawsBackground = flag;
}

- (void)setEditable:(BOOL)flag
{
  isEditable = flag;
  // If we are editable then we are selectable
  if (flag)
    isSelectable = YES;
}

- (void)setImportsGraphics:(BOOL)flag
{
  importsGraphics = flag;
}

- (void)setRichText:(BOOL)flag
{
  isRichText = flag;
}

- (void)setSelectable:(BOOL)flag
{
  isSelectable = flag;
  // If we are not selectable then we must not be editable
  if (!flag)
    isEditable = NO;
}

//
// Managing Font and Color
//
- (NSColor *)backgroundColor
{
  return textBackgroundColor;
}

- (void)changeFont:(id)sender
{
  [self notImplemented:_cmd];//FIXME
}

- (NSFont *)font
{
  return defaultFont;
}

- (void)setBackgroundColor:(NSColor *)color
{
  [textBackgroundColor release];//FIXME: Is this correct? Should I make a copy?
  textBackgroundColor = [color retain];
}

- (void)setTextColor:(NSColor *)color
	 range:(NSRange)range
{
  if(color == nil)
  {
     [_textStorage removeAttribute:NSForegroundColorAttributeName
        range:range];
  }
  else
  {
    [_textStorage addAttribute:NSForegroundColorAttributeName
       value:color range:range];
  }
}

- (void)setFont:(NSFont *)obj
{
  [defaultFont release];//FIXME: is this correct? should I make a copy?
  defaultFont = [obj retain];
}

- (void)setFont:(NSFont *)font
	range:(NSRange)range
{
  if(font == nil)
  {
     [_textStorage removeAttribute:NSFontAttributeName
        range:range];
  }
  else
  {
    [_textStorage addAttribute:NSFontAttributeName
       value:font range:range];
  }
}

- (void)setTextColor:(NSColor *)color
{
  [textColor release];//FIXME: Is this correct? Should I make a copy?
  textColor = [color retain];
}

- (void)setUsesFontPanel:(BOOL)flag
{
  usesFontPanel = flag;
}

- (NSColor *)textColor
{
  return textColor;
}

- (BOOL)usesFontPanel
{
  return usesFontPanel;
}

//
// Managing the Selection
//
- (NSRange)selectedRange
{
  return selectedRange;
}

- (void)setSelectedRange:(NSRange)range
{
  selectedRange = range;
}

//
// Sizing the Frame Rectangle
//
- (void)setFrame:(NSRect)frameRect
{
  [super setFrame:frameRect];
}

- (BOOL)isHorizontallyResizable
{
  return isHorizontallyResizable;
}

- (BOOL)isVerticallyResizable
{
  return isVerticallyResizable;
}

- (NSSize)maxSize
{
  return maxSize;
}

- (NSSize)minSize
{
  return minSize;
}

- (void)setHorizontallyResizable:(BOOL)flag
{
  isHorizontallyResizable = flag;
}

- (void)setMaxSize:(NSSize)newMaxSize
{
  maxSize = newMaxSize;
}

- (void)setMinSize:(NSSize)newMinSize
{
  minSize = newMinSize;
}

- (void)setVerticallyResizable:(BOOL)flag
{
  isVerticallyResizable = flag;
}

- (void)sizeToFit
{
  [self notImplemented:_cmd];//FIXME
}

//
// Responding to Editing Commands
//
- (void)alignCenter:(id)sender
{
  alignment = NSCenterTextAlignment;
}

- (void)alignLeft:(id)sender
{
  alignment = NSLeftTextAlignment;
}

- (void)alignRight:(id)sender
{
  alignment = NSRightTextAlignment;
}

- (void)copy:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)copyFont:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)copyRuler:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)cut:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
  //[textContents deleteCharactersInRange:selectedRange];
}

- (void)delete:(id)sender
{
  //[textContents deleteCharactersInRange:selectedRange];
  [self notImplemented:_cmd];//FIXME: Send notifications...
}

- (void)paste:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)pasteFont:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)pasteRuler:(id)sender
{
  [self notImplemented:_cmd];//FIXME: NSPasteboard not implemented yet
}

- (void)selectAll:(id)sender
{
  selectedRange.location = 0;
  selectedRange.length = [_textStorage length];
}

- (void)subscript:(id)sender
{
  if([self isRichText])
    [_textStorage subscriptRange:selectedRange];
  else
    [_textStorage subscriptRange:NSMakeRange(0,[_textStorage length])];
}

- (void)superscript:(id)sender
{
  if([self isRichText])
    [_textStorage superscriptRange:selectedRange];
  else
    [_textStorage superscriptRange:NSMakeRange(0,[_textStorage length])];
}

- (void)underline:(id)sender
{
  [self notImplemented:_cmd];//FIXME
}

- (void)unscript:(id)sender
{
  if([self isRichText])
    [_textStorage unscriptRange:selectedRange];
  else
    [_textStorage unscriptRange:NSMakeRange(0,[_textStorage length])];
}

//
// Managing the Ruler
//
- (BOOL)isRulerVisible
{
  return isRulerVisible;
}

- (void)toggleRuler:(id)sender
{
  if(isRulerVisible)
    isRulerVisible = NO;
  else
    isRulerVisible = YES;
}

//
// Spelling
//
- (void)checkSpelling:(id)sender
{
  [self notImplemented:_cmd];//FIXME
}

- (void)showGuessPanel:(id)sender
{
  [self notImplemented:_cmd];//FIXME
}

//
// Scrolling
//
- (void)scrollRangeToVisible:(NSRange)range
{
  [self notImplemented:_cmd];//FIXME
}

//
// Reading and Writing RTFD Files
//
- (BOOL)readRTFDFromFile:(NSString *)path
{
  [self notImplemented:_cmd];//FIXME
  return NO;
}

- (BOOL)writeRTFDToFile:(NSString *)path
	     atomically:(BOOL)flag
{
  [self notImplemented:_cmd];//FIXME
  return NO;
}

//
// Managing the Field Editor
//
- (BOOL)isFieldEditor
{
  return isFieldEditor;
}

- (void)setFieldEditor:(BOOL)flag
{
  isFieldEditor = flag;
}

//
// Handling Events 
//
- (void)mouseDown:(NSEvent *)theEvent
{
  // If not selectable then don't recognize the mouse down
  if (!isSelectable) return;

  [[self window] makeFirstResponder:self];
}

- (void)mouseUp:(NSEvent *)theEvent
{
  // If not selectable then don't recognize the mouse up
  if (!isSelectable) return;
}

- (void)mouseMoved:(NSEvent *)theEvent
{
  // If not selectable then don't recognize the mouse moved
  if (!isSelectable) return;
}

- (void)keyDown:(NSEvent *)theEvent
{
  // If not editable then don't recognize the key down
  if (!isEditable) return;
  [_textStorage replaceCharactersInRange:NSMakeRange(0,0)
     withString:[theEvent characters]];
}

- (void)keyUp:(NSEvent *)theEvent
{
  // If not editable then don't recognize the key up
  if (!isEditable) return;
}

- (BOOL)acceptsFirstResponder
{
  if ([self isSelectable])
    return YES;
  else
    return NO;
}

- (BOOL)becomeFirstResponder
{
  if ([self isEditable])
  {
    return YES;
  }
  else
  {
    return NO;
  }
}

//
// Managing the Delegate
//
- (id)delegate
{
  return delegate;
}

- (void)setDelegate:(id)anObject
{
  delegate = anObject;
}

//
// Implemented by the Delegate
//
- (void)textDidBeginEditing:(NSNotification *)aNotification
{
  //[[NSNotificationCenter defaultCenter] 
  //  postNotificationName: NSTextDidBeginEditingNotification
  //  object: self];
  if ([delegate respondsToSelector:@selector(textDidBeginEditing:)])
    [delegate textDidBeginEditing:nil];
}

- (void)textDidChange:(NSNotification *)aNotification
{
  if ([delegate respondsToSelector:@selector(textDidChange:)])
    [delegate textDidChange:nil];
}

- (void)textDidEndEditing:(NSNotification *)aNotification
{
  if ([delegate respondsToSelector:@selector(textDidEndEditing:)])
    [delegate textDidEndEditing:nil];
}

- (BOOL)textShouldBeginEditing:(NSText *)textObject
{
  if ([delegate respondsToSelector:@selector(textShouldBeginEditing:)])
    return [delegate textShouldBeginEditing:nil];
  else
    return YES;
}

- (BOOL)textShouldEndEditing:(NSText *)textObject
{
  if ([delegate respondsToSelector:@selector(textShouldEndEditing:)])
    return [delegate textShouldEndEditing:nil];
  else
    return YES;
}

- (void)fillTextBuffer
{
  /*char *strPtr;
  int tmpBufferLoc;
  NSString *tmpStr;

  strPtr = textStartPtr + bufferDrawRange.location;
  tmpBufferLoc = bufferDrawRange.location + bufferDrawRange.length;
  
  while(*strPtr!=' ' && *strPtr!='\t' && *strPtr!='\n' && tmpBufferLoc<textDrawLength)
  {
    tmpBufferLoc++;
    strPtr++;
    if(selectedRange.location==tmpBufferLoc)
      break;
  }
  tmpStr = [[textContents
              substringFromRange:
                NSMakeRange(bufferDrawRange.location,
                            tmpBufferLoc-bufferDrawRange.location)]
              autorelease];
  if(*strPtr=='\n' || tmpBufferLoc>=textDrawLength)
  {
    makeLineFeed = YES;
    tmpBufferLoc++;
  }
  else
  {
    //Test if possible...
    stringWidth = pxkWidthOfString(tmpStr,defaultFont);
    if((stringWidth+currentTextPos.x)>drawSizeLimit.width)
    {
      if(possibleBreakLoc>0)
      {
        tmpBufferLoc = possibleBreakLoc;
        strPtr = textStartPtr + tmpBufferLoc;
        tmpStr = [[textContents
                    substringFromRange:
                      NSMakeRange(bufferDrawRange.location,tmpBufferLoc-bufferDrawRange.location)]
                    autorelease];
      }
      makeLineFeed = YES;
    }
    else
      possibleBreakLoc = tmpBufferLoc;
  }
  bufferDrawRange.length = (tmpBufferLoc - bufferDrawRange.location);
  if(makeLineFeed)
    possibleBreakLoc = 0;*/
}

- (void)drawTextBuffer
{
  /*NSString *tmpStr;
  
  tmpStr = [[textContents
              substringFromRange:bufferDrawRange]
              autorelease];
  pxkDrawString([tmpStr cString],currentTextPos);
  if(makeLineFeed)
  {
    currentTextPos.y += lineFeedHeight;
    currentTextPos.x = 0;
    makeLineFeed = NO;
  }
  else
    currentTextPos.x += stringWidth;
  bufferDrawRange.location += bufferDrawRange.length;
  bufferDrawRange.length = 0;*/
}

- (void)drawText
{
  int cnt,maxCnt;
  int textPos;
  NSLineDesc lineDesc;
  NSString *substring;
  int lineFeedHeight,pointSize;
  NSRange effectiveRange,clippedRange;
  NSDictionary *attrs;
  NSPoint drawPos;
  NSFont *font,*defFont;
  BOOL flipped;

  [self lockFocus];
  flipped = [self isFlipped];

  pxkCalcTextStorageLineBreaks(_textStorage,breakArray,[self frame].size);
  maxCnt = breakArray->chunk.used/sizeof(NSLineDesc);
  textPos = 0;
  defFont = (NSFont *)[NSFont fontWithName:@"Helvetica" size:12];
  for(cnt=0;cnt<maxCnt;cnt++)
  {
    //Get line length
    lineDesc = breakArray->breaks[cnt];
    
    //Calc linefeed height
    lineFeedHeight = 0;
    effectiveRange = NSMakeRange(textPos,0);
    while(NSMaxRange(effectiveRange) < textPos+lineDesc)
    {
      attrs = [_textStorage attributesAtIndex:NSMaxRange(effectiveRange)
                 effectiveRange:&effectiveRange];
      font = (NSFont *)[attrs objectForKey:NSFontAttributeName];
      if(font==nil)
        font = defFont;
      pointSize = [font pointSize];
      if(lineFeedHeight < pointSize)
        lineFeedHeight = pointSize;
    }
    //Make line feed
    drawPos.y += lineFeedHeight;
    drawPos.x = 0;
    //Draw line
    effectiveRange = NSMakeRange(textPos,0);
    while(NSMaxRange(effectiveRange) < textPos+lineDesc)
    {
      attrs = [_textStorage attributesAtIndex:NSMaxRange(effectiveRange)
                 effectiveRange:&effectiveRange];
      font = (NSFont *)[attrs objectForKey:NSFontAttributeName];
      if(font==nil)
        font = defFont;
      clippedRange = NSIntersectionRange(effectiveRange,NSMakeRange(textPos,lineDesc));
      substring = [[_textStorage string] substringFromRange:
                    clippedRange];
      [substring drawAtPoint:drawPos withAttributes:attrs];
      drawPos.x += [font widthOfString:substring];
    }
    textPos += lineDesc;
  }
  [self unlockFocus];

  /*
  - (void)__drawAtPoint:(NSPoint)point inRange:(NSRange)range
{
  NSFont *font;
  NSPoint truePos = point;
  NSRange effRange;
  NSString *text,*txtPiece;
  int strBegin,strEnd,tmpBufferLoc,attrBufferLoc;
  int textPieceEnd,textEdge,fontHeight,lineFeedHeight;
  NSDictionary *attr;
  char *totalStrPtr,*strPtr;
  BOOL flipped;
  BOOL doBreak = NO;
  NSView *fView;
  NSColor *defColor;
  NSFont  *defFont;
  
  //This method shall NOT take line-sweep-direction into account, and it doesn't
  fView = [NSView focusView];
  if(!fView)
    [NSException raise:NSGenericException format:
      @"No focued view in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];

  textEdge = point.x;

  strBegin = range.location;
  effRange = NSMakeRange(strBegin,0);
  text = [self string];
  strEnd = NSMaxRange(range);
  if(strEnd>[text length])
  {
    [NSException raise:NSGenericException format:
      @"Range too long in class NSAttributedString (NSAttributedStringAdditions) in method - __drawAtPoint:inRange:"];
  }
  totalStrPtr = (char *)[text cString];
  flipped = [fView isFlipped];
  if(!flipped)
  {
    truePos.y = [fView frame].size.height - point.y;
  }

  defColor = [NSColor blackColor];
  defFont = [NSFont fontWithName:@"Helvetica" size:12];

  tmpBufferLoc = strBegin;
  strPtr = totalStrPtr + strBegin;

  while(tmpBufferLoc<strEnd)
  {
    attrBufferLoc = tmpBufferLoc;
    if(NSMaxRange(effRange) <= attrBufferLoc)
    {
      attr = [self attributesAtIndex:NSMaxRange(effRange) effectiveRange:&effRange];
      _txt_setAttributes(attr,&font,defFont,defColor);
      fontHeight = [font pointSize];
    }
    lineFeedHeight = fontHeight;
    while(*strPtr!='\n' && tmpBufferLoc<strEnd)
    {
      tmpBufferLoc++;
      strPtr++;
    }
    while(effRange.location < tmpBufferLoc)
    {
      if(NSMaxRange(effRange) <= attrBufferLoc)
      {
        attr = [self attributesAtIndex:NSMaxRange(effRange) effectiveRange:&effRange];
        _txt_setAttributes(attr,&font,defFont,defColor);
        fontHeight = [font pointSize];
        if(lineFeedHeight<fontHeight)
          lineFeedHeight = fontHeight;
      }

      if(NSMaxRange(effRange)>=tmpBufferLoc)
      {
        textPieceEnd = tmpBufferLoc;
        doBreak = YES;
      }
      else
        textPieceEnd = NSMaxRange(effRange);

      txtPiece =
        [text substringFromRange:
        NSMakeRange(attrBufferLoc,textPieceEnd-attrBufferLoc)];
      _txt_drawString([txtPiece cString],truePos,flipped,YES,fontHeight);
      truePos.x += [font widthOfString:txtPiece];
      attrBufferLoc = textPieceEnd;
      if(doBreak)
      {
        doBreak = NO;
        break;
      }
    }
    if(tmpBufferLoc < strEnd)
    {
      //Forced line-feed
      if(flipped)
        truePos.y += lineFeedHeight;
      else
        truePos.y -= lineFeedHeight;
      truePos.x = textEdge;
      tmpBufferLoc++;//Skip '/n'
      strPtr++;//Skip '/n'
      strBegin = tmpBufferLoc;
    }
  }
}

  */
}

//
// Displaying
//
- (void)display
{
  NSRect rect = [self frame];
  rect.origin.x = 0;
  rect.origin.y = 0;
  
  [super display];
  
  [(NSView *)self lockFocus];
  NS_DURING
    if(drawsBackground)
    {
      [textBackgroundColor set];
      NSRectFill(rect);
    }
  
    [self drawText];
  NS_HANDLER
    [(NSView *)self unlockFocus];
    [localException raise];
  NS_ENDHANDLER
}

- (void)drawRect:(NSRect)rect
{
  [super drawRect:rect];//FIXME: what should this method do?
}

//
// NSCoding protocol
//
- (void)encodeWithCoder:aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObjectReference: delegate withName: @"Delegate"];
  //FIXME: Doesn't encode text contents
  [aCoder encodeValueOfObjCType: "I" at: &alignment];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &drawsBackground];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isEditable];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isRichText];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isSelectable];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &importsGraphics];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &usesFontPanel];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isHorizontallyResizable];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isVerticallyResizable];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isRulerVisible];
  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &isFieldEditor];
  [aCoder encodeObject: backgroundColor];
  [aCoder encodeObject: textColor];
  [aCoder encodeObject: defaultFont];
  //[aCoder encodeValueOfObjCType: @encode(NSRange) at: &selectedRange];
    //FIXME: Should the selected range be stored?
  [aCoder encodeValueOfObjCType: @encode(NSSize) at: &maxSize];
  [aCoder encodeValueOfObjCType: @encode(NSSize) at: &minSize];
}

- initWithCoder:aDecoder
{
  [super initWithCoder:aDecoder];

  [aDecoder decodeObjectAt: &delegate withName: NULL];
  //FIXME: Doesn't decode text contents
  [aDecoder decodeValueOfObjCType: "I" at: &alignment];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &drawsBackground];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isEditable];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isRichText];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isSelectable];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &importsGraphics];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &usesFontPanel];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) 
	    at: &isHorizontallyResizable];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isVerticallyResizable];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isRulerVisible];
  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &isFieldEditor];
  backgroundColor = [[aDecoder decodeObject] retain];
  textColor = [[aDecoder decodeObject] retain];
  defaultFont = [[aDecoder decodeObject] retain];
  //[aDecoder decodeValueOfObjCType: @encode(NSRange) at: &selectedRange];
    //FIXME: Should the selected range be stored?
  selectedRange = NSMakeRange(0,0);
  [aDecoder decodeValueOfObjCType: @encode(NSSize) at: &maxSize];
  [aDecoder decodeValueOfObjCType: @encode(NSSize) at: &minSize];

  return self;
}

@end
/*
Copyright 1996 by NeXT Software, Inc.  All Rights Reserved.

Inherits From:	NSView : NSResponder : NSObject

Conforms To:	NSChangeSpelling

NSIgnoreMisspelledWords
NSCoding (NSResponder)
NSObject (NSObject)


Declared In:	AppKit/NSText.h


Purpose

NSText declares the most general programmatic interface for objects that manage text. 
You usually use one of its subclasses, NSTextView or NSCStringText, but both share 
the methods and other definitions of this class.


Principal Attributes

	Supports rich text and graphics		Provides delegation and notification

	Works with the Font Panel and menu		Works with the pasteboard

	Works with the Services facility		Works with the spell-checking service


Creation

See the class description.


Commonly Used Methods

- readRTFDFromFile:	Reads an .rtf or .rtfd file.
- writeRTFDToFile:atomically:	Writes the receivers text to a file.
- string	Returns the receivers text, without attributes.
- RTFFromRange:	Returns the receivers text with attributes.
- RTFDFromRange:	Returns the receivers text with attributes and attachments.


Class Description


The NSText class declares the most general programmatic interface to objects that 
manage text. NSText is an abstract class with two concrete subclasses, NSCStringText 
and NSTextView. You can create an NSText object when constructing your applications 
interface with Interface Builder, or at run time using initWithFrame:. For the most 
generic and portable use of text, this is the recommended approach. NSCStringText is 
most suitable for use when backward compatibility with the NEXTSTEP Text object is 
needed. NSTextView is a NeXT addition to the OpenStep specification that acts as the 
front end to NeXTs revised text system. Instances of any of these classes are 
generically called text objects.

Text objects are used by the Application Kit wherever text appears in interface 
objects: A text object draws the title of a window, the commands in a menu, the 
title of a button, and the items in a browser. Your application can also create 
text objects for its own purposes.

The text classes are unlike most other classes in the Application Kit in the 
richness and complexity of their interface. One of their design goals is to provide a 
comprehensive set of text-handling features so that youll rarely need to create a 
subclass. Among other things, a text object can:

	Control whether the user can select or edit text.

	Control the font and layout characteristics of its text by working with the 
Font Panel and menu.

	Let the user control the format of paragraphs by manipulating a ruler.

	Control the color of its text and background.

	Wrap text on a word or character basis.

	Display graphic images within its text.

	Write text to or read text from files in the form of RTFDRich Text Format 
files that contain TIFF or EPS images, or attached files.

	Let another object, the delegate, dynamically control its properties.

	Let the user copy and paste text within and between applications.

	Let the user copy and paste font and format information between NSText objects.

	Let the user check the spelling of words in its text.


Graphical user-interface building tools (such as Interface Builder) may give you 
access to text objects in several different configurations, such as those found 
in the NSTextField, NSForm, and NSScrollView objects. These classes configure a 
text object for their own specific purposes. Additionally, all NSTextFields, 
NSForms, NSButtons within the same windowin short, all objects that access a text 
object through associated cellsshare the same text object, reducing the memory 
demands of an application. Thus, its generally best to use one of these classes 
whenever it meets your needs, rather than create text objects yourself. If one of 
these classes doesnt provide enough flexibility for your purposes, you can create 
text objects programmatically.

Text objects typically work closely with various other objects. Some of thesesuch 
as the delegate or an embedded graphic objectrequire a degree of programming on 
your part. Otherssuch as the Font Panel, spell checker, or rulertake no effort 
other than deciding whether the service should be enabled or disabled. Several of 
the following sections discuss these interrelationships.



Plain and Rich Text Objects



Text objects are differentiated into two groups: those that allow only one set of 
text attributes for all of their text, and those that allow multiple fonts, sizes, 
indents, and other attributes for different sets of characters and paragraphs. 
Text objects in the former group are called plain text objects, while those in the 
latter are called rich text objects. You can control whether a text object is plain 
or rich using the setRichText: method. Rich text objects are also capable of 
allowing the user to drag images and files into them. This behavior is controlled by 
the setImportsGraphics: method.

A rich NSText object can use RTF (Rich Text Format) as an interchange format. 
Not all RTF control words are supported: On input, an NSText object ignores 
any control word it doesnt recognize; some of those it can read and interpret it 
doesnt write out. The table below lists the RTF control words that any text 
object recognizes. Subclasses may recognize more. 


Control Word	Read	Write

\ansi	yes	yes
\b	yes	yes
\cb	yes	yes
\cf	yes	yes
\colortbl	yes	yes
\dnn	yes	yes
\fin	yes	yes
\fn	yes	yes
\fonttbl	yes	yes
\fsn	yes	yes
\i	yes	yes
\lin	yes	yes
\margrn	yes	yes
\paperwn	yes	yes
\mac	yes	no
\margln	yes	yes
\par	yes	yes
\pard	yes	no
\pca	yes	no
\qc	yes	yes
\ql	yes	yes
\qr	yes	yes
\sn	yes	no
\tab	yes	yes
\upn	yes	yes



Notifying a Text Objects Delegate



Many of an NSText objects actions can be controlled through an associated object, 
the NSText objects delegate. The delegate can be any object you choose, and one 
delegate can control multiple NSText objects. If it implements any of the following 
methods, the delegate receives the corresponding message at the appropriate time:


textShouldBeginEditing:
textDidBeginEditing:
textDidChange:
textShouldEndEditing:
textDidEndEditing:


Of special note are the two textShould methods. These methods are requests for 
permission. Any time a text object begins an operation that would change its text or 
attributes, it uses textShouldBeginEditing: to request approval for the change. 
The delegate can return YES to permit the change, or NO to forbid it. Similarly, 
textShouldEndEditing: enables the delegate to prevent a text object from ending 
editing, such as when it contains an invalid value.


Adding Graphics and Other Attachments to the Text



A rich text object may allow graphics or other file attachments to be embedded in the 
text. Each graphic is treated as a single (possibly large) character: The texts 
line height and character placement are adjusted to accommodate the graphic. Graphics 
are embedded in the text in either of two ways: programmatically or directly through 
user actions. In the programmatic approach, graphic objects can be added using 
replaceRange:WithRTFD:, or through a more specific method defined by a subclass.

An alternate means of adding an image or other attachment to the text is for the 
user to drag an image or other file directly into the text object. The text object 
automatically creates an attachment object to manage the display of the image 
(the implementation of attachment differs between NSTextView and NSCStringText). 
This feature requires a rich text object that has been configured to receive 
dragged images using the setImportsGraphics: method. 

Images that have been imported can be written as part of an RTFD document. RTFD 
documents use a file package, or directory, to store the components of the document 
(the D stands for directory). The file package has the name of the document plus 
an .rtfd extension. The file package always contains a file called TXT.rtf for the 
text of the document, and one or more TIFF or EPS files for the images, plus the 
files for other attachments. A text object can transfer information in an RTFD 
document to a file and read it from a file using the writeRTFDToFile:atomically: and 
readRTFDFromFile: methods.

Working with the Font Panel



Text objects are designed to work with the Application Kits font conversion system, 
defined by the NSFontPanel and NSFontManager classes. By default, a text object 
keeps the Font Panel updated with the first font in its selection, or of its typing 
attributes (defined below). It also changes the font in response to messages from 
the Font Panel and Font menu. Such changes apply to the selected text or typing 
attributes for a rich text object, or to all the text in a plain text object. You 
can turn this behavior off using the setUsesFontPanel: method. Doing so is 
recommended for a text object that serves as a field editor, for example.

Working with Rulers and Paragraph Styles



Text objects also provide for a ruler, by which the user can edit paragraph attributes 
such as indents and tabs. NSCStringText uses its own ruler object, and defines some 
methods for altering paragraph attributes. NSTextView works with the public 
NSRulerView class and uses the NSTextStorage and NSParagraphStyle classes to handle 
paragraph attributes. To show or hide a text objects ruler, use the toggleRuler: 
action method. Similar to the Font Panel, NSTextView can be set not to use a ruler 
with the setUsesRuler: method.

Adopted Protocols


NSChangeSpelling	- changeSpelling:

NSIgnoreMisspelledWords	- ignoreSpelling:


Method Types


Creating instances	- initWithFrame:

Getting the characters	- string

Setting graphic attributes	- setBackgroundColor:

- backgroundColor
- setDrawsBackground:
- drawsBackground


Setting behavioral attributes	- setEditable:

- isEditable
- setSelectable:
- isSelectable
- setFieldEditor:
- isFieldEditor
- setRichText:
- isRichText
- setImportsGraphics:
- importsGraphics


Using the Font Panel and menu	- setUsesFontPanel:

- usesFontPanel


Using the ruler	- toggleRuler:

- isRulerVisible


Changing the selection	- setSelectedRange:

- selectedRange


Replacing text	- replaceCharactersInRange:withRTF:

- replaceCharactersInRange:withRTFD:
- replaceCharactersInRange:withString:
- setString:


Action methods for editing	- selectAll:

- copy:
- cut:
- paste:
- copyFont:
- pasteFont:
- copyRuler:
- pasteRuler:
- delete:


Changing the font	- changeFont:

- setFont:
- setFont:range:
- font


Setting text alignment	- setAlignment:

- alignCenter:
- alignLeft:
- alignRight:
- alignment


Setting text color	- setTextColor:

- setTextColor:range:
- textColor


Setting super- and subscripting	- superscript:

- subscript:
- unscript:


Underlining text	- underline:

Reading and writing RTF	- readRTFDFromFile:

- writeRTFDToFile:atomically:
- RTFDFromRange:
- RTFFromRange:


Checking spelling	- checkSpelling:

- showGuessPanel:


Constraining size	- setMaxSize:

- maxSize
- setMinSize:
- minSize
- setVerticallyResizable:
- isVerticallyResizable
- setHorizontallyResizable:
- isHorizontallyResizable
- sizeToFit


Scrolling	- scrollRangeToVisible:

Setting the delegate	- setDelegate:

- delegate


Instance Methods


alignCenter:


- (void)alignCenter:(id)sender


This action method applies center alignment to selected paragraphs (or all text if 
the receiver is a plain text object).


See also:	- alignLeft:, - alignRight:, - alignment, - setAlignment:



alignLeft:


- (void)alignLeft:(id)sender


This action method applies left alignment to selected paragraphs (or all text if the 
receiver is a plain text object).


See also:	- alignCenter:, - alignRight:, - alignment, - setAlignment:



alignRight:


- (void)alignRight:(id)sender


This action method applies right alignment to selected paragraphs (or all text if 
the receiver is a plain text object).


See also:	- alignLeft:, - alignCenter:, - alignment, - setAlignment:



alignment


- (NSTextAlignment)alignment


Returns the alignment of the first selected paragraph, or of all text for a plain text 
object. This value is one of:


NSLeftTextAlignment
NSRightTextAlignment
NSCenterTextAlignment
NSJustifiedTextAlignment
NSNaturalTextAlignment (realized as one of the above depending on the script)



backgroundColor


- (NSColor *)backgroundColor


Returns the receivers background color.


See also:	- drawsBackground, - setBackgroundColor:



changeFont:


- (void)changeFont:(id)sender


This action method changes the font of the selection for a rich text object, or of 
all text for a plain text object. If the receiver doesnt use the Font Panel, 
however, this method does nothing.

This method changes the font by sending convertFont: messages to sender (which is 
presumed to be an NSFontManager or similarly capable object) and applying each 
NSFont returned to the appropriate text. If a rich text objects selection contains 
multiple fonts, convertFont: is invoked for each contiguous range of characters 
that share a font. See the NSFontManager class specification for more information 
on font conversion.

See also:	- usesFontPanel



checkSpelling:


- (void)checkSpelling:(id)sender


This action method searches for a misspelled word in the receivers text. The search 
starts at the end of the selection and continues until it reaches a word suspected of 
being misspelled or the end of the text. If a word isnt recognized by the spelling 
server. A showGuessPanel: message then opens the Guess panel and allows the user 
to make a correction or add the word to the local dictionary.


See also:	- showGuessPanel:



copy:


- (void)copy:(id)sender


This action method copies the selected text onto the general pasteboard, in as many 
formats as the receiver supports. A plain text object uses NSStringPboardType for 
plain text, and a rich text object also uses NSRTFPboardType.


See also:	- copyFont:, - copyRuler:, - cut:, - paste:



copyFont:


- (void)copyFont:(id)sender


This action method copies the font information for the first character of the 
selection (or for the insertion point) onto the font pasteboard, as NSFontPboardType.


See also:	- copy:, - copyRuler:, - cut:, - paste:



copyRuler:


- (void)copyRuler:(id)sender


This action method copies the paragraph style information for first selected paragraph 
onto the ruler pasteboard, as NSRulerPboardType, and expands the selection to 
paragraph boundaries.


See also:	- copy:, - copyFont:, - cut:, - paste:



cut:


- (void)cut:(id)sender


This action method deletes the selected text and places it onto the general 
pasteboard, in as many formats as the receiver supports. A plain text object uses 
NSStringPboardType for plain text, and a rich text object also uses NSRTFPboardType.


See also:	- delete:, - copy:, - copyFont:, - copyRuler:, - paste:



delegate


- (id)delegate


Returns the receivers delegate, or nil if it has none.


See also:	- setDelegate:



delete:


- (void)delete:(id)sender


This action method deletes the selected text.


See also:	- cut:



drawsBackground


- (BOOL)drawsBackground


Returns YES if the receiver draws its background, NO if it doesnt.


See also:	- backgroundColor, - setDrawsBackground:



font


- (NSFont *)font


Returns the font of the first character in the receivers text, or of the insertion 
point if theres no text.


See also:	- setFont:, - setFont:range:



importsGraphics


- (BOOL)importsGraphics


Returns YES if the receiver allows the user to import files by dragging, NO if it 
doesnt. A text object that accepts dragged files is also a rich text object.


See also:	- isRichText, - setImportsGraphics:



initWithFrame:


- (id)initWithFrame:(NSRect)frameRect


Initializes the receiver with frameRect as its frame rectangle. This method 
actually substitutes an instance of a concrete subclass of NSText, such as 
NSCStringText or NSTextView, depending on the platform, and configures that 
instance to archive itself in a manner portable across OpenStep implementations.



isEditable


- (BOOL)isEditable


Returns YES if the receiver allows the user to edit text, NO if it doesnt. You 
can change the receivers text programmatically regardless of this setting.

If the receiver is editable, its also selectable.


See also:	- isSelectable, - setEditable:



isFieldEditor


- (BOOL)isFieldEditor


Returns YES if the receiver interprets Tab, Shift-Tab, and Return (Enter) as cues 
to end editing, and possibly to change the first responder; no if it accepts them 
as text input. See the NSWindow class specification for more information on field 
editors. By default, NSText objects dont behave as field editors.


See also:	- setFieldEditor:



isHorizontallyResizable


- (BOOL)isHorizontallyResizable


Returns YES if the receiver automatically changes its width to accommodate the 
width of its text, NO if it doesnt.


See also:	- isVerticallyResizable, - setHorizontallyResizable



isRichText


- (BOOL)isRichText


Returns YES if the receiver allows the user to apply attributes to specific ranges 
of the text, NO if it doesnt.


See also:	- importsGraphics, - setRichText:



isRulerVisible


- (BOOL)isRulerVisible


Returns YES if the receivers enclosing scroll view shows its ruler, NO otherwise.


See also:	- usesRuler, - toggleRuler:



isSelectable


- (BOOL)isSelectable


Returns YES if the receiver allows the user to select text, NO if it doesnt.


See also:	- isEditable, - setSelectable:



isVerticallyResizable


- (BOOL)isVerticallyResizable


Returns YES if the receiver automatically changes its height to accommodate the 
height of its text, NO if it doesnt.


See also:	- isHorizontallyResizable, - setVerticallyResizable



maxSize


- (NSSize)maxSize


Returns the receivers maximum size.


See also:	- minSize, - setMaxSize:



minSize


- (NSSize)minSize


Returns the receivers minimum size.


See also:	- maxSize, - setMinSize:



paste:


- (void)paste:(id)sender


This action method pastes text from the general pasteboard at the insertion point 
or over the selection.


See also:	- copy:, - cut:, - pasteFont:, - pasteRuler:



pasteFont:


- (void)pasteFont:(id)sender


This action method pastes font information from the font pasteboard onto the 
selected text or insertion point of a rich text object, or over all text 
of a plain text object.


See also:	- copyFont:, - pasteRuler:



pasteRuler:


- (void)pasteRuler:(id)sender


This action method pastes paragraph style information from the ruler pasteboard onto 
the selected paragraphs of a rich text object. It doesnt apply to a plain text 
object.


See also:	- copyFont:, - pasteRuler:



readRTFDFromFile:


- (BOOL)readRTFDFromFile:(NSString *)path


Attempts to read the RTFD file at path, returning YES if successful and NO if not. 
path should be the path for an .rtf file or an .rtfd file wrapper, not for 
the RTF file within an .rtfd file wrapper. 


See also:	- writeRTFDToFile:atomically:



replaceCharactersInRange:withRTF:


- (void)replaceCharactersInRange:(NSRange)aRange withRTF:(NSData *)rtfData


Replaces the characters in aRange with RTF text interpreted from rtfData. This 
method applies only to rich text objects.


See also:	- replaceCharactersInRange:withRTFD:, - replaceCharactersInRange:withString:



replaceCharactersInRange:withRTFD:


- (void)replaceCharactersInRange:(NSRange)aRange withRTFD:(NSData *)rtfdData


Replaces the characters in aRange with RTFD text interpreted from rtfdData. This 
method applies only to rich text objects.


See also:	- replaceCharactersInRange:withRTF:, 
- replaceCharactersInRange:withString:



replaceCharactersInRange:withString:


- (void)replaceCharactersInRange:(NSRange)aRange withString:(NSString *)aString


Replaces the characters in aRange with aString. For a rich text object, the text of 
aString is assigned the formatting attributes of the first character of the text 
it replaces, or of the character immediately before aRange if the ranges length 
is zero. If the ranges location is zero, the formatting attributes of the first 
character in the receiver are used.


See also:	- replaceCharactersInRange:withRTF:, 
- replaceCharactersInRange:withRTFD:



RTFDFromRange:


- (NSData *)RTFDFromRange:(NSRange)aRange


Returns an NSData object that contains an RTFD stream corresponding to the 
characters and attributes within aRange. Raises an NSRangeException if any part of 
aRange lies beyond the end of the receivers characters.

When writing data to the pasteboard, you can use the NSData object as the first 
argument to NSPasteboards setData:forType: method, with a second argument of 
NSRTFDPboardType.


See also:	- RTFFromRange:



RTFFromRange:


- (NSData *)RTFFromRange:(NSRange)range


Returns an NSData object that contains an RTF stream corresponding to the characters 
and attributes within aRange, omitting any attachment characters and attributes. 
Raises an NSRangeException if any part of aRange lies beyond the end of the receivers 
characters.

When writing data to the pasteboard, you can use the NSData object as the first 
argument to NSPasteboards setData:forType: method, with a second argument of 
NSRTFPboardType.


See also:	- RTFDFromRange:documentAttributes:



scrollRangeToVisible:


- (void)scrollRangeToVisible:(NSRange)aRange


Scrolls the receiver in its enclosing scroll view so that the first characters of 
aRange are visible.



selectAll:


- (void)selectAll:(id)sender


This action method selects all of the receivers text.



selectedRange


- (NSRange)selectedRange


Returns the range of selected characters.


See also:	- setSelectedRange:



setAlignment:


- (void)setAlignment:(NSTextAlignment)mode


Sets the alignment of all the receivers text to mode, which may be one of:


NSLeftTextAlignment
NSRightTextAlignment
NSCenterTextAlignment
NSJustifiedTextAlignment
NSNaturalTextAlignment (realized as one of the above depending on the script)


See also:	- alignment, - alignLeft:, - alignCenter:, - alignRight:



setBackgroundColor:


- (void)setBackgroundColor:(NSColor *)aColor


Sets the receivers background color to aColor.


See also:	- setDrawsBackground:, - backgroundColor



setDelegate:


- (void)setDelegate:(id)anObject


Sets the receivers delegate to anObject, without retaining it.


See also:	- delegate



setDrawsBackground:


- (void)setDrawsBackground:(BOOL)flag


Controls whether the receiver draws its background. If flag is YES, the receiver 
fills its background with the background color; if flag is NO, it doesnt.


See also:	- setBackgroundColor:, - drawsBackground



setEditable:


- (void)setEditable:(BOOL)flag


Controls whether the receiver allows the user to edit its text. If flag is YES, 
the receiver allows the user to edit text and attributes; if flag is NO, it 
doesnt. You can change the receivers text programmatically regardless 
of this setting. If the receiver is made editable, its also made selectable. 
NSText objects are by default editable.


See also:	- setSelectable:, - isEditable



setFieldEditor:


- (void)setFieldEditor:(BOOL)flag


Controls whether the receiver interprets Tab, Shift-Tab, and Return (Enter) as cues 
to end editing, and possibly to change the first responder. If flag is YES, it does; 
if flag is NO, it doesnt, instead accepting these characters as text input. See 
the NSWindow class specification for more information on field editors. By default, 
NSText objects dont behave as field editors.


See also:	- isFieldEditor



setFont:


- (void)setFont:(NSFont *)aFont


Sets the font of all the receivers text to aFont.


See also:	- setFont:range:, - font



setFont:range:


- (void)setFont:(NSFont *)aFont range:(NSRange)aRange


Sets the font of characters within aRange to aFont. This method applies only to a 
rich text object.


See also:	- setFont:, - font



setHorizontallyResizable:


- (void)setHorizontallyResizable:(BOOL)flag


Controls whether the receiver changes its width to fit the width of its text. If 
flag is YES it does; if flag is NO it doesnt.


See also:	- setVerticallyResizable:, - isHorizontallyResizable



setImportsGraphics:


- (void)setImportsGraphics:(BOOL)flag


Controls whether the receiver allows the user to import files by dragging. If flag 
is YES, it does; if flag is NO, it doesnt. If the receiver is set to accept 
dragged files, its also made a rich text object. Subclasses may or may not accept 
dragged files by default.


See also:	- setRichText:, - importsGraphics



setMaxSize:


- (void)setMaxSize:(NSSize)aSize


Sets the receivers maximum size to aSize.


See also:	- setMinSize:, - maxSize



setMinSize:


- (void)setMinSize:(NSSize)aSize


Sets the receivers minimum size to aSize.


See also:	- setMaxSize:, - minSize



setRichText:


- (void)setRichText:(BOOL)flag


Controls whether the receiver allows the user to apply attributes to specific ranges
of the text. If flag is YES it does; if flag is NO it doesnt. If flag is NO, the 
receiver is also set not to accept dragged files. Subclasses may or may not let the 
user apply multiple attributes to the text and accept drag files by default.


See also:	- isRichText, - setImportsGraphics:



setSelectable:


- (void)setSelectable:(BOOL)flag


Controls whether the receiver allows the user to select its text. If flag is YES, the 
receiver allows the user to select text; if flag is NO, it doesnt. You can set 
selections programmatically regardless of this setting. If the receiver is made not 
selectable, its also made not editable. NSText objects are by default editable 
and selectable.


See also:	- setEditable:, - isSelectable



setSelectedRange:


- (void)setSelectedRange:(NSRange)aRange


Selects the receivers characters within aRange.


See also:	- selectedRange



setString:


- (void)setString:(NSString *)aString


Replaces the receivers entire text with aString, applying the formatting attributes 
of the old first character to its new contents.



setTextColor:


- (void)setTextColor:(NSColor *)color


Sets the text color of all characters in the receiver to aColor. Removes the text 
color attribute if aColor is nil.


See also:	- setTextColor:range - textColor:



setTextColor:range:


- (void)setTextColor:(NSColor *)aColor range:(NSRange)aRange


Sets the text color of characters within aRange to aColor. Removes the text color 
attribute if aColor is nil. This method applies only to rich text objects.


See also:	- setTextColor: - textColor:



setUsesFontPanel:


- (void)setUsesFontPanel:(BOOL)flag


Controls whether the receiver uses the Font Panel and Font menu. If flag is YES, 
the receiver responds to messages from the Font Panel and from the Font menu, and 
updates the Font Panel with the selection font whenever it changes. If flag is NO 
the receiver doesnt do any of this. By default, an NSText object uses the Font 
Panel and menu.


See also:	- usesFontPanel



setVerticallyResizable:


- (void)setVerticallyResizable:(BOOL)flag


Controls whether the receiver changes its height to fit the height of its text. If 
flag is YES it does; if flag is NO it doesnt.


See also:	- setHorizontallyResizable:, - isVerticallyResizable



showGuessPanel:


- (void)showGuessPanel:(id)sender


This action method opens the Spelling panel, allowing the user to make a correction 
during spell checking.


See also:	- checkSpelling:



sizeToFit


- (void)sizeToFit


Resizes the receiver to fit its text.


See also:	- isHorizontallyResizable, - isVerticallyResizable



string


- (NSString *)string


Returns the characters of the receivers text.


See also:	- setString:



subscript:


- (void)subscript:(id)sender


This action method applies a subscript attribute to selected text (or all text if the 
receiver is a plain text object), lowering its baseline offset by a predefined amount.


See also:	- subscript:, - unscript:, - lowerBaseline: (NSTextView)



superscript:


- (void)superscript:(id)sender


This action method applies a superscript attribute to selected text (or all text if 
the receiver is a plain text object), raising its baseline offset by a predefined 
amount.


See also:	- subscript:, - unscript:, - raiseBaseline: (NSTextView)



textColor


- (NSColor *)textColor


Returns the color of the receivers first character, or for the insertion point if 
theres no text.


See also:	- setTextColor:, - setTextColor:range:



toggleRuler:


- (void)toggleRuler:(id)sender


This action method shows or hides the ruler, if the receiver is enclosed in a 
scroll view.



underline:


- (void)underline:(id)sender


This action method underlines selected text for a rich text object, or all text 
for a plain text object.



unscript:


- (void)unscript:(id)sender


This action method removes any superscripting or subscripting from selected text 
(or all text if the receiver is a plain text object).


See also:	- subscript:, - superscript:, - raiseBaseline: (NSTextView), 
- lowerBaseline: (NSTextView)



usesFontPanel


- (BOOL)usesFontPanel


Returns YES if the receiver uses the Font Panel, NO otherwise.


See also:	- setUsesFontPanel:



writeRTFDToFile:atomically:


- (BOOL)writeRTFDToFile:(NSString *)path atomically:(BOOL)flag


Writes the receivers text as RTF with attachments to a file or directory at path. 
Returns YES on success and NO on failure. If atomicFlag is YES, attempts to write 
the file safely so that an existing file at path is not overwritten, nor does 
a new file at path
 actually get created, unless the write is successful.


See also:	- RTFFromRange:, - RTFDFromRange:, - readRTFDFromFile:



Methods Implemented By the Delegate


textDidBeginEditing:


- (void)textDidBeginEditing:(NSNotification *)aNotification


Informs the delegate that the text object has begun editing (that it has become 
first responder). The name of aNotification is NSTextViewDidBeingEditingNotification.



textDidChange:


- (void)textDidChange:(NSNotification *)aNotification


Informs the delegate that the text object has changed its characters or formatting 
attributes. The name of aNotification is NSTextViewDidChangeNotification.



textDidEndEditing:


- (void)textDidEndEditing:(NSNotification *)aNotification


Informs the delegate that the text object has finished editing (that it has resigned 
first responder status). The name of aNotification is 
NSTextViewDidEndEditingNotification.



textShouldBeginEditing:


- (BOOL)textShouldBeginEditing:(NSText *)aTextObject


Invoked from a text objects implementation of becomeFirstResponder, this method 
requests permission for aTextObject to begin editing. If the delegate returns YES, 
the text object proceeds to make changes. If the delegate returns NO, the text 
object abandons the editing operation. This method is invoked whenever aTextObject 
attempts to become first responder.


See also:	- makeFirstResponder: (NSWindow), - becomeFirstResponder (NSResponder)



textShouldEndEditing:


- (BOOL)textShouldEndEditing:(NSText *)aTextObject


Invoked from a text objects implementation of resignFirstResponder, this method 
requests permission for aTextObject to end editing. If the delegate returns YES, 
the text object proceeds to finish editing and resign first responder status. 
If the delegate returns NO, the text object selects all of its text and remains 
the first responder.


See also:	- resignFirstResponder (NSResponder)



Notifications


NSTextDidBeginEditingNotification



Posted when an NSText object begins any operation that changes characters or 
formatting attributes.

The notification contains:


Notification Object	The notifying NSText object.

Userinfo	None



NSTextDidChangeNotification



Posted after an NSText object performs any operation that changes characters or 
formatting attributes.

The notification contains:


Notification Object	The notifying NSText object.

Userinfo	None



NSTextDidEndEditingNotification



Posted when an NSText object resigns first responder status.

The notification contains:


Notification Object	The notifying NSText object.

Userinfo	None


Key	Value


NSTextMovement	The keyboard operation that ended editing, an NSNumber (int).


The value for NSTextMovement is one of:


NSIllegalTextMovement
NSReturnTextMovement
NSTabTextMovement
NSBacktabTextMovement
NSLeftTextMovement
NSRightTextMovement
NSUpTextMovement
NSDownTextMovement


The value for NSTextMovement is typically one of the first four.
NSIllegalTextMovement indicates that the text object ended editing because of 
something other than keyboard input, typically a mouse click in another view. 
NSReturnTextMovement, NSTabTextMovement, and NSBacktabTextMovement indicate 
that a field editor ended editing because the user pressed the key named in 
the value. The remaining four values indicate the same for arrow keys.

*/