/* 
   NSTextAttachment.m

   Implementation of textattachment-class

   Copyright (C) 1997 Free Software Foundation, Inc.

   Written by: ANOQ of the Sun <anoq@vip.cybercity.dk>
   Date: November 1997
   
   This file is part of NSXKit

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "NSTextAttachment.h"

@implementation NSTextAttachmentCell

//Cell size and position 
- (NSSize)cellSize;
{
  /*Returns the size of the attachment's icon.
  See also: - icon (NSFileWrapper), - fileWrapper (NSTextAttachment)*/
  return NSMakeSize(0,0);
}

- (NSPoint)cellBaselineOffset;
{
  /*Returns the position where the attachment cell's image should be drawn in text, relative to the current point 
  established in the glyph layout. The image should be drawn so that its lower left corner lies on this point.
  See also: - icon (NSFileWrapper)*/
  return NSMakePoint(0,0);
}

//Drawing 
-(void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)aView;
{
  /*Draws the receiver's image within cellFrame in aView, which should be the focus view.
  See also: - drawWithFrame:inView: (NSCell), - lockFocus (NSView)*/
}

-(void)highlight:(BOOL)flag withFrame:(NSRect)cellFrame inView:(NSView *)aView;
{
  /*Draws the receiver's image with highlighting if flag is YES within cellFrame in aView, which should be the focus 
  view.
  See also: - highlight:withFrame:inView: (NSCell), - lockFocus (NSView)*/
}

//Setting the attachment 
-(void)setAttachment:(NSTextAttachment *)anAttachment;
{
  _textAttachment = anAttachment;
  /*Sets the text attachment object that owns the receiver to anAttachment, without retaining it (the text attachment, 
  as the owner, retains the cell).
  See also: - attachment, - setAttachmentCell: (NSTextAttachment)*/
}

- (NSTextAttachment *)attachment;
{
  return _textAttachment;
}

//Event handling
- (BOOL)trackMouse:(NSEvent *)theEvent inRect:(NSRect)cellFrame ofView:(NSView *)aTextView untilMouseUp:(BOOL)flag;
{
  /*Handles a mouse-down event on the receiver's image. theEvent is the mouse-down event. cellFrame is the region of
  aTextView in which further mouse events should be tracked. aTextView is the view which received the event.
  It's assumed to be an NSTextView, and should be the focus view. If flag is YES, the receiver tracks the mouse until 
  a mouse-up event occurs; if flag is NO, it stops tracking when a mouse-dragged event occurs outside of cellFrame. 
  Returns YES if the receiver successfully finished tracking the mouse (typically through a mouse-up event), 
  NO otherwise (such as when the mouse is dragged outside cellFrame).
  NSTextAttachmentCell's implementation of this method calls upon aTextView's delegate to handle the event. If 
  theEvent is a mouse-up event for a double click, the text attachment cell sends the delegate a 
  textView:doubleClickedOnCell:inRect: message and returns YES. Otherwise, depending on whether the user clicks 
  or drags the cell, it sends the delegate a textView:clickedOnCell:inRect: or a textView:draggingCell:inRect:event: 
  message and returns YES. NSTextAttachmentCell's implementation returns NO only if flag is NO and the mouse is 
  dragged outside of cellFrame. The delegate methods are invoked only if the delegate responds.
  See also: - wantsToTrackMouse, - trackMouse:inRect:ofView:untilMouseUp: (NSCell),- lockFocus (NSView)*/

  /*Handles a mouse-down event on the receiver's image. NSTextAttachmentCell's implementation of this method calls 
  upon aTextView's delegate to handle the event. If theEvent concludes as a double click, the text attachment cell 
  sends the delegate a textView:doubleClickedOnCell:inRect: message and returns
  YES. Otherwise, depending on whether the user clicks or drags the cell, it sends the delegate a 
  textView:clickedOnCell:inRect: or a textView:draggingCell:inRect:event: message and returns YES. 
  NSTextAttachmentCell's implementation returns NO only if flag is NO and the mouse is dragged outside of 
  cellFrame. The delegate methods are invoked only if the delegate responds.
  See also: - wantsToTrackMouse, - trackMouse:inRect:ofView:untilMouseUp: (NSCell), - lockFocus (NSView)*/
  return NO;
}

- (BOOL)wantsToTrackMouse;
{
  /*Returns YES if the receiver will handle a mouse event occurring over its image (to support dragging, for example), 
  NO otherwise. NSTextAttachmentCell's implementation of this method returns YES. The NSView containing the cell 
  should invoke this method before sending a trackMouse:inRect:ofView:untilMouseUp: message.
  For an attachment in an attributed string, if the attachment cell returns NO its attachment character should be 
  selected rather than the cell being asked to track the mouse. this results in the attachment icon behaving as any 
  regular glyph in text.*/
  
  /*Returns YES. NSTextAttachmentCell objects support dragging. An NSTextView invokes this method before sending 
  trackMouse:inRect:ofView:untilMouseUp: to the text attachment cell.
  A more static subclass might override this method to return NO, which results in the attachment image behaving 
  as any other glyph in the text, and not allow itself to be dragged or to perform a method on being clicked.*/
  return NO;
}

@end //implementation NSTextAttachmentCell

@implementation NSTextAttachment

//Creating an NSTextAttachment
-(id)initWithFileWrapper:(NSFileWrapper *)aWrapper
{
  attachmentCell = [[NSTextAttachmentCell alloc] init];
  fileWrapper = [aWrapper retain];

  /*Initializes a newly allocated NSTextAttachment to contain aWrapper and to use an NSTextAttachmentCell as its 
  attachment cell. If aWrapper contains an image file that the receiver can interpret as an NSImage object, it 
  sets the attachment cell's image to the NSImage rather than to aWrapper's icon.
  This method is the designated initializer for the NSTextAttachment class. Returns self.
  See also: - setFileWrapper:, - setAttachmentCell:*/
  return self;
}

//Setting the file wrapper
- (void)setFileWrapper:(NSFileWrapper *)aWrapper
{
  if(fileWrapper)
    [fileWrapper release];
  fileWrapper = aWrapper;
  if(fileWrapper)
    [fileWrapper retain];
}

- (NSFileWrapper *)fileWrapper
{
  return fileWrapper;
}

//Setting the attachment cell
-(void)setAttachmentCell:(id <NSTextAttachmentCell>)aCell
{
  if(attachmentCell)
    [attachmentCell release];
  attachmentCell = aCell;
  if(attachmentCell)
    [attachmentCell retain];
  /*Sets the object used to draw the icon for the attachment and to handle mouse events to aCell.
  See also: - setFileWrapper:, - setImage: (NSCell), - icon (NSFileWrapper), - attachmentCell*/
}

- (id <NSTextAttachmentCell>)attachmentCell
{
  /*Returns the object used to draw the icon for the attachment and to handle mouse events. An NSTextAttachment by 
  default uses an NSTextAttachmentCell that displays the attached file's icon, or its contents if the file 
  contains an image.
  See also: - fileWrapper, - image (NSCell), - icon (NSFileWrapper), - setAttachmentCell:*/
  return attachmentCell;
}

- (void)dealloc
{
  if(fileWrapper)
    [fileWrapper release];
  if(attachmentCell)
    [attachmentCell release];
  [super dealloc];
}

@end //NSTextAttachment