/*
 * Filename: NSView.m
 * Project:  Pascal's X kit
 * Module:   Basic application components
 *
 * Function: This is the NSView class
 *
 * Author:   Pascal Forget
 * Date:     January 1996
 *
 * Copyright (C) 1995 Pascal Forget.  All Rights Reserved.
 */

#include "NSView.h"
#include "PXKDrawing.h"
#include "PXKGeometry.h"
#include "misc.h"

static NSView *focusView;

@implementation NSView

+ (NSView *)focusView;
{
    return focusView;
}

- (void)addSubview:(NSView *)aView;
{
    [subviews addObject:aView];
    [aView setSuperview:self];
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent;
{
    return YES; /* Default behavior.  Is this the proper behavior? */
}

- (NSView *)ancestorSharedWithView:(NSView *)aView;
{
    if (([aView isDescendantOf:self]) || (aView == self)) {
	return self;
    } else if (aView == superview) {
	return aView;
    } else if (superview != nil) {
	return [superview ancestorSharedWithView:aView];
    } else {
	return nil;
    }
}

- (BOOL)autoresizesSubviews;
{
    return autoresizesSubviews;
}

- (BOOL)canDraw;
{
    return YES; /* Subclasses may override */
}

- (NSPoint)convertPoint:(NSPoint)aPoint fromView:(NSView *)aView;
{
    NSPoint p = aPoint;

    /* Not implemented yet */

    return p;
}

/*
 * Incomplete implementation used only in hitTest: method.
 */
- (NSPoint)convertPoint:(NSPoint)aPoint toView:(NSView *)aView;
{
    NSPoint p = aPoint;
    NSRect subviewFrame = [aView frame];
    int i, count;
    NSView *aSubview;

    if ([aView isDescendantOf:self] == NO)
    {
	/* I don't know what to do, I give up.  (Sorry!) */
	fprintf(stderr, "NSView in convertPoint:toView: aView is not one "
		"of my descendants.  I don't know what to do.  Goodbye!\n");
	exit(-1);
    }

    /*
     * if aView is one of my subviews, then we can easily
     * convert to its coordinates system.  If not, then
     * if aView is a descendant, then I can reduce the
     * problem to the previous case by converting the
     * coordinates to my subview witch also has aView
     * as its descendant, and recursively until I reach
     * aView's superview.
     */

    if ([subviews containsObject:aView] == YES)
    {
	p.x -= subviewFrame.origin.x;
	p.y -= subviewFrame.origin.y;
    }
    else
    {
	count = [subviews count];

	for (i=0; i<count; i++)
	{
	    aSubview = (NSView *)[subviews objectAtIndex:i];

	    if ([aView isDescendantOf:aSubview] == YES)
	    {
		p = [self convertPoint:p toView:aSubview];
		return [aSubview convertPoint:p toView:aView];
	    }
	}
    }

    return p;
}

- (void)display;
{
    if (viewNum < 0) {

	dbug(msg("NSView display, creating the XWindow\n"));
	
	if (superview != nil) {
	
	    /*
	     * make sure the superview has already it's own backing
	     */
	    if ([superview viewNum] < 0) {
		[superview display];
		return;
	    } else {
		dbug(msg("NSView display, creating the XWindow using superview\n"));
		viewNum = createView(self, [superview viewNum]);
	    }
	} else {
	    dbug(msg("NSView display, creating the XWindow using window\n"));
	    viewNum = createView(self, [window windowNumber]);
	}

	if (viewNum <0) {
	    dbug(msg("NSView display: error creating the view's "
		    "windowing system-specific implementation.\n")); 
	    /* Should raise exception here */
	} else {
	    dbug(msg("NSView display set view frame.\n"));
	    setViewFrame(viewNum, frame);
	}	    
    } else {
	dbug("View display\n");
    }
    
    displayView(viewNum);
    
    /*
     * If this is a window's frame view (the view has no superview)
     * then paint the background the same color as the window's
     */
    if (superview == nil) {
	drawViewBackground(viewNum);
    }
    
    [subviews makeObjectsPerform:@selector(display)];
}

- (void)displayIfNeeded;
{
    if ((superview != nil) && (needsDisplay == YES)) {
	
	if (viewNum < 0) {
	    viewNum = createView(self, [superview viewNum]);

	    if (viewNum <0) {
		dbug(msg("NSView display: error creating the view's "
			"windowing system-specific implementation.\n")); 
		/* Should raise exception here */
	    }
	}

	displayView(viewNum);

	if (isOpaque == YES) {
	    [subviews makeObjectsPerform:@selector(display)];
	}
    }
}

- (void)displayIfNeededIgnoringOpacity;
{
    if ((superview != nil) && (needsDisplay == YES)) {
	
	if (viewNum < 0) {
	    viewNum = createView(self, [superview viewNum]);

	    if (viewNum <0) {
		fprintf(stderr, "NSView display: error creating the view's "
			"windowing system-specific implementation.\n"); 
		/* Should raise exception here */
	    }
	}

	displayView(viewNum);
	[subviews makeObjectsPerform:@selector(display)];
    }
}

- (void)displayRect:(NSRect)aRect;
{
    [self display]; /* Dumb implementation to be replaced later */
}

- (void)displayRectIgnoringOpacity:(NSRect)aRect;
{
    [self display]; /* Dumb implementation to be replaced later */
}

- (void)drawRect:(NSRect)rect;
{
    /*
     * NSView's implementation is used for faking transparency
     * and for painting every window's frame view background
     */
    drawViewBackground(viewNum);
    
    /* Implemented by subclasses to supply drawing instructions */
}

- (void)fax:(id)sender;
{
    /* Not implemented yet. */
}

- (float)frameRotation;
{
    return frameRotation; /* This will be hard to implement w/o DPS */
}

- (NSRect)frame;
{
    return frame;
}

- (NSView *)hitTest:(NSPoint)aPoint;
{
    int count = [subviews count] - 1;
    NSView *aView;
    NSView *hit;
    NSPoint p;
    NSRect subviewFrame;

    /*
     * Let's see if the point is enclosed in my frame
     */
    
    if ((aPoint.x >= frame.origin.x) &&
	(aPoint.x <= frame.origin.x + frame.size.width) &&
	(aPoint.y >= frame.origin.y) &&
	(aPoint.y <= frame.origin.y + frame.size.height))
    {
	 // See if the point is inside one of my subviews
    
	while (count > -1) {

	    aView = (NSView *)[subviews objectAtIndex:count];

	    // Convert the point to the subview's coordinates

	    p = [self convertPoint: aPoint fromView: [self superview]];

	    hit = [aView hitTest: p];
	    
	    if (hit != nil) {
		return hit;
	    }
	    count--;
	}

	/* 
	 * The point is not inside one of my subviews... but it is inside
	 * my frame rect, therefore a hit is detected for myself
	 */
	return self;
    }
    
    return nil;  // No hit was detected
}

- (id)init;
{
    [super init];
    viewNum = -1;
    superview = nil;
    backgroundColor = [NSColor whiteColor];
    frameRotation = 0.0;
    subviews = [[NSMutableArray alloc] init];
    window = nil;
    tag = 0;
    cursorRects = [[NSMutableArray alloc] init]; /* Should be a hash table. */
    isOpaque = YES;
    needsDisplay = YES;
    return self;
}

- (id)initWithFrame:(NSRect)frameRect;
{
    [self init];

    frame = frameRect;
#if 0
    frame.origin.x = 0.0;
    frame.origin.y = 0.0;
    frame.size.width = 100.0;
    frame.size.height = 100.0;
#endif
    return self;
}

- (BOOL)isDescendantOf:(NSView *)aView;
{
    if (superview != nil) {
	if (superview == aView) {
	    return YES;
	} else {
	    return [superview isDescendantOf: aView];
	}
    } else {
	return NO;
    }
}

- (BOOL)isFlipped; // returns NO; subclasses override as needed.
{
    return NO;
}

- (BOOL)isOpaque;
{
    return isOpaque;
}

- (BOOL)isRotatedFromBase;
{
    return NO; // This library does not support frame rotation (yet?)
}

- (void)lockFocus;
{
    focusView = self;
}

- (void)mouseDown:(NSEvent *)theEvent;  /* subclasses should override */
{
    int i, count = [subviews count];
    NSView *aSubview;
    NSView *hit = nil;

    dbug(msg("%s: mouseDown: at: %i %i\n",
	    [[self description] cString],
	    (int)[theEvent locationInWindow].x,
	    (int)[theEvent locationInWindow].y));
    
    for (i=0; i<count; i++)
    {
	aSubview = [subviews objectAtIndex: i];
	
	hit = [aSubview hitTest: [theEvent locationInWindow]];

	if (hit != nil)
	{
	    if (([self isKeyWindow] == YES) || 
		([hit acceptsFirstMouse: theEvent] == YES))
	    {
		[hit mouseDown: theEvent];
		return;
	    }
	}
    }

    // At this point, we know that no subview was a hit;
    // We call NSResponder's implementation of the method.
    [super mouseDown: theEvent];
}

- (void)mouseUp:(NSEvent *)theEvent;  /* subclasses should override */
{
    int i, count = [subviews count];
    NSView *aSubview;
    NSView *hit = nil;

    msg("%s: mouseUp: at: %i %i\n",
	    [[self description] cString],
	    (int)[theEvent locationInWindow].x,
	    (int)[theEvent locationInWindow].y);
    
    for (i=0; i<count; i++)
    {
	aSubview = [subviews objectAtIndex: i];
	
	hit = [aSubview hitTest: [theEvent locationInWindow]];

	if (hit != nil)
	{
	    if (([self isKeyWindow] == YES) || 
		([hit acceptsFirstMouse: theEvent] == YES))
	    {
		[hit mouseUp: theEvent];
		return;
	    }
	}
    }

    // At this point, we know that no subview was a hit;
    // We call NSResponder's implementation of the method.
    [super mouseUp: theEvent];
}

- (BOOL)needsDisplay;
{
    return needsDisplay;
}

- (NSView *)opaqueAncestor;
{
    if (superview != nil) {
	if ([superview isOpaque]) {
	    return superview;
	} else {
	    return [superview opaqueAncestor];
	}
    }

    return nil;
}   

- (BOOL)postsFrameChangedNotifications;
{
    return notifiesAncestors;
}

- (void)print:(id)sender;
{
    /* Not implemented yet. */
}

/*
 * removeFromSuperview removes the receiver from the view hierarchy.
 * My (Pascal) interpretation of the previous line is that the view
 * will not only be removed from the view hierarchy, but it will
 * effectively disappear from the screen.
 */

- (void)removeFromSuperview;
{
    if (superview != nil) {
	[(NSMutableArray *)[superview subviews] removeObject:self];
    }

    /*
     * Destroy the X Window that provided
     *  the on-screen representation for this view
     */
    if (viewNum > -1) {       
	destroyView(viewNum);
	viewNum = -1;
	window = nil;
    }
}

- (void)replaceSubview:(NSView *)oldView with:(NSView *)newView;
{
    unsigned int index = [subviews indexOfObject:oldView];

    if (index != NSNotFound) {
	if (newView != nil) {
	    [subviews replaceObjectAtIndex:index withObject:newView];
	    [newView setSuperview:self];
	} else {
	    [subviews removeObject:oldView];
	}
	[oldView setSuperview:nil];
    }
}
	
- (void)rotateByAngle:(float)angle;
{
    frameRotation += angle;
    setViewFrameRotation(viewNum, frameRotation);
}

- (void)setAutoresizesSubviews:(BOOL)flag;
{
    autoresizesSubviews = flag;
}

- (void)setFrame:(NSRect)frameRect;
{
    NSSize oldSize = frame.size;
    
    frame = frameRect;
    setViewFrame(viewNum, frame);
    [self resizeSubviewsOldSize:oldSize newSize:frame.size];
}

- (void)setFrameOrigin:(NSPoint)newOrigin;
{
#if 1
    frame.origin.x = newOrigin.x;
    frame.origin.y = newOrigin.y;

    setViewFrameOrigin(viewNum, newOrigin);

    if ((notifiesAncestors == YES) && (superview != nil))
    {
	/* Notify superview.  (OpenStep does not specify selector -Duh.) */
    }
#endif    
}

- (void)setFrameRotation:(float)angle;
{
    if (frameRotation != angle) {
	frameRotation = angle;
	setViewFrameRotation(viewNum, frameRotation);
	
	if ((notifiesAncestors == YES) && (superview != nil)) {
	    /* Notify superview.  (OpenStep does not specify selector -Duh.) */
	}
    }
}

- (void)setFrameSize:(NSSize)newSize;
{
    NSSize oldSize = frame.size;
#if 1    
    frame.size = newSize;
    
    setViewFrameSize(viewNum, newSize);

    if ((notifiesAncestors == YES) && (superview != nil)) {
	/* Notify superview.  (OpenStep does not specify selector -Duh.) */
    }

    [self resizeSubviewsOldSize:oldSize newSize:newSize];
#endif    
}

- (void)setNeedsDisplay:(BOOL)flag;
{
    needsDisplay = flag;
}

/*
 * setNeedsDisplayInRect: has an incomplete implementation.
 * It will be made OpenStep compliant in a future release.
 * For the moment being, when a view needs to be redisplayed,
 * it's the whole view that will be redisplayed.  Redrawing
 * only a portion of a view will speed up drawing Matrices
 * in a Scroll view where only a small portion of the view
 * is actually visible.
 */
- (void)setNeedsDisplayInRect:(NSRect)invalidRect; /* dumb implementation */
{
    needsDisplay = YES;
}

- (void)setPostsFrameChangedNotifications:(BOOL)flag;
{
    notifiesAncestors = flag;
}

/*
 * The setTag: method does not appear in the OpenStep specification.
 * I sure hope that this is a mistake, as it really ought to be in there...
 */

- (void)setTag:(int)anInt; 
{
    tag = anInt;
}

- (void)sortSubviewsUsingFunction:(int (*)(id,id,void *))compare
			  context:(void *)context;
{
    /* Not implemented yet */
}

- (NSArray *)subviews;
{
    return subviews;
}

- (NSView *)superview;
{
    return superview;
}

- (void)superviewSizeChanged:(NSSize)oldSize;
{

}

- (int)tag;
{
    return tag;
}

- (void)unlockFocus;
{
    if (focusView == self) {
	focusView = nil;
    }
}

- (void)viewWillMoveToWindow:(NSWindow *)newWindow;
{
    int i, count;
    
    window = newWindow;

    count = [subviews count];

    for (i=0; i<count; i++) {
	/*
	 * The following line has the effect
	 * of changing the 'window' attribute
	 * in every subview
	 */
	[(NSView *)[subviews objectAtIndex:i] setSuperview:self];
    }
}

- (id)viewWithTag:(int)aTag;
{
    unsigned int i, count = [subviews count];
    NSView *aSubview;
    
    for (i=0; i<count; i++) {

	aSubview = [subviews objectAtIndex:i];

	if ([aSubview tag] == aTag) {
	    return aSubview;
	}
    }
    return nil; /* Not found */
}

- (NSWindow *)window;
{
    return window;
}

@end

@implementation NSView (NonGNUStep)

- (NSColor *)backgroundColor;
{
    return backgroundColor;
}

- (void)resizeSubviewsOldSize:(NSSize)oldSize newSize:(NSSize)newSize;
{
    if ((autoresizesSubviews == YES) &&
	((oldSize.width != newSize.width) ||
	 (oldSize.height != newSize.height)))
    {
	[self resizeSubviewsOldSize:oldSize];
    }
}

- (void)resizeSubviewsOldSize:(NSSize)oldSize;
{
    NSView *aSubview;
    int index = [subviews count];
	
    while (index > 0) {
	aSubview = [subviews objectAtIndex:index];
	[aSubview superviewSizeChanged:oldSize];
	index--;
    }
}

- (void)setOpaque:(BOOL)flag;
{
    isOpaque = flag;
}

- (void)setSuperview:(NSView *)aView;
{
    superview = aView;
    
    if (aView != nil) {
	window = [superview window];
    } else {
	window = nil;
	/*
	 * Destroy the X Window that provided
	 *  the on-screen representation for this view
	 */
	if (viewNum > -1) {       
	    destroyView(viewNum);
	    viewNum = -1;
	}
    }
}

- (int)viewNum;
{
    return viewNum;
}


- (NSRect)bounds;// FIXME: Dummy implementation
{
  NSRect tmpRect = {{0,0},{0,0}};
  return tmpRect;
}


@end
