/*
 * Filename: NSWindow.m
 * Project:  Pascal's OpenStep Application Kit for X Windows
 * Module:   Basic application components
 *
 * Function: This is the NSWindow class
 *
 * Author:   Pascal Forget <pascal@wsc.com>
 * Date:     February 1997
 *
 * Copyright (C) 1997 Paascal Forget
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "NSWindow.h"
#include "PXKList.h"
#include "NSView.h"
#include "PXKGeometry.h"

#define NS_DEFAULT_WINDOW_TITLE "Window"

@implementation NSWindow

- (NSColor *)backgroundColor;
{
    return backgroundColor;
}

- (NSBackingStoreType)backingType;
{
    return backingType;
}

- (void)becomeKeyWindow;
{
    if ([self canBecomeKeyWindow])
    {
	[[self sharedApplication] setKeyWindow:self];	
	// Should send notification here
    }
}

- (void)becomeMainWindow;
{
    if ([self canBecomeMainWindow])
    {
	[[self sharedApplication] setMainWindow:self];	
	// Should send notification here
    }
}

- (BOOL)canBecomeKeyWindow;
{
    return YES; // subclasses override as needed
}

- (BOOL)canBecomeMainWindow;
{
    return YES; // subclasses override as needed
}

- (void)center; /* Incomplete implementation */
{
    if (windowNum < 0) {
	windowNum = createWindow(self);
    }
    
    if (windowNum > -1) {
	/* centerWindow(windowNum); */
    }
}

- (void)close;
{
    if (windowNum > -1 ) {
#if 0
	if ((delegate == nil) ||
	    (((BOOL)[delegate respondsToSelector:
			      @selector(windowWillClose:)]) &&
	     
	     ((BOOL)[(id)delegate windowWillClose:self])))
	{
	    closeWindow(windowNum);
	}
#else
	closeWindow(windowNum);
#endif
    }
    isVisible = NO;
}

- (id)contentView;
{
    return contentView;
}

- (id)delegate;
{
    return delegate;
}

- (void)disableFlushWindow;
{
    isFlushWindowDisabled = YES;
    /* send notification to the drawing engine */
}

- (void)enableFlushWindow;
{
    isFlushWindowDisabled = NO;
    /* send notification to the drawing engine */
}

- (NSRect)frame;
{
    return frame;
}

- (id)init;
{
    title = [[NSString alloc] initWithCString:NS_DEFAULT_WINDOW_TITLE];
    isDocumentEdited = NO;
    isFlushWindowDisabled = NO;
    delegate = nil;
    backgroundColor = [NSColor lightGrayColor];
    return self;
}

- (id)initWithContentRect:(NSRect)contentRect
		styleMask:(unsigned int)aStyle
		  backing:(NSBackingStoreType)bufferingType
		    defer:(BOOL)flag;
{
    [self init];

    styleMask = aStyle;
    backingType = bufferingType;
    deferred = flag;
    _contentRect = contentRect;
    
    if (deferred == NO) {
	windowNum = createWindow(self);
	frameView = [[NSView alloc] initWithFrame:contentRect];
	contentView = [[NSView alloc] initWithFrame:contentRect];
	[frameView addSubview:contentView];

#ifdef TRANSPARENCY
	[contentView setOpaque:YES];
#endif
	[frameView viewWillMoveToWindow:self];
    } else {
	windowNum = -1;
    }
    return self;
}

- (void)dealloc;
{
    if (backgroundColor != nil) {
	[backgroundColor release];
    }

    if (representedFilename != nil) {
	[representedFilename release];
    }

    if (title != nil) {
	[title release];
    }

    if (windowNum >-1 ) {
	destroyWindow(windowNum);
    }

    [super dealloc];
}

- (void)display;
{
    if (isFlushWindowDisabled == NO) {
	displayWindow(windowNum);
	[frameView display]; // frameView displays the content view
	isVisible = YES;
    }
}

- (BOOL)isDocumentEdited;
{
    return isDocumentEdited;
}
	
- (BOOL)isFlushWindowDisabled;
{
    return isFlushWindowDisabled;
}

- (BOOL)isKeyWindow; /* Stub implementation - always returns YES */
{
    return YES;
}

- (BOOL)isVisible;
{
    return isVisible;
}

- (BOOL)makeFirstResponder:(NSResponder *)aResponder
{
  return NO;//FIXME: Dummy
}

- (void)mouseDown:(NSEvent *)theEvent;
{
    NSView *hit = [contentView hitTest:[theEvent locationInWindow]];
    
    if (hit == nil) {
    	[self noResponderFor:@selector(mouseDown:)];
	return;
    } else {
	if (([self isKeyWindow] == YES) || 
	    ([hit acceptsFirstMouse:theEvent] == YES))
	{
	    [hit mouseDown:theEvent];
	}
    }
}

- (void)mouseUp:(NSEvent *)theEvent;
{
    NSView *hit = [contentView hitTest: [theEvent locationInWindow]];
    
    if (hit == nil) {
    	[self noResponderFor: @selector(mouseUp:)];
	return;
    } else {
	fprintf(stderr, "Got hit\n");
	if (([self isKeyWindow] == YES) || 
	    ([hit acceptsFirstMouse: theEvent] == YES))
	{
	    [hit mouseUp: theEvent];
	}
    }
}

- (NSString *)representedFilename;
{
    return representedFilename;
}

- (void)rightMouseDown:(NSEvent *)theEvent;
{
    NSView *hit = [contentView hitTest:[theEvent locationInWindow]];
    
    if (hit == nil) {
    	[self noResponderFor:@selector(rightMouseDown:)];
	return;
    } else {
	if (([self isKeyWindow] == YES) || 
	    ([hit acceptsFirstMouse:theEvent] == YES))
	{
	    [hit rightMouseDown:theEvent];
	}
    }    
}

- (void)setBackgroundColor:(NSColor *)color;
{
    if (color != backgroundColor) {
	if (backgroundColor != nil) {
	    [backgroundColor release];
	}

	backgroundColor = color;

	if (color != nil) {
	    [color retain];
	}
    }
}

- (void)setBackingType:(NSBackingStoreType)type;
{
    backingType = type; /* Incomplete implementation */
}

- (void)setContentSize:(NSSize)aSize;
{

#if 0    
    _contentRect.size = aSize;
    
    if (contentView == nil) {
	contentView = [[NSView alloc] initWithFrame:_contentRect];
    }

    [contentView setFrameSize:aSize];
#endif    
}

- (void)setContentView:(NSView *)aView;
{
    if (aView != contentView) {
	if (contentView != nil) {
	    [contentView release];
	}

	contentView = aView;
        [contentView viewWillMoveToWindow:self];	
    
	if (aView != nil) {
	    [aView retain];
	}
    }
}

- (void)setDelegate:(id)anObject;
{
    if (delegate != anObject) {

	if (delegate != nil) {
	    [delegate release];
	}

	delegate = anObject;

	if (anObject != nil) {
	    [anObject retain];
	}
    }
}

- (void)setDocumentEdited:(BOOL)flag;
{
    isDocumentEdited = flag;
}

- (void)setFrame:(NSRect)frameRect display:(BOOL)flag;
{
    NSSize aSize;
    frame = frameRect;
    
    setWindowFrame(windowNum, frame);

    aSize = frameRect.size;
    
    aSize.height -= 21.0; /* Title bar is always there in this release */

    [self setContentSize:aSize];
    
    if (flag == YES) {
	[self display];
    }
}

- (void)setFrameOrigin:(NSPoint)aPoint;
{
    frame.origin = aPoint;
    setWindowFrameOrigin(windowNum, aPoint);
}

- (void)setFrameTopLeftPoint:(NSPoint)aPoint;
{
    frame.origin.x = aPoint.x;
    frame.origin.y = aPoint.y - frame.size.height;
    setWindowFrameOrigin(windowNum, aPoint);
}

- (void)setRepresentedFilename:(NSString *)aString;
{
    if (aString != representedFilename) {
	if (representedFilename != nil) {
	    [representedFilename release];
	}

	representedFilename = aString;

	if (aString != nil) {
	    [aString retain];
	}
    }
}

- (void)setTitle:(NSString *)aString;
{
    if (aString != title) {
	if (title != nil) {
	    [title release];
	}

	title = aString;

	if (title != nil) {
	    [title retain];
	}

	if (windowNum >- 1) {
	    setWindowTitle(windowNum, title);
	}
    }
}

- (void)setTitleWithRepresentedFilename:(NSString *)aString;
{
    [self setRepresentedFilename:aString];
    [self setTitle:representedFilename];
}

- (unsigned int)styleMask;
{
    return styleMask;
}

- (NSString *)title;
{
    return title;
}

- (int)windowNumber;
{
    return windowNum;
}

@end
