/*
 * Filename: PXKDrawing.m
 * Project:  Pascal's AppKit for X Windows
 * Module:   Drawing Engine
 *
 * Function: Basic drawing functions of the kit
 *
 * Author:   Pascal Forget <pascal@wsc.com>
 * Date:     January 1997
 *
 * Copyright (C) 1997 Pascal Forget.  All Rights Reserved.
 */

#include "PXKDrawing.h"
#include "PXKGeometry.h"
#include "NSFont.h"
#include <Foundation/NSException.h>

#define PXKDE_DEBUG 0
#define PXK_MAX_COLOR 65535.0

//FIXME:
//------
//In the NSBitmapImageRep class I allocate 32 bits per pixel for the
//internal storage format when the display is running 32bits per
//pixel. I do it to avoid an image-conversion from 24bit to 32bit
//when sending the image to the X-server. I don't know if this
//violates the OpenStep-specification. I haven't found any other
//ways to avoid this image-conversion, and the image-conversion
//takes a lot of time. The image-conversion has been written
//though and is in PXKDrawing.m in the function NSDrawBitmap().

/*
 * Below are declarations of the drawing engine's global variables.
 */

static NSColor *currentColor;
static Display *pxkDisplay;
//static NSFont *_currentFont;

NSPoint pxkCurrentPos;

GC pxkBlackGC;         /* Global drawing contexts for the most commonly      */
GC pxkWhiteGC;         /* used colors: Black, White, Light Gray, and         */
GC pxkLightgrayGC;     /* Dark Gray.  The drawing engine is smart            */
GC pxkDarkgrayGC;      /* enough to use these contexts when passed NSColors  */
                       /* that either have grayscale or RGB backings that    */
                       /* correspond to any of these predefined contexts.    */
GC pxkColorTextGC;

/* Private function, to be used solely by PXKGeometry and this source file */

void setCurrentColor(NSColor *aColor)
{
    currentColor = aColor;
}

/* Rectangle Drawing Functions */

void NSEraseRect(NSRect aRect)           /* fills the rect with white.       */
{
    pxkFillRect(aRect, [NSColor whiteColor]);    
}

void NSHighlightRect(NSRect aRect)       /* LTGRAY->WHITE, and vice-versa.   */
{
}

void NSRectClip(NSRect aRect)            /* Intersects the current clip path */
{
}

void NSRectClipList(const NSRect *rects, int count) 
{
}

void NSRectFill(NSRect aRect)            /* Fills rect using current color   */
{
    pxkFillRect(aRect, currentColor);
}

void NSRectFillList(const NSRect *rects, int count) 
{
    int i;
    
    for (i=0; i<count; i++) {
	pxkFillRect(rects[i], currentColor);
    }
}

/*
 * NSRectFillWithGrays() Fills each rectangle in the array 'rects'
 * with the gray whose value is stored at the corresponding location
 * in the array 'grays'.  Both arrays must be count elements long.
 * Avoid overlapping rectangles, because the drawing order can't be
 * guaranteed.
 */

void NSRectFillListWithGrays(const NSRect *rects,const float *grays,int count) 
{
}


/* Draw a line (private function) */

void
pxkDrawLine(NSPoint a, NSPoint b, NSColor *color)
{
    NSView *focusView = [NSView focusView];
    int windowId;
    
    pxkDisplay = xDisplay();
    
    if (focusView != nil) {
	windowId = [focusView viewNum];
	
	/* Let's cheat a little in order to speed up things */

	if (color == [NSColor blackColor]) {
	    //puts("black color");
	    //fprintf(stdout, "windowId: %d\n", windowId);
	    XDrawLine(pxkDisplay,
		      xWindowWithId(windowId),
		      pxkBlackGC,
		      a.x,a.y,b.x,b.y);
	} else if (color == [NSColor whiteColor]) {
	    XDrawLine(pxkDisplay,
		      xWindowWithId(windowId),
		      pxkWhiteGC,
		      a.x,a.y,b.x,b.y);
	} else if (color == [NSColor darkGrayColor]) {
	    XDrawLine(pxkDisplay,
		      xWindowWithId(windowId),
		      pxkDarkgrayGC,
		      a.x,a.y,b.x,b.y);
	} else if (color == [NSColor lightGrayColor]) {
	    XDrawLine(pxkDisplay,
		      xWindowWithId(windowId),
		      pxkLightgrayGC,
		      a.x,a.y,b.x,b.y);
	} else if (color == [NSColor redColor]) {
	    XDrawLine(pxkDisplay,
		      xWindowWithId(windowId),
		      pxkColorTextGC,
		      a.x,a.y,b.x,b.y);
	}
    } else {
	puts("No view in focus!!!");
    }
}

/*
 * Draw a line that's defined as an NSRect instead of 2 points
 */

void
pxkDrawLineRect(NSRect line, NSColor *color)
{
    NSPoint a = line.origin;
    NSPoint b;

    /* Do not display lines where width or height < 0 */
    
    if ((line.size.width >= 0) && (line.size.height >= 0)) {
	b.x = line.origin.x + line.size.width;
	b.y = line.origin.y + line.size.height;    
	pxkDrawLine(a, b, color);
    }
}

void
pxkFillRect(NSRect rect, NSColor *color)
{
    NSView *focusView = [NSView focusView];
    int windowId;

    pxkDisplay = xDisplay();

    if (focusView != nil) {
	windowId = [focusView viewNum];
	
	/* Let's cheat a little in order to speed up things */

	if (color == [NSColor blackColor]) {
	    XFillRectangle(pxkDisplay,
			   xWindowWithId(windowId),
			   pxkBlackGC,
			   rect.origin.x,
			   rect.origin.y,
			   rect.size.width,
			   rect.size.height);
	} else if (color == [NSColor whiteColor]) {
	    XFillRectangle(pxkDisplay,
			   xWindowWithId(windowId),
			   pxkWhiteGC,
			   rect.origin.x,
			   rect.origin.y,
			   rect.size.width,
			   rect.size.height);
	} else if (color == [NSColor darkGrayColor]) {
	    XFillRectangle(pxkDisplay,
			   xWindowWithId(windowId),
			   pxkDarkgrayGC,
			   rect.origin.x,
			   rect.origin.y,
			   rect.size.width,
			   rect.size.height);
	} else if (color == [NSColor lightGrayColor]) {
	    XFillRectangle(pxkDisplay,
			   xWindowWithId(windowId),
			   pxkLightgrayGC,
			   rect.origin.x,
			   rect.origin.y,
			   rect.size.width,
			   rect.size.height);
	} else {
	    
	    pxkSetColorInGC();
	    
	    XFillRectangle(pxkDisplay,
			   xWindowWithId(windowId),
			   pxkColorTextGC,
			   rect.origin.x,
			   rect.origin.y,
			   rect.size.width,
			   rect.size.height);
	}
    } else {
	pxkDebugMsg("No view in focus!!!");
    }
}


/* Private functions, not part of the OpenStep specification */
NSPoint intersectionPoint(NSRect lineA, NSRect lineB)
{
    NSPoint aPoint;
    float r, s;
    float XA = lineA.origin.x;
    float XB = lineA.origin.x + lineA.size.width;
    float XC = lineB.origin.x;
    float XD = lineB.origin.x + lineB.size.width;
    float YA = lineA.origin.y;
    float YB = lineA.origin.y + lineA.size.height;
    float YC = lineB.origin.y;
    float YD = lineB.origin.y + lineB.size.height;
    float denominator = (XB-XA)*(YD-YC)-(YB-YA)*(XD-XC);
    
    /*
      The following code taken from comp.graphics.algorithms FAQ.
      
      This problem can be extremely easy or extremely difficult
      depends on your applications. If all you want is the
      intersection point, the following should work:
      
      Let A,B,C,D be 2-space position vectors. Then the directed
      line segments AB and CD are given by:
	  
	  AB=A+r(B-A), r in [0,1]
	  CD=C+s(D-C), s in [0,1]
		  
          If AB and CD intersect, then

	  A+r(B-A)=C+s(D-C)

          or

	  XA+r(XB-XA)=XC+s(XD-XC)
	  YA+r(YB-YA)=YC+s(YD-YC)  for some r,s in [0,1]
	  
          Solving the above for r and s yields
 
	       (YA-YC)(XD-XC)-(XA-XC)(YD-YC)
          r = -----------------------------  (eqn 1)
	       (XB-XA)(YD-YC)-(YB-YA)(XD-XC)
	  
	       (YA-YC)(XB-XA)-(XA-XC)(YB-YA)
	   s = -----------------------------  (eqn 2)
	       (XB-XA)(YD-YC)-(YB-YA)(XD-XC)
	       
	   */
    
    if ((denominator < 0.00000001) && (denominator > -0.00000001)) {
	//fprintf(stderr, "in function intersect(), lines are parallel.\n");
	aPoint.x = -1;
	aPoint.y = -1; /* x and y are always positive in this toolkit */
    }
    
    r = ((YA-YC)*(XD-XC)-(XA-XC)*(YD-YC)) / denominator;
    s = ((YA-YC)*(XB-XA)-(XA-XC)*(YB-YA)) / denominator;
    
         /*
          Let I be the position vector of the intersection point, then

	  I=A+r(B-A)

          or

	  XI=XA+r(XB-XA)
	  YI=YA+r(YB-YA)    

	  */
	
    aPoint.x = XA+r*(XB-XA);
    aPoint.y = YA+r*(YB-YA);
    
    /*
	  By examining the values of r and s, you can also determine
	  some other limiting conditions:
	  
	  If 0 <= r <= 1 and 0 <= s <= 1, intersection exists
	  r < 0 or r>1 or s < 0 or s>1 line segments do not intersect
	  
	  If the denominator in eqn 1 is zero, AB and CD are parallel
          If the numerator in eqn 1 is also zero, AB and CD are
          coincident

          If the intersection point of the 2 lines are needed (lines in
          this context mean infinite lines) regardless whether the two
          line segments intersect, then

          If r > 1, I is located on extension of AB
          If r < 0, I is located on extension of BA
          If s > 1, I is located on extension of CD
          If s < 0, I is located on extension of DC

          Also note that the denominators of equations 1 and 2 are
          identical.

          References:
	  
          [O'Rourke] pp. 249-51
          [Gems3] pp. 199-202 "Faster Line Segment Intersection,"
	  */

    return aPoint;
}

float dist2D(NSPoint a, NSPoint b)
{
    return (a.x-b.x)*(a.x - b.x) / ((a.y-b.y)*(a.y-b.y));
}

NSRect clipLineToRect(NSRect aLine, NSRect aRect)
{
    NSRect resultLine = aLine;
    NSRect clipLine;
    NSPoint aPoint;
    float XA, XB, YA, YB;
    
    XA = aLine.origin.x;
    XB = aLine.origin.x + aLine.size.width;
    YA = aLine.origin.y;
    YB = aLine.origin.y + aLine.size.height;

    /*
     * Set height = width = -1  for lines that are totally
     * outside of the clipping rectangle.
     */

    if (((XA < aRect.origin.x) && (XB < aRect.origin.x)) ||
	((XA > aRect.origin.x + aRect.size.width) &&
	 (XB > aRect.origin.x + aRect.size.width)) ||
	((YA < aRect.origin.y) && (YB < aRect.origin.y)) ||
	((YA > aRect.origin.x + aRect.size.width) &&
	 (YB > aRect.origin.x + aRect.size.width)))
    {
	resultLine.size.width = resultLine.size.height = -1;
	return resultLine;
    }
    
    /* clip left */
    
    if ((XA <= aRect.origin.x) && (XB >= aRect.origin.x)) {

#if PXKDE_DEBUG
	pxkDebugMsg("clip left");
#endif
	clipLine = aRect;
	clipLine.size.width = 0.0;
	
	aPoint = intersectionPoint(aLine, clipLine);
    
	resultLine.origin.x = aPoint.x;
	resultLine.origin.y = aPoint.y;
	resultLine.size.width = XB - aPoint.x;
	resultLine.size.height = YB - aPoint.y;
    }

    /* clip right */
    
    if ((XA <= aRect.origin.x + aRect.size.width) &&
	(XB >= aRect.origin.x + aRect.size.width))
    {

#if PXKDE_DEBUG	
	pxkDebugMsg("clip right");
#endif
 
	clipLine = aRect;
	clipLine.origin.x += clipLine.size.width;
	clipLine.size.width = 0.0;
	
	aPoint = intersectionPoint(aLine, clipLine);
    
	resultLine.origin.x = aPoint.x;
	resultLine.origin.y = aPoint.y;
	resultLine.size.width = XB - aPoint.x;
	resultLine.size.height = YB - aPoint.y;
    }

    /* clip bottom */
    
    if ((YA <= aRect.origin.y + aRect.size.height) &&
	(YB >= aRect.origin.y + aRect.size.height))
    {
	
#if PXKDE_DEBUG
	pxkDebugMsg("clip bottom");
#endif

	clipLine = aRect;
	clipLine.size.height = 0.0;
	
	aPoint = intersectionPoint(aLine, clipLine);
    
	resultLine.origin.x = aPoint.x;
	resultLine.origin.y = aPoint.y;
	resultLine.size.width = XB - aPoint.x;
	resultLine.size.height = YB - aPoint.y;
    }

    /* clip top */
    
    if ((YA <= aRect.origin.y) && (YB >= aRect.origin.y))
    {
#if PXKDE_DEBUG
	pxkDebugMsg("clip top");	
#endif
	clipLine = aRect;
	clipLine.origin.x += clipLine.size.height;
	clipLine.size.height = 0.0;
	
	aPoint = intersectionPoint(aLine, clipLine);
    
	resultLine.origin.x = aPoint.x;
	resultLine.origin.y = aPoint.y;
	resultLine.size.width = XB - aPoint.x;
	resultLine.size.height = YB - aPoint.y;
    }

    return resultLine;
}

/* Draw a Bordered Rectangle */

void NSDrawButton(NSRect aRect, NSRect clipRect)     /* Bordered ltgray rect */
{
    NSRect line;
    
    /* Draw the top white line */    
    line = aRect;
    line.size.height = 0;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);
    
    /* Draw the left white line */
    line = aRect;
    line.size.width = 0;
    line.size.height -= 1; /* line doesn't draw down to the last pixel */
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);    
    
    /* Draw the bottom black line */
    line = aRect;
    line.origin.y += line.size.height;
    line.size.height = 0;    
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor blackColor]);

    /* Draw the bottom darkgray line above the black line */
    line.origin.y--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);

    /* Draw the right black line */
    line = aRect;
    line.origin.x += line.size.width;
    line.size.width = 0;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor blackColor]);

    /* Draw the right darkgray line */
    line.origin.x--;
    line.origin.y++;
    line.size.height -= 2.0; /* line doesn't draw down to the last pixel */
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);  
}

void NSDrawGrayBezel(NSRect aRect, NSRect clipRect)  /* pushed-in button     */
{ /* Like the border of a TextField? */
    NSRect line;
    
    /* Draw the top darkgray line */ 
    line = aRect;
    line.size.height = 0;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);
    
    /* Draw the top darkgray line below the darkgray line */    
    line.origin.y++;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);
    
    /* Draw the left darkgray line */
    line = aRect;
    line.size.width = 0;
    line.size.height--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);

    /* Draw the left darkgray line, besides the vertical darkgray line */
    line.origin.x++;
    line.size.height--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);    
    
    /* Draw the bottom white line */
    line = aRect;
    line.origin.y += line.size.height;
    line.size.height = 0;    
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);

    /* Draw the bottom lightgray line above the white line */
    line.origin.x++;
    line.size.width--;
    line.origin.y--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor lightGrayColor]);

    /* Draw the right white line */
    line = aRect;
    line.origin.x += line.size.width;
    line.size.width = 0;
    line.origin.y++;
    line.size.height--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);

    /* Draw the right lightgray line */
    line.origin.x--;
    line.origin.y++;
    line.size.height -= 2.0; /* line doesn't draw down to the last pixel */
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor lightGrayColor]);
}

void NSDrawGroove(NSRect aRect, NSRect clipRect)     /* like a Box           */
{
    NSRect line;
    
    /* Draw the top darkgray line */    
    line = aRect;
    line.size.height = 0;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);
    
    /* Draw the top white line below the darkgray line */    
    line.origin.y++;
    line.origin.x++;
    line.size.width--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);
    
    /* Draw the left darkgray line */
    line = aRect;
    line.size.width = 0;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);

    /* Draw the left white line, besides the vertical darkgray line */
    line.origin.x++;
    line.origin.y++;
    line.size.height--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);    
    
    /* Draw the bottom white line */
    line = aRect;
    line.origin.x++;
    line.size.width--;
    line.origin.y += line.size.height;
    line.size.height = 0;    
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);

    /* Draw the bottom darkgray line above the white line */
    line.origin.x++;
    line.size.width--;
    line.origin.y--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);

    /* Draw the right white line */
    line = aRect;
    line.origin.x += line.size.width;
    line.size.width = 0;
    line.origin.y++;
    line.size.height--;
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor whiteColor]);

    /* Draw the right darkgray line */
    line.origin.x--;
    line.origin.y++;
    line.size.height -= 2.0; /* line doesn't draw down to the last pixel */
    pxkDrawLineRect(clipLineToRect(line, clipRect), [NSColor darkGrayColor]);  
}

#if 0
void NSDrawTiledRects(NSRect boundsRect,
		      NSRect clipRect,
		      const NSRectEdge *sides,
		      const float *grays,
		      int count) 
{
}
#endif

void NSDrawWhiteBezel(NSRect aRect, NSRect clipRect)  /* like a TextField    */
{
}

void NSFrameRect(NSRect aRect)  /* Draws a frame of width 1.0 use curr color */
{
    NSPoint a, b;
    
    a = aRect.origin;
    b.x = aRect.origin.x + aRect.size.width;
    b.y = aRect.origin.y;
    pxkDrawLine(a, b, currentColor);

    a.x = b.x;
    a.y += aRect.origin.y + aRect.size.height;
    pxkDrawLine(a, b, currentColor);

    b.x = aRect.origin.x;
    b.y = a.y;
    pxkDrawLine(a, b, currentColor);

    pxkDrawLine(aRect.origin, b, currentColor);
}

void NSFrameRectWithWidth(NSRect aRect, float frameWidth) 
{
}

void NSDrawBitmap(NSRect rect,int pixelsWide,int pixelsHigh,
       int bitsPerSample,int samplesPerPixel,int bitsPerPixel,
       int bytesPerRow,BOOL isPlanar,BOOL hasAlpha,
       NSString *colorSpaceName,const unsigned char *const data[5])
{
  XImage *ximg;
  NSPoint imgPos;
  NSView *focusView = [NSView focusView];
  int windowId;
  Visual *visual;
  int defDepth;
  NSMutableData *tmpDataArea;
  unsigned char *tmpDataPtr,*tmpPtr;
  int tmpDataLength;
  int cnt;
  
  if(!focusView)
  {
    puts("No view in focus!");
    return;
  }
  windowId = [focusView viewNum];
  imgPos = pxkGetCurrentPos();
  NSCAssert(data!=NULL,@"data is NULL!");
  NSCAssert(data[0]!=NULL,@"data[0] is NULL!");
  pxkDisplay = xDisplay();
  NSCAssert(pxkDisplay!=NULL,@"pxkDisplay is NULL!");
  puts("Testing XImage...");
  
  visual = DefaultVisual(pxkDisplay,xScreen());
  NSCAssert(visual!=NULL,@"visual is NULL!");
  defDepth = DefaultDepth(pxkDisplay,xScreen());
  printf("bits_per_rgb: %ld\n",(long)visual->bits_per_rgb);

  printf("bitsPerPixel:%ld\n",(long)bitsPerPixel);
  if(defDepth!=bitsPerPixel)
  {
    puts("data conversion of imagedata is needed...SO LAME!");
    //FIXME: Doesn't X-Windows provide methods to do this when the imagedata
    //       is sent to the X-server with XPutImage...?
    tmpDataLength = defDepth * pixelsWide * pixelsHigh;
    tmpDataArea = [NSMutableData dataWithLength:tmpDataLength];
    if(!tmpDataArea)
    {
      puts("Out of memory!");
      return;
    }
    [tmpDataArea autorelease];
    tmpDataPtr = [tmpDataArea mutableBytes];
    tmpPtr = (unsigned char *)data[0];
    if(bitsPerPixel==24 && defDepth==32)
    {
      for(cnt=0;cnt<pixelsWide*pixelsHigh;cnt++)
      {
        //Converting from BGR to ABGR
        *tmpDataPtr++ = (unsigned char)0;
        *tmpDataPtr++ = *tmpPtr++;
        *tmpDataPtr++ = *tmpPtr++;
        *tmpDataPtr++ = *tmpPtr++;
      }
      tmpDataPtr = [tmpDataArea mutableBytes];
    }
  }
  else
    tmpDataPtr = (unsigned char *)data[0];
  ximg = XCreateImage(pxkDisplay,DefaultVisual(pxkDisplay,xScreen()),
           defDepth,ZPixmap,0,tmpDataPtr,pixelsWide,
           pixelsHigh,8,0);// FIXME: Not always ZPixmap, padding???
  if(!ximg)
  {
     puts("XCreateImage failed...");
     return;
  }
  puts("XImage created...");
  ximg->byte_order = MSBFirst;

  puts("Putting image...");
  XPutImage(pxkDisplay,xWindowWithId(windowId),pxkBlackGC,ximg,0,0,
    imgPos.x,imgPos.y,pixelsWide,pixelsHigh);
  puts("XImage has been put!");
  XFree(ximg);
  return;
}

void
pxkSetColorInGC(void)
{
    XColor c1;
    
    pxkDisplay = xDisplay();

    c1.red = [currentColor redComponent] * PXK_MAX_COLOR;
    c1.green = [currentColor greenComponent] * PXK_MAX_COLOR;
    c1.blue = [currentColor blueComponent] * PXK_MAX_COLOR;

#if PXKDE_DEBUG
    fprintf(stdout, "rgb: %2.2f %2.2f %2.2f\n",
	    [currentColor redComponent],
	    [currentColor greenComponent],
	    [currentColor blueComponent]);
#endif
    
    XAllocColor(pxkDisplay,
		XDefaultColormap(pxkDisplay, xScreen() ),
		&c1);
	
    XSetForeground(pxkDisplay, pxkColorTextGC, c1.pixel);
}

void
pxkDrawString(const char *string, NSPoint location)
{
    NSView *focusView = [NSView focusView];
    int windowId;
    XFontStruct *font_info;
    BOOL skipFont = NO;
    NSFont *curr = [NSFont currentFont];
    char fontname[256];   

#if 0
    char *system = "-misc-fixed-medium-r-normal--15-*-iso8859-1";
    char *helvetica = "-adobe-helvetica-medium-r-normal--14*-iso8859-1";
    char *times = "*times*-r-normal--14";
    char *courier = "*courier*-r-normal--14";
#endif
    
    char slant;
    char weight[10];
    
    if (focusView == nil) {
	return;
    }
    
    pxkDisplay = xDisplay();
    windowId = [focusView viewNum];
    
    // build the font name

    if ([curr isBold]) {
	strcpy(weight, "bold");
    } else {
	strcpy(weight, "medium");
    }
    
    if ([curr isItalic]) { 
	slant = 'o';
    } else {
	slant = 'r';
    }
    
    if (! strcmp([[curr familyName] cString], "fixed"))
    {
	//strcpy(fontname, "-adobe-helvetica-bold-o-normal--24*-iso8859-1");
	
	sprintf(fontname, "-sony-fixed-%s-%c-normal--%i*-iso8859-1",
		weight, slant, [curr pointSize]);

    } else {
	sprintf(fontname, "-adobe-%s-%s-%c-normal--%i*-iso8859-1", 
		[[curr familyName] cString],
		weight,
		slant,
		(int)[curr pointSize]);
    }
    
#if PXKDE_DEBUG
    puts(fontname);
#endif
    
    /* Load font and get font information structure. */
    if ((font_info = XLoadQueryFont(pxkDisplay,fontname)) == NULL)
    {
	(void) fprintf( stderr, "Drawing Engine error: cannot open"
			" 9x15 font\n");
	skipFont = YES;
    }
    
    // Set the drawing color for the text
    
    if (currentColor == nil) {
	currentColor = [NSColor blackColor];	

	// Set the font
	if (skipFont == NO) {
	    XSetFont(pxkDisplay, pxkBlackGC, font_info->fid);

#if PXKDE_DEBUG
	    puts("setting font for black colored text");
#endif
	}
	
	XDrawString(pxkDisplay,
		    xWindowWithId(windowId),
		    pxkBlackGC,
		    location.x,
		    location.y,
		    string,
		    strlen(string));
	
    } else {
	
	// Set the font
	if (skipFont == NO) {
	    XSetFont(pxkDisplay, pxkColorTextGC, font_info->fid);
#if PXKDE_DEBUG
	    puts("setting font for colored text");
#endif
	}

	pxkSetColorInGC();
	
	XDrawString(pxkDisplay,
		    xWindowWithId(windowId),
		    pxkColorTextGC,
		    location.x,
		    location.y,
		    string,
		    strlen(string));
    }
}

int
pxkWidthOfString(NSString *string, NSFont *aFont)
{
    XFontStruct *font_info;
    BOOL skipFont = NO;
    char fontname[256];   
    char slant;
    char weight[10];

    if ((string == nil) || ([string length] == 0))
    {
	return 0;
    }
    
    pxkDisplay = xDisplay();
    
    // build the font name

    if ([aFont isBold]) {
	strcpy(weight, "bold");
    } else {
	strcpy(weight, "medium");
    }
    
    if ([aFont isItalic]) { 
	slant = 'o';
    } else {
	slant = 'r';
    }
    
    if (! strcmp([[aFont familyName] cString], "fixed"))
    {
	//strcpy(fontname, "-adobe-helvetica-bold-o-normal--24*-iso8859-1");
		
	sprintf(fontname, "-sony-fixed-%s-%c-normal--%i*-iso8859-1",
		weight, slant, [aFont pointSize]);

    } else {
	sprintf(fontname, "-adobe-%s-%s-%c-normal--%i*-iso8859-1", 
		[[aFont familyName] cString],
		weight,
		slant,
		(int)[aFont pointSize]);
    }

#if PXKDE_DEBUG    
    fprintf(stdout, "font for width: %s\n", fontname);
#endif
    
    /* Load font and get font information structure. */
    if ((font_info = XLoadQueryFont(pxkDisplay,fontname)) == NULL)
    {
	(void) fprintf( stderr, "Drawing Engine error: cannot open"
			" 9x15 font\n");
	skipFont = YES;
    }
    
    return XTextWidth(font_info, [string cString], [string length]);
}

void pxkSetCurrentPos(NSPoint newPosition)
{
  pxkCurrentPos = newPosition;
}

NSPoint pxkGetCurrentPos(void)
{
  return pxkCurrentPos;
}

unsigned int pxkFVRBitsPerSample(void)
{
  return 8;
}

unsigned int pxkFVRSamplesPerPixel(void)
{
  if(DefaultDepth(xDisplay(),xScreen())==32)
    return 4;
  else
    return 3;
}

BOOL pxkFVRHasAlpha(void)
{
  return (pxkFVRSamplesPerPixel()>3);
}

BOOL pxkFVRIsPlanar(void)
{
  return NO;
}

NSMutableData *pxkFVRGrab(NSRect rect)
{
  XImage *ximg;
  Window srcXWin;
  NSMutableData *newImageData;
  NSView *focusView = [NSView focusView];
  int windowId;
  unsigned int dataLength;
  unsigned int cnt;
  unsigned char *srcPtr,*destPtr;

  if(!focusView)
  {
    puts("No view in focus!");
    return nil;
  }
  windowId = [focusView viewNum];
  srcXWin = xWindowWithId(windowId);
  NSCAssert(srcXWin!=0,@"srcXWin is 0");
  pxkDisplay = xDisplay();
  NSCAssert(pxkDisplay!=NULL,@"pxkDisplay is NULL!");
  
  ximg = XGetImage(pxkDisplay, srcXWin,rect.origin.x,rect.origin.y,
    rect.size.width, rect.size.height, AllPlanes, ZPixmap);
  puts("Got XImage...");
  if(!ximg)
  {
     puts("XGetImage failed...");
     return nil;
  }
  dataLength = ximg->bytes_per_line*ximg->height;
  newImageData = [NSMutableData dataWithLength:dataLength];
  //newImageData = [NSMutableData dataWithBytes:ximg->data
  //                 length:ximg->bytes_per_line*ximg->height];
  if(newImageData)
    [newImageData retain];
  else
  {
    XDestroyImage(ximg);
    return nil;
  }
  puts("Got mutable imagedata...");
  destPtr = [newImageData mutableBytes];
  srcPtr = ximg->data;
  if(ximg->byte_order == LSBFirst)
  {
    if(ximg->depth==32)
    {
      dataLength /= 4;
      for(cnt=0;cnt<dataLength;cnt++)
      {
        //Swap bytes
        *destPtr++ = *(srcPtr+3);
        *destPtr++ = *(srcPtr+2);
        *destPtr++ = *(srcPtr+1);
        *destPtr++ = *srcPtr;
        srcPtr+=4;
      }
    }
    else if(ximg->depth==24)
    {
      dataLength /= 3;
      for(cnt=0;cnt<dataLength;cnt++)
      {
        //Swap bytes
        *destPtr++ = *(srcPtr+2);
        *destPtr++ = *(srcPtr+1);
        *destPtr++ = *srcPtr;
        srcPtr+=3;
      }
    }
  }
  XDestroyImage(ximg);
  puts("Destroyed ximg");
  return newImageData;
}

BOOL pxkCopyWindowRect(NSWindow *win,NSRect rect)
{
  NSPoint imgPos;
  Window srcXWin,destXWin;
  NSView *focusView = [NSView focusView];
  NSView *srcView;
  int windowId;
  
  if(!focusView)
  {
    puts("No view in focus!");
    return NO;
  }
  NSCAssert(win!=nil,@"win is nil!");
  srcView = [win contentView];
  NSCAssert(srcView!=nil,@"srcView is nil!");
  windowId = [srcView viewNum];
  srcXWin = xWindowWithId(windowId);
  NSCAssert(srcXWin!=0,@"srcXWin is 0");
  windowId = [focusView viewNum];
  destXWin = xWindowWithId(windowId);
  NSCAssert(destXWin!=0,@"destXWin is 0");
  
  imgPos = pxkGetCurrentPos();
  pxkDisplay = xDisplay();
  NSCAssert(pxkDisplay!=NULL,@"pxkDisplay is NULL!");
  
  XCopyArea(pxkDisplay,srcXWin,destXWin,pxkBlackGC,
    rect.origin.x,rect.origin.y,
    rect.size.width,rect.size.height,
    imgPos.x,imgPos.y);
  return YES;
}