/*
 * Filename: PXKEventFactory.m
 * Project:  NSXKit
 * Module:   Main application components
 *
 * Function: See header file
 *
 * Author:   Pascal Forget
 * Date:     Mars 1996
 *
 *
 * Updated 10/7/97 by Stuart Croy to include Tracking-Rectangle Events

 * Copyright (C) 1996 Free Software Foundation, Inc.
 */

#include "PXKEventFactory.h"
//#include "NSApplication.h"
#include "PXKDrawing.h"
#include "PXKGeometry.h"
#include "misc.h"

/*
 * Current event settings stored here as we don't want to
 * waste CPU cycles creating / destroying these variables
 * each time we go through a cycle in the event loop
 */
XEvent xEvent;
NSEvent *_currentEvent;
extern NSEvent *NULL_EVENT;
NSEventType eventType;
unsigned short eventModifierFlags;
NSPoint eventLocation;
unsigned int eventFlags;
NSTimeInterval eventTime;
KeySym keysym;
unsigned short keyCode;

/*
 * Private functions (function calls are faster than method calls)
 */
void
processKeyDown(void);

void
processKeyUp(void);

void
processMouseDown(void);

void
processMouseMoved(void);

void
processMouseUp(void);

void
setKeyCode(void);

void
setModifierFlags(unsigned int state);

void
updateKeyAndMainWindow(void);

NSEvent *
currentEvent(void)
{
    return _currentEvent;
}

NSEvent *
nextEvent(void)
{
    Display *pxkDisplay = xDisplay(); /* xDisplay() is in PXKDrawingEngine.m */
//    XFontStruct *font_info;
    NSSize size;

    _currentEvent = NULL_EVENT;
	
    XNextEvent(pxkDisplay, &xEvent);
    switch  (xEvent.type) {

      case Expose: break; /* Not implemented yet */	    

      case ConfigureNotify:
	/* window has been resized, change width and
	 * height to send to draw_text and draw_graphics
	 * in next Expose */

	updateKeyAndMainWindow();

	size.width = (float)xEvent.xconfigure.width;
	size.height = (float)xEvent.xconfigure.height; 
	[[NSApp keyWindow] setContentSize:size];
	break;
	    
      case ButtonPress:   processMouseDown();   break;
      case ButtonRelease: processMouseUp();     break;
      case MotionNotify:  processMouseMoved();  break;
	    
      case KeyPress:
	processKeyDown();
//	XRefreshKeyboardMapping(&xEvent);
	break;
	
      case KeyRelease:
	processKeyUp(); /* FALL THROUGH */	    
	    
      case MappingNotify:
//	XRefreshKeyboardMapping(&xEvent);
	break;
	
      case EnterNotify:
	msg("Enter notify\n");
	break;

      case LeaveNotify:
	msg("Leave notify\n");
	break;
	    
      default:
	/* all events selected by StructureNotifyEventMask
	 * except ConfigureNotify are thrown away here,
	 * since nothing is done with them */
	break;
	
    } /* end event type switch */    

    return _currentEvent;
}

/*
 * implementation of the private functions
 *
 * I've tried to create as few variables as possible
 * in the following functions as they are called over
 * and over again.
 *
 * The following functions are not part of GNUstep.
 * They are used to convert X events into NSEvents.
 *
 * - Pascal Forget <pascal@wsc.com>
 */

void
processMouseDown(void)
{
    XEvent event0, event1, event2;
    int clickCount = 1;

    Display *display = xDisplay();

    updateKeyAndMainWindow();
    
    setModifierFlags(xEvent.xbutton.state);
    
    if (xEvent.xbutton.button == Button1) {
	eventType = NSLeftMouseDown;
    } else {
	eventType = NSRightMouseDown;
    }
    
    eventLocation = pxkLocationInWindow(xEvent.xbutton.window,
					xEvent.xbutton.x,
					xEvent.xbutton.y);
    while(1) {
	/* Get rid of presses on other buttons */
      while(XCheckTypedEvent(display, ButtonPress, &event0)){};
	
	/* Wait for release; if on correct button, exit */ 
      XCheckMaskEvent(display, ButtonReleaseMask, &event0);

#if 1
	break;
#else	
	if (event0.xbutton.button == xEvent.xbutton.button)
	{
	    break;
	}
#endif	
    }

    /* Wait 0.1s, leaving time for the other events to arrive */
    usleep(10000); 

    /*
     * Now check if the same button has been pressed again.
     * If so, we consider the whole mouse-up-down-up sequence
     * as a single event, namely a Left mouseDown double-click.
     */

    if (XPending(display)>1) {
	XNextEvent(display, &event1);
	XNextEvent(display, &event2);

	if ((event1.type == ButtonPress) &&
	    (event1.xbutton.button == xEvent.xbutton.button) &&
	    (event2.type == ButtonRelease) &&
	    (event2.xbutton.button == xEvent.xbutton.button))
	{
	    clickCount = 2;
	    XPutBackEvent(display, &event2);
	} else {
	    XPutBackEvent(display, &event2);
	    XPutBackEvent(display, &event1);
	    XPutBackEvent(display, &event0);
	}
    }
    
    _currentEvent = [NSEvent mouseEventWithType: eventType
			     location: eventLocation
		      	     modifierFlags: eventModifierFlags
			     timestamp: (NSTimeInterval)xEvent.xbutton.time
			     windowNumber: numberForXWindow(
				 xEvent.xbutton.window)
			     context: nil
			     eventNumber: xEvent.xbutton.serial
			     clickCount: clickCount
			     pressure: 1.0];
}

/*
 * The processMouseMoved function generates a mouseEvent.
 * It either sends a mouseMoved or a mouseDragged message
 * to the key window, depending if a button was depressed
 * while the mouse moved.
 */

void
processMouseMoved(void)
{
    setModifierFlags(xEvent.xbutton.state);

    eventLocation = pxkLocationInWindow(xEvent.xmotion.window,
					xEvent.xmotion.x,
					xEvent.xmotion.y);
    
    if (xEvent.xbutton.button == Button1) {

	_currentEvent = [NSEvent mouseEventWithType:NSLeftMouseDragged
				location:eventLocation
				modifierFlags:eventModifierFlags
				timestamp:(NSTimeInterval)xEvent.xmotion.time
				windowNumber:numberForXWindow(
				    xEvent.xany.window)
				context:nil
				eventNumber:xEvent.xmotion.serial
				clickCount:1
				pressure:1.0];
    } else if (xEvent.xbutton.button == Button2) {

	_currentEvent = [NSEvent mouseEventWithType:NSRightMouseDragged
				location:eventLocation
				modiferFlags:eventModifierFlags
				timestamp:(NSTimeInterval)xEvent.xbutton.time
				windowNumber:numberForXWindow(
				    xEvent.xbutton.window)
				context:nil
				eventNumber:xEvent.xany.serial
				clickCount:1
				pressure:1.0];

    } else {

	_currentEvent = [NSEvent mouseEventWithType:NSMouseMoved
				location:eventLocation
				modifierFlags:eventModifierFlags
				timestamp:(NSTimeInterval)xEvent.xbutton.time
				windowNumber:numberForXWindow(
				    xEvent.xbutton.window)
				context:nil
				eventNumber:xEvent.xany.serial
				clickCount:0
				pressure:0.0];
    }
}

void
processMouseUp(void)
{
    updateKeyAndMainWindow();
    
    setModifierFlags(xEvent.xbutton.state);
        
    if (xEvent.xbutton.button == Button1) {
	eventType = NSLeftMouseUp;
    } else {
	eventType = NSRightMouseUp;
    }

    eventLocation = pxkLocationInWindow(xEvent.xbutton.window,
					xEvent.xbutton.x,
					xEvent.xbutton.y);
    
    _currentEvent = [NSEvent mouseEventWithType:eventType
			    location:eventLocation
			    modifierFlags:xEvent.xbutton.state
			    timestamp:(NSTimeInterval)xEvent.xbutton.time
			    windowNumber:numberForXWindow(
				xEvent.xbutton.window)
			    context:nil
			    eventNumber:xEvent.xany.serial
			    clickCount:1
			    pressure:1.0];
}

void
processKeyDown(void)
{
    char buf[256];
    int count;
    XComposeStatus compose;
    NSString *keys, *ukeys;

    updateKeyAndMainWindow();
    
    setModifierFlags(xEvent.xkey.state);
    
    count = XLookupString(&xEvent, buf, 256, &keysym, &compose);

    /* Make sure that the string is properly terminated */
		      
    if (count > 255) {
	buf[255] = '\0';
    } else if (count < 1) {
	buf[0] = '\0';
    } else {
	buf[count] = '\0';
    }

    eventLocation = pxkLocationInWindow(xEvent.xkey.window,
					xEvent.xkey.x,
					xEvent.xkey.y);
    
    setKeyCode();
    
    ukeys = [NSString stringWithCString:buf];
    keys = ukeys; /* Stupid implementation (to be replaced) */
    
    _currentEvent = [NSEvent keyEventWithType:NSKeyDown
			    location:eventLocation
			    modifierFlags:eventFlags
			    timestamp:(NSTimeInterval)xEvent.xkey.time
			    windowNumber:numberForXWindow(
				xEvent.xkey.window)
			    context:nil
			    characters:keys
			    charactersIgnoringModifiers:ukeys
			    isARepeat:NO
			    keyCode:keyCode];
}

void
processKeyUp(void)
{
    char buf[256];
    int count;
    XComposeStatus compose;
    NSString *keys, *ukeys;

    updateKeyAndMainWindow();

    setModifierFlags(xEvent.xkey.state);
    
    count = XLookupString(&xEvent, buf, 256, &keysym, &compose);

    /* Make sure that the string is properly terminated */
    
    if (count > 255) {
	buf[255] = '\0';
    } else if (count < 1) {
	buf[0] = '\0';
    } else {
	buf[count] = '\0';
    }

    ukeys = [NSString stringWithCString:buf];
    keys = ukeys; /* Stupid implementation (to be replaced) */    


    eventLocation = pxkLocationInWindow(xEvent.xkey.window,
					xEvent.xkey.x,
					xEvent.xkey.y);
    
    _currentEvent = [NSEvent keyEventWithType:NSKeyUp
			    location:eventLocation
			    modifierFlags:eventModifierFlags
			    timestamp:(NSTimeInterval)xEvent.xkey.time
			    windowNumber:numberForXWindow(xEvent.xany.window)
			    context:nil
			    characters:keys
			    charactersIgnoringModifiers:ukeys
			    isARepeat:NO
			    keyCode:keyCode];
}

void
setKeyCode(void)
{
    if ((keysym == XK_Return) ||
	(keysym == XK_KP_Enter) ||
	(keysym == XK_Linefeed))
    {
	/* do nothing for now */
	
    } else if ((keysym >= XK_Shift_L) && (keysym <= XK_Hyper_R)) {

	/*
	 * The shift button is depressed.  This should have already
	 * been detected in the setModifierFlags function.  Therefore,
	 * nothing is done here...
	 */
    }

    /* if a function key was pressed */
    
    if ((keysym >= XK_F1) && (keysym <= XK_F35)) {

	eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	
	switch(xEvent.xkey.keycode) {
	  case XK_F1: keyCode = NSF1FunctionKey; break;
	  case XK_F2: keyCode = NSF2FunctionKey; break;
	  case XK_F3: keyCode = NSF3FunctionKey; break;
	  case XK_F4: keyCode = NSF4FunctionKey; break;
	  case XK_F5: keyCode = NSF5FunctionKey; break;
	  case XK_F6: keyCode = NSF6FunctionKey; break;
	  case XK_F7: keyCode = NSF7FunctionKey; break;
	  case XK_F8: keyCode = NSF8FunctionKey; break;
	  case XK_F9: keyCode = NSF9FunctionKey; break;
	  case XK_F10: keyCode = NSF10FunctionKey; break;
	  case XK_F11: keyCode = NSF11FunctionKey; break;
	  case XK_F12: keyCode = NSF12FunctionKey; break;
	  case XK_F13: keyCode = NSF13FunctionKey; break;
	  case XK_F14: keyCode = NSF14FunctionKey; break;
	  case XK_F15: keyCode = NSF15FunctionKey; break;
	  case XK_F16: keyCode = NSF16FunctionKey; break;
	  case XK_F17: keyCode = NSF17FunctionKey; break;
	  case XK_F18: keyCode = NSF18FunctionKey; break;
	  case XK_F19: keyCode = NSF19FunctionKey; break;
	  case XK_F20: keyCode = NSF20FunctionKey; break;
	  case XK_F21: keyCode = NSF21FunctionKey; break;
	  case XK_F22: keyCode = NSF22FunctionKey; break;
	  case XK_F23: keyCode = NSF23FunctionKey; break;
	  case XK_F24: keyCode = NSF24FunctionKey; break;
	  case XK_F25: keyCode = NSF25FunctionKey; break;
	  case XK_F26: keyCode = NSF26FunctionKey; break;
	  case XK_F27: keyCode = NSF27FunctionKey; break;
	  case XK_F28: keyCode = NSF28FunctionKey; break;
	  case XK_F29: keyCode = NSF29FunctionKey; break;
	  case XK_F30: keyCode = NSF30FunctionKey; break;
	  case XK_F31: keyCode = NSF31FunctionKey; break;
	  case XK_F32: keyCode = NSF32FunctionKey; break;
	  case XK_F33: keyCode = NSF33FunctionKey; break;
	  case XK_F34: keyCode = NSF34FunctionKey; break;
	  case XK_F35: keyCode = NSF35FunctionKey; break;

	  default: /* do nothing */
	}
    } else {
	if ((keysym == XK_BackSpace) || (keysym == XK_Delete)) {
	    keyCode = NSDeleteFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Home) {
	    keyCode = NSHomeFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Left) {
	    keyCode = NSLeftArrowFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Up) {
	    keyCode = NSUpArrowFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Right) {
	    keyCode = NSRightArrowFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Down) {
	    keyCode = NSDownArrowFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Prior) {
	    keyCode = NSPrevFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Page_Up) {
	    keyCode = NSPageUpFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Next) {
	    keyCode = NSNextFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Page_Down) {
	    keyCode = NSPageDownFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_End) {
	    keyCode = NSEndFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Begin) {
	    keyCode = NSBeginFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Select) {
	    keyCode = NSSelectFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Print) {
	    keyCode = NSPrintFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Execute) {
	    keyCode = NSExecuteFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Insert) {
	    keyCode = NSInsertFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Undo) {
	    keyCode = NSUndoFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Redo) {
	    keyCode = NSRedoFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Menu) {
	    keyCode = NSMenuFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Find) {
	    keyCode = NSFindFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 

#if 0
	} else if (keysym == XK_Cancel) {
	    keyCode = NSSystemFunctionKey; /* ??? */		
#endif

	} else if (keysym == XK_Help) {
	    keyCode = NSHelpFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Break) {
	    keyCode = NSBreakFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask;    
	} else if (keysym == XK_Mode_switch) {
	    keyCode = NSModeSwitchFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 

#if 0	
	} else if (keysym == XK_script_switch) {
	    keyCode = NSSystemFunctionKey; /* ??? */
	} else if (keysym == XK_Num_Lock) {
	    keyCode = NSSystemFunctionKey; /* ??? */
#endif
	} else if (keysym == XK_Sys_Req) {
	    keyCode = NSSysReqFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Scroll_Lock) {
	    keyCode = NSScrollLockFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Pause) {
	    keyCode = NSPauseFunctionKey;		
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if (keysym == XK_Clear) {
	    keyCode = NSClearDisplayFunctionKey;
	    eventModifierFlags = eventModifierFlags | NSFunctionKeyMask; 
	} else if ((keysym == XK_Shift_L) || (keysym == XK_Shift_R)) {
	    eventModifierFlags = eventModifierFlags |
				 NSFunctionKeyMask |
				 NSShiftKeyMask; 
	} else if ((keysym == XK_Control_L) || (keysym == XK_Control_R)) {
	    eventModifierFlags = eventModifierFlags |
				 NSFunctionKeyMask |
				 NSControlKeyMask; 
	} else if (keysym == XK_Alt_R) {
	    eventModifierFlags = eventModifierFlags |
				 NSFunctionKeyMask |
				 NSAlternateKeyMask; 
	} else if (keysym == XK_Alt_L) {
	    eventModifierFlags = eventModifierFlags |
				 NSFunctionKeyMask |
				 NSCommandKeyMask; 
	}
    }
    
    /* If the key press originated from the key pad */
    
    if ((keysym >= XK_KP_Space) && (keysym <= XK_KP_9)) {
	eventModifierFlags = eventModifierFlags | NSNumericPadKeyMask;

	switch(keysym) {
	  case XK_KP_F1:        keyCode = NSF1FunctionKey;         break;
	  case XK_KP_F2:        keyCode = NSF2FunctionKey;         break;
	  case XK_KP_F3:        keyCode = NSF3FunctionKey;         break;
	  case XK_KP_F4:        keyCode = NSF4FunctionKey;         break;
	  case XK_KP_Home:      keyCode = NSHomeFunctionKey;       break;
	  case XK_KP_Left:      keyCode = NSLeftArrowFunctionKey;  break;
	  case XK_KP_Up:        keyCode = NSUpArrowFunctionKey;    break;
	  case XK_KP_Right:     keyCode = NSRightArrowFunctionKey; break;
	  case XK_KP_Down:      keyCode = NSDownArrowFunctionKey;  break;
	  case XK_KP_Prior:     keyCode = NSPrevFunctionKey;       break;
	    //case XK_KP_Page_Up:   keyCode = NSPageUpFunctionKey;     break;
	  case XK_KP_Next:      keyCode = NSNextFunctionKey;       break;
	    //case XK_KP_Page_Down: keyCode = NSPageDownFunctionKey;   break;
	  case XK_KP_End:       keyCode = NSEndFunctionKey;        break;
	  case XK_KP_Begin:     keyCode = NSBeginFunctionKey;      break;
	  case XK_KP_Insert:    keyCode = NSInsertFunctionKey;     break;
	  case XK_KP_Delete:    keyCode = NSDeleteFunctionKey;     break;
	    
	  default:  break;  /* Nothing to do */
	}
    } /* End of keypad stuff */

    if (((keysym >= XK_KP_Space) && (keysym <= XK_KP_9)) ||
	((keysym >= XK_space) && (keysym <= XK_asciitilde)))
    {	
	/* Not processed */
    }
}

/*
 * setModifierFlags determines which modifier keys (Command, Control,
 * Shift, and so forth) were held down while the event occured and
 * sets the eventModifierFlags instance variable accordingly.
 */
    
void
setModifierFlags(unsigned int state)
{
    eventModifierFlags = 0; /* Clear the mask */

    if (state & ControlMask) {
	eventModifierFlags = eventModifierFlags | NSControlKeyMask;
    }

    if (state & ShiftMask) {
	eventModifierFlags = eventModifierFlags | NSShiftKeyMask;
    }

    if (state & Mod1Mask) {
	eventModifierFlags = eventModifierFlags | NSCommandKeyMask; /* ? */
	fprintf(stdout, "setButtonModifierFlags(): Mod1Mask\n");
    }

    if (state & Mod2Mask) {
	eventModifierFlags = eventModifierFlags | NSAlternateKeyMask;/* ? */
	fprintf(stdout, "setButtonModifierFlags(): Mod2Mask\n");
    }

    if (state & Mod3Mask) {
	eventModifierFlags = eventModifierFlags | NSAlphaShiftKeyMask;/* ? */
	fprintf(stdout, "setButtonModifierFlags(): Mod3Mask\n");
    }

    if (state & Mod4Mask) {
	eventModifierFlags = eventModifierFlags | NSHelpKeyMask; /* ? */
	fprintf(stdout, "setButtonModifierFlags(): Mod4Mask\n");
    }

    if (state & Mod5Mask) {
	eventModifierFlags = eventModifierFlags | NSControlKeyMask; /* ? */
	fprintf(stdout, "setButtonModifierFlags(): Mod5Mask\n");
    }
}

void
updateKeyAndMainWindow(void)
{
    [NSApp updateKeyAndMainWindow];
}
